-- Configure MySQL server operation

SET sql_mode = 'ANSI';

-- Create database

DROP DATABASE IF EXISTS "unb2test";
CREATE DATABASE "unb2test" COLLATE 'utf8_bin';
USE "unb2test";

-- Create base relations

-- Key lists table, keeps lists of keys for several purposes on several other tables
CREATE TABLE "keylist" (
	-- All entries in a key list are referenced by the key list ID
	"KeylistId" INTEGER UNSIGNED NOT NULL AUTO_INCREMENT PRIMARY KEY)
ENGINE 'InnoDB' CHARACTER SET 'utf8' COLLATE 'utf8_bin';

-- Key lists table, keeps lists of keys for several purposes on several other tables
CREATE TABLE "keylist_keys" (
	-- All entries in a key list are referenced by the key list ID
	"KeylistId" INTEGER UNSIGNED NOT NULL,
	-- A key entry in the list
	"UserId" INTEGER UNSIGNED NOT NULL,
	PRIMARY KEY ("KeylistId", "UserId"))
ENGINE 'InnoDB' CHARACTER SET 'utf8' COLLATE 'utf8_bin';

-- Users table, keeps all registered user account and their additional keys
CREATE TABLE "user" (
	-- User ID, also called "key"
	"UserId" INTEGER UNSIGNED NOT NULL PRIMARY KEY,
	-- Time when this user was created
	"CreatedTime" DATETIME NOT NULL,
	-- Time when this user has last logged in
	"LastLoginTime" DATETIME,
	-- Display name
	"DisplayName" VARCHAR(255),
	-- Prepared display name for uniqueness
	"DisplayNamePrep" VARCHAR(255),
	-- ID of the key list that contains all additional keys that are assigned to this user
	"AdditionalKeylistId" INTEGER UNSIGNED,
	-- Login name
	"LoginName" VARCHAR(255),
	-- Login password, usually only a secure hash of it
	"LoginPassword" VARCHAR(255))
ENGINE 'InnoDB' CHARACTER SET 'utf8' COLLATE 'utf8_bin';
CREATE UNIQUE INDEX "user_DisplayNamePrep_unq" ON "user" ("DisplayNamePrep");
CREATE UNIQUE INDEX "user_LoginName_unq" ON "user" ("LoginName");

-- Users VCard table, keeps all users' vcard entries
CREATE TABLE "user_vcard" (
	-- ID of the user to which this vcard entry belongs
	"UserId" INTEGER UNSIGNED NOT NULL,
	-- Entry type
	-- (e.g. 'RealName', 'EMail', 'Birthday', 'ICQ', ...)
	"Type" VARCHAR(255) NOT NULL,
	-- Entry data
	"Data" TEXT NOT NULL,
	-- Determines to whom this vcard entry is visible
	-- (0 = public, 1 = internal, 2 = confidential, 3 = private)
	"Visibility" TINYINT UNSIGNED NOT NULL DEFAULT 0,
	PRIMARY KEY ("UserId", "Type"))
ENGINE 'InnoDB' CHARACTER SET 'utf8' COLLATE 'utf8_bin';
CREATE INDEX "user_vcard_UserId_idx" ON "user_vcard" ("UserId");

-- Users configuration table, keeps all users' configuration settings
CREATE TABLE "user_config" (
	-- ID of the user to which this config entry belongs
	"UserId" INTEGER UNSIGNED NOT NULL,
	-- Setting name
	"Name" VARCHAR(255) NOT NULL,
	-- Setting data
	"Data" VARCHAR(255) NOT NULL,
	PRIMARY KEY ("UserId", "Name"))
ENGINE 'InnoDB' CHARACTER SET 'utf8' COLLATE 'utf8_bin';
CREATE INDEX "user_config_UserId_idx" ON "user_config" ("UserId");

-- Messages table, keeps all messages
CREATE TABLE "message" (
	-- Message ID
	"MessageId" INTEGER UNSIGNED NOT NULL AUTO_INCREMENT PRIMARY KEY,
	-- Time when this message was created
	"CreatedTime" DATETIME NOT NULL,
	-- User to whom this message is assigned, usually the user who created the message
	"Owner" INTEGER UNSIGNED NOT NULL,
	-- Message ID on which this message primarily references, used for tree visualisation or replies
	"PrimaryReference" INTEGER UNSIGNED,
	-- Determines whether this message is a discussion entry point
	"EntryPoint" BOOLEAN NOT NULL DEFAULT FALSE,
	-- Message's page name for Wiki-like usage
	"PageName" VARCHAR(255),
	-- Determines whether this message is a draft
	"Draft" BOOLEAN NOT NULL DEFAULT FALSE,
	-- ID of the key list that contains all users that are allowed to read this message (if not NULL)
	"ReadAccessKeylistId" INTEGER UNSIGNED,
	-- ID of the key list that contains all users that are allowed to alter this message (if not NULL)
	"AlterAccessKeylistId" INTEGER UNSIGNED,
	-- ID of the key list that contains all users that are allowed to reply to this message (if not NULL)
	"ReplyAccessKeylistId" INTEGER UNSIGNED,
	-- Moderation flags of the entire message
	-- (1 = Locked, 2 = Hidden, 4 = EnforceApproval)
	"ModerationFlags" TINYINT UNSIGNED NOT NULL DEFAULT 0,
	-- Determines whether this message has attached data (is a file)
	"FileType" TINYINT UNSIGNED NOT NULL DEFAULT 0,
	-- Content of this message that shall be regarded for searching, usually a simplified plain-text version of the latest approved message revision
	"SearchOptimisedContent" MEDIUMTEXT,
	-- Subject of this message that shall be regarded for searching, usually of the latest approved message revision
	"SearchOptimisedSubject" VARCHAR(255),
	-- Message revision ID that shall be regarded for tag-related searching, usually of the latest approved message revision
	"SearchRevisionId" INTEGER UNSIGNED,
	-- Time when a user last began editing this message, used for a warning message to potential concurrent editors
	"LastBeginEdit" DATETIME)
ENGINE 'InnoDB' CHARACTER SET 'utf8' COLLATE 'utf8_bin';
CREATE UNIQUE INDEX "message_PageName_unq" ON "message" ("PageName");

-- Message references table, keeps all references of a message
CREATE TABLE "message_reference" (
	-- ID of the message that is referencing another message
	"MessageId" INTEGER UNSIGNED NOT NULL,
	-- ID of the message that is being referenced
	"Reference" INTEGER UNSIGNED NOT NULL,
	PRIMARY KEY ("MessageId", "Reference"))
ENGINE 'InnoDB' CHARACTER SET 'utf8' COLLATE 'utf8_bin';

-- Message revisions table, keeps all revisions of a message
CREATE TABLE "message_revision" (
	-- Message revision ID
	"MessageRevisionId" INTEGER UNSIGNED NOT NULL AUTO_INCREMENT PRIMARY KEY,
	-- ID of the message that this revision belongs to
	"MessageId" INTEGER UNSIGNED NOT NULL,
	-- Time when this revision was created
	"CreatedTime" DATETIME NOT NULL,
	-- ID of the user who created this revision
	"Author" INTEGER UNSIGNED NOT NULL,
	-- Subject of the message (revision)
	"Subject" VARCHAR(255),
	-- Content
	"Content" MEDIUMTEXT,
	-- Cached HTML output of the parsed content, used to accelerate displaying of the message (revision)
	"HtmlContent" MEDIUMTEXT,
	-- Summary of the changes in this revivion to the previous one
	"Summary" VARCHAR(255),
	-- Moderation state of this single revision
	-- (0 = WaitingForApproval, 1 = Approved, 2 = Locked)
	"ModerationState" TINYINT UNSIGNED NOT NULL DEFAULT 0,
	-- Determines whether this message revision is a draft
	"Draft" BOOLEAN NOT NULL DEFAULT FALSE,
	-- MIME content type of attached data
	"ContentType" VARCHAR(255),
	-- Data of the attached file, if not stored in a separate file in the local filesystem
	"Data" MEDIUMBLOB)
ENGINE 'InnoDB' CHARACTER SET 'utf8' COLLATE 'utf8_bin';
CREATE INDEX "message_revision_CreatedTime_idx" ON "message_revision" ("CreatedTime");

-- Tags table, keeps all known tags
CREATE TABLE "tag" (
	-- Tag ID
	"TagId" INTEGER UNSIGNED NOT NULL AUTO_INCREMENT PRIMARY KEY,
	-- Displayed name of the tag
	"TagName" VARCHAR(255) NOT NULL,
	-- ID of the parent tag for this tag, used to organise tags in a tree hierarchy
	"ParentTag" INTEGER UNSIGNED,
	-- Special meaning of this tag
	"Meaning" TINYINT UNSIGNED,
	-- Description of this tag that explains the user when and why this tag should be used
	"Description" VARCHAR(255),
	-- ID of the key list that contains all users that are allowed to use/set this tag in messages (if not NULL)
	"UseAccessKeylistId" INTEGER UNSIGNED,
	-- ID of the key list that contains all users that are allowed to read messages that have this tag set (if not NULL)
	"ReadAccessKeylistId" INTEGER UNSIGNED,
	-- Flags for special tag behaviour
	-- (1 = NotSelectable, 2 = Invisible, 4 = EnforceApproval, 8 = UserData)
	"Flags" TINYINT UNSIGNED NOT NULL DEFAULT 0)
ENGINE 'InnoDB' CHARACTER SET 'utf8' COLLATE 'utf8_bin';

-- Message revisions' tags table, keeps all tag associations for messages
CREATE TABLE "message_revision_tag" (
	-- ID of the message revision that this tag is set for
	"MessageRevisionId" INTEGER UNSIGNED NOT NULL,
	-- ID of the tag that is set for the message
	"TagId" INTEGER UNSIGNED NOT NULL,
	-- Data that is set for this tag and the message (if tag.Flags has UserData set)
	"TagData" VARCHAR(255),
	PRIMARY KEY ("MessageRevisionId", "TagId"))
ENGINE 'InnoDB' CHARACTER SET 'utf8' COLLATE 'utf8_bin';

-- Message revisions' embedded files table, keeps all file attachment associations for messages
CREATE TABLE "message_revision_file" (
	-- ID of the message revision that this association is set for
	"MessageRevisionId" INTEGER UNSIGNED NOT NULL,
	-- ID of the (file) message that is embedded in the message
	"MessageId" INTEGER UNSIGNED NOT NULL,
	-- Relation type of the file to the message revision
	"Relation" VARCHAR(255),
	PRIMARY KEY ("MessageRevisionId", "MessageId"))
ENGINE 'InnoDB' CHARACTER SET 'utf8' COLLATE 'utf8_bin';

-- Rating table, keeps all information about users rating messages
CREATE TABLE "message_rating" (
	-- ID of the user that has rated the message
	"UserId" INTEGER UNSIGNED NOT NULL,
	-- ID of the message that is rated
	"MessageId" INTEGER UNSIGNED NOT NULL,
	-- Rating value, where 0 is a neutral rating, positive values are good, negative are bad
	"Rating" TINYINT NOT NULL,
	-- Rating weight that tells how relevant a single rating is for the total rating
	"Weight" TINYINT UNSIGNED NOT NULL,
	PRIMARY KEY ("UserId", "MessageId"))
ENGINE 'InnoDB' CHARACTER SET 'utf8' COLLATE 'utf8_bin';
CREATE INDEX "message_rating_MessageId_idx" ON "message_rating" ("MessageId");

-- Rating table, keeps all information about users rating messages
CREATE TABLE "log" (
	-- Log entry number
	"Num" INTEGER UNSIGNED NOT NULL AUTO_INCREMENT PRIMARY KEY,
	-- Date/time of the log entry
	"Time" DATETIME NOT NULL,
	-- Event type
	"Type" TINYINT UNSIGNED NOT NULL,
	-- Event ID
	"EventId" INTEGER UNSIGNED NOT NULL,
	-- User ID, intentionally not referencing the user table
	"UserId" INTEGER UNSIGNED,
	-- Client IP address
	"ClientAddress" VARCHAR(100),
	-- Event data in XML format
	"Data" VARCHAR(65535))
ENGINE 'InnoDB' CHARACTER SET 'utf8' COLLATE 'utf8_bin';
CREATE INDEX "log_Time_idx" ON "log" ("Time");
CREATE INDEX "log_EventId_idx" ON "log" ("EventId");

-- Create foreign keys for referential integrity

ALTER TABLE "keylist_keys" ADD FOREIGN KEY ("KeylistId") REFERENCES "keylist" ("KeylistId") ON DELETE CASCADE;
ALTER TABLE "keylist_keys" ADD FOREIGN KEY ("UserId") REFERENCES "user" ("UserId") ON DELETE CASCADE;

ALTER TABLE "user" ADD FOREIGN KEY ("AdditionalKeylistId") REFERENCES "keylist" ("KeylistId") ON DELETE SET NULL;

ALTER TABLE "user_vcard" ADD FOREIGN KEY ("UserId") REFERENCES "user" ("UserId") ON DELETE CASCADE;

ALTER TABLE "user_config" ADD FOREIGN KEY ("UserId") REFERENCES "user" ("UserId") ON DELETE CASCADE;

ALTER TABLE "message" ADD FOREIGN KEY ("Owner") REFERENCES "user" ("UserId");
ALTER TABLE "message" ADD FOREIGN KEY ("PrimaryReference") REFERENCES "message" ("MessageId") ON DELETE SET NULL;
ALTER TABLE "message" ADD FOREIGN KEY ("ReadAccessKeylistId") REFERENCES "keylist" ("KeylistId") ON DELETE SET NULL;
ALTER TABLE "message" ADD FOREIGN KEY ("AlterAccessKeylistId") REFERENCES "keylist" ("KeylistId") ON DELETE SET NULL;
ALTER TABLE "message" ADD FOREIGN KEY ("ReplyAccessKeylistId") REFERENCES "keylist" ("KeylistId") ON DELETE SET NULL;
ALTER TABLE "message" ADD FOREIGN KEY ("SearchRevisionId") REFERENCES "message_revision" ("MessageRevisionId");

ALTER TABLE "message_reference" ADD FOREIGN KEY ("MessageId") REFERENCES "message" ("MessageId") ON DELETE CASCADE;
ALTER TABLE "message_reference" ADD FOREIGN KEY ("Reference") REFERENCES "message" ("MessageId") ON DELETE CASCADE;

ALTER TABLE "message_revision" ADD FOREIGN KEY ("MessageId") REFERENCES "message" ("MessageId") ON DELETE CASCADE;
ALTER TABLE "message_revision" ADD FOREIGN KEY ("Author") REFERENCES "user" ("UserId");

ALTER TABLE "tag" ADD FOREIGN KEY ("ParentTag") REFERENCES "tag" ("TagId");
ALTER TABLE "tag" ADD FOREIGN KEY ("UseAccessKeylistId") REFERENCES "keylist" ("KeylistId") ON DELETE SET NULL;
ALTER TABLE "tag" ADD FOREIGN KEY ("ReadAccessKeylistId") REFERENCES "keylist" ("KeylistId") ON DELETE SET NULL;

ALTER TABLE "message_revision_tag" ADD FOREIGN KEY ("MessageRevisionId") REFERENCES "message_revision" ("MessageRevisionId") ON DELETE CASCADE;
ALTER TABLE "message_revision_tag" ADD FOREIGN KEY ("TagId") REFERENCES "tag" ("TagId") ON DELETE CASCADE;

ALTER TABLE "message_revision_file" ADD FOREIGN KEY ("MessageRevisionId") REFERENCES "message_revision" ("MessageRevisionId") ON DELETE CASCADE;
ALTER TABLE "message_revision_file" ADD FOREIGN KEY ("MessageId") REFERENCES "message" ("MessageId") ON DELETE CASCADE;

ALTER TABLE "message_rating" ADD FOREIGN KEY ("UserId") REFERENCES "user" ("UserId") ON DELETE CASCADE;
ALTER TABLE "message_rating" ADD FOREIGN KEY ("MessageId") REFERENCES "message" ("MessageId") ON DELETE CASCADE;

-- Create triggers

DELIMITER |
CREATE TRIGGER "keylist_empty_trg"
AFTER DELETE ON "keylist_keys" FOR EACH ROW
	IF NOT EXISTS (SELECT "KeylistId" FROM "keylist_keys" WHERE "KeylistId" = OLD."KeylistId") THEN
		DELETE FROM "keylist" WHERE "KeylistId" = OLD."KeylistId";
	END IF;
|
DELIMITER ;

