<?php
// UNB Components
// Copyright since 2007 by Yves Goergen
// Website: http://unclassified.de/go/unb2
// Licence: GNU General Public License (GPL) Version 3
//
// This file contains the UnbDatabase class.

if (!class_exists('UnbConfiguration', false)) require_once dirname(__FILE__) . '/configuration.class.php';
if (!class_exists('UnbDatabaseException', false)) require_once dirname(__FILE__) . '/unb.class.php';
if (!class_exists('UnbUser', false)) require_once dirname(__FILE__) . '/user.class.php';

// The database connection class provides basic query abstraction support and methods to easily
// execute SQL statements and query record sets, records or single values with an SQL query.
//
// It establishes the connection to the database provider with PHP Data Objects (PDO) which
// technically allows using diverse database backends. Still the submitted queries must be
// formulated such that each DBMS to be used understands them.
//
// UnbDatabase is a singleton class, there can only be one single instance of it during runtime.
// This way, the database connection can be used from multiple classes without passing its
// reference everywhere.
//
class UnbDatabase
{
	// Display queries and error messages when an error occurs.
	const DisplayQueryErrors = true;
	// Display information for all issued queries. (Ineffective if DisplayQueryErrors is {{false}}.)
	const DisplayQueryInfo = false;

	// Singleton instance
	private static $instance = null;

	// Array with information on how to connect to the database
	private $connectData = null;
	// PDO connection
	private $connection = null;
	// PDO driver name for the current database connection
	private $driverName = '';
	// Table prefix, allows storing several UNB tables instances in a single database (schema)
	private $tablePrefix = '';
	// Current transaction nesting level (only one transaction at a time is used in the database backend)
	private $transactionLevel = 0;
	// Remembers whether the database is locked for the current transaction (SQLite only)
	private $databaseLocked = false;
	// Statistics data: Number of queries run in this session
	private $queryCount = 0;
	// Statistics data: Total time that all queries took
	private $totalTime = 0.0;
	// Cached value: Number of rows affected by the last SQL statement (only for DELETE, INSERT and UPDATE)
	private $rowsAffected = 0;
	// Indicates whether the last query caused a constraint violation
	private $constraintError = false;
	// Indicates whether the last query ran against a lock and should be repeated (SQLite only)
	private $busyError = false;
	// Last error message
	private $errorMessage = null;
	// Connection ID (for debugging concurrent connections, not supported by every database driver)
	private $connectionId = 0;
	// Last generated auto value (from auto increment column or implicitly used sequence, triggered
	// by a UnbDbAutoValue parameter)
	private $lastAutoValue = null;

	// Private constructor, does nothing but preventing an unwanted instantiation of this class.
	//
	private function __construct()
	{
	}

	// Gets the instance of this class.
	//
	// If called for the first time, the instance is created.
	//
	// Returns the only UnbDatabase instance
	//
	public static function GetInstance()
	{
		if (!isset(self::$instance))
		{
			self::$instance = new self();
		}
		return self::$instance;
	}

	// Gets a separate instance of this class.
	//
	// This should be used only when needed, like for the logging functions.
	//
	// Returns a new UnbDatabase instance
	//
	public static function GetNewInstance()
	{
		return new self();
	}

	// Gets a value indicating whether the database connection is open.
	//
	// Returns (bool)
	//
	public function IsOpen()
	{
		return isset($this->connection);
	}

	// Gets the driver name used for the database connection.
	//
	// Returns (string)
	//
	public function GetDriverName()
	{
		return $this->driverName;
	}

	// Gets statistic data.
	//
	// Returns an array with the keys 'queryCount' (number of queries issued) and
	// 'totalTime' (time for all queries in seconds)
	//
	public function GetStatistics()
	{
		return array(
			'queryCount' => $this->queryCount,
			'totalTime' => $this->totalTime);
	}

	// Opens a database connection.
	//
	// connectData = (array) Database connection configuration array containing the following items:
	//                       * 'driver' = PDO driver to be used ('mysql', 'pgsql' or 'sqlite')
	//                       * 'dsn' = DSN string (only with ODBC)
	//                       * 'hostname' = Database server hostname (not with SQLite)
	//                       * 'port' = Database server port number (optional; not with SQLite)
	//                       * 'unix-socket' = Database server UNIX socket (optional; only with MySQL, not with hostname/port)
	//                       * 'dbname' = Database name (not with SQLite)
	//                       * 'username' = Connection login username (not with SQLite)
	//                       * 'password' = Connection login password (not with SQLite)
	//                       * 'filename' = Database filename, relative to this file (only with SQLite)
	//                       * 'schema' = Database schema name (only with pgsql)
	//                       * 'synchronous' = Disk synchronisation mode (optional; 'OFF', 'NORMAL' or 'FULL'; only with SQLite)
	//                       * 'table-prefix' = Table name prefix (optional)
	//
	// Throws InvalidArgumentException
	// Throws UnbDatabaseException
	// Throws PDOException
	//
	public function Open($connectData = null)
	{
		if (isset($connectData))
		{
			if (!is_array($connectData))
				throw new InvalidArgumentException('Invalid argument type: $connectData');
		}
		else
		{
			// No database connection configuration is set, go read it from the global configuration.
			$entries = UnbConfiguration::Get('database.*');
			$connectData = array();
			foreach ($entries as $name => $data)
			{
				$connectData[substr($name, 9)] = $data;
			}
		}

		$this->connectData = $connectData;
		if ($connectData['table-prefix'] != null)
			$this->tablePrefix = $connectData['table-prefix'];

		switch ($this->connectData['driver'])
		{
			case 'mysql':
				$dsn = 'mysql:dbname=' . $this->connectData['dbname'] .
					(isset($this->connectData['hostname']) ? ';host=' . $this->connectData['hostname'] : '') .
					(isset($this->connectData['port']) ? ';port=' . $this->connectData['port'] : '') .
					(isset($this->connectData['unix-socket']) ? ';unix_socket=' . $this->connectData['unix-socket'] : '');
				break;

			case 'odbc':
				$dsn = 'odbc:' . $this->connectData['dsn'];
				break;

			case 'pgsql':
				$dsn = 'pgsql:host=' . $this->connectData['hostname'] .
					(isset($this->connectData['port']) ? ';port=' . $this->connectData['port'] : '') .
					';dbname=' . $this->connectData['dbname'];
				break;

			case 'sqlite':
				if (!file_exists(dirname(__FILE__) . '/' . $this->connectData['filename']))
					throw new UnbDatabaseException('[UNB#37] Database file does not exist');
				if (!filesize(dirname(__FILE__) . '/' . $this->connectData['filename']))
					throw new UnbDatabaseException('[UNB#37] Database file is empty (0 bytes long)');
				$dsn = 'sqlite:' . dirname(__FILE__) . '/' . $this->connectData['filename'];
				break;

			default:
				throw new InvalidArgumentException('Unsupported database driver name. Check your configuration.');
		}
		$this->driverName = $this->connectData['driver'];

		$parameters = array();
		if ($connectData['persistent'])
			$parameters[PDO::ATTR_PERSISTENT] = true;

		$this->connection = new PDO($dsn, $this->connectData['username'], $this->connectData['password'], $parameters);

		// Driver-specific configuration
		switch ($this->driverName)
		{
			case 'mysql':
				$minVersion = '5.0.15';
				$version = $this->QueryValue('SELECT VERSION()');
				if (version_compare($version, $minVersion) < 0)
					throw new Exception('Unsupported MySQL server version ' . $version . '; version ' . $minVersion . ' or newer is required.');

				$this->connection->exec('SET sql_mode = \'ANSI\'');
				$this->connection->exec('SET storage_engine = \'InnoDB\'');
				$this->connection->exec('SET CHARACTER SET \'utf8\'');
				$this->connection->exec('SET NAMES \'utf8\'');
				$this->connectionId = $this->QueryValue('SELECT CONNECTION_ID()');
				break;

			case 'pgsql':
				if (isset($this->connectData['schema']))
					$this->connection->exec('SET search_path TO ' . $this->connection->quote($this->connectData['schema']));
				break;

			case 'sqlite':
				$minVersion = '3.6.19';
				$version = $this->QueryValue('SELECT sqlite_version()');
				if (version_compare($version, $minVersion) < 0)
					throw new Exception('Unsupported SQLite library version ' . $version . '; version ' . $minVersion . ' or newer is required.');

				if (isset($this->connectData['synchronous']))
					$this->connection->exec('PRAGMA synchronous = ' . $this->connectData['synchronous']);
				$this->connection->exec('PRAGMA foreign_keys = ON');
				// Register a user-defined function with SQLite to perform Unicode-safe LOWER()
				// operations on strings. Since PHP comes with such a feature but SQLite does not,
				// we make SQLite use PHP's function here.
				$this->connection->sqliteCreateFunction('LOWER', 'mb_strtolower', 1);
				$this->connection->sqliteCreateFunction('EXTRACTVALUE', array(__CLASS__, 'SQLite_ExtractValue'), 2);
				break;
		}
	}

	// Closes the database connection.
	//
	public function Close()
	{
		$this->connection = null;
	}

	// SQLite UDF (user defined function) for XML access.
	//
	public static function SQLite_ExtractValue($xml, $xpath)
	{
		$xml = new SimpleXMLElement($xml);
		$data = $xml->xpath($xpath);
		if (is_array($data) && count($data) > 0)
			return $data[0];
		return '';
	}

	// Shows query information and errors.
	//
	// ps = (PDO/PDOStatement) Prepared statement or PDO connection used to retrieve the error details
	// r = (bool) Result value. {{true}}: The query was successful. {{false}}: There was an error that should be displayed.
	// parameters = (array) Bound parameters used to execute the SQL query.
	//                      Those will be inserted for every "?" when displaying the SQL query.
	// time = (float) Time in seconds it took to execute the query.
	//
	// Uses UnbDatabase::DisplayQueryInfo
	//
	private function ShowQueryInfo($ps, $r, $parameters, $time)
	{
		if (!$r || self::DisplayQueryInfo)
		{
			echo '<div class="sql-info" onclick="alert(&quot;' . htmlspecialchars(str_replace("\n", '\n', addslashes(UnbDebug::GetCallTrace(1)))) . '&quot;)"><b>SQL execute:</b> ';
			if ($ps instanceof PDOStatement)
			{
				$sql = htmlspecialchars($ps->queryString, ENT_NOQUOTES);
			}
			else if ($ps instanceof PDO)
			{
				$sql = htmlspecialchars($parameters, ENT_NOQUOTES);
				$parameters = null;
			}
			if ($parameters)
			{
				foreach ($parameters as $p)
				{
					if (!isset($p))
						$p = 'NULL';
					else if (is_bool($p))
						$p = $p ? '1' : '0';
					else if (!is_numeric($p))
						$p = '\'' . htmlspecialchars(mb_substr($p, 0, 30), ENT_NOQUOTES) . '\'';
					$sql = preg_replace('_\?_', '<i>' . $p . '</i>', $sql, 1);
				}
			}
			echo $sql;
			if ($time >= 0)
			{
				echo ' &ndash; ' . ($time >= 0.005 ? '<b>' : '') . round($time * 1000, 1) . 'ms' . ($time >= 0.005 ? '</b>' : '');
			}
			echo '</div>';
		}

		if (!$r)
		{
			// $ps may be a PDOStatement or a PDO object.
			$ei = $ps->errorInfo();
			echo '<div class="sql-error"><b>SQL error:</b> [SQLSTATE:' . $ei[0] . ', ' . $ei[1] . '] ' . htmlspecialchars($ei[2], ENT_NOQUOTES) . '</div>';
		}
	}

	// Finds out what the problem was. :-)
	//
	// ps = (PDO/PDOStatement) Prepared statement or PDO connection used to retrieve the error details
	//
	private function DetermineErrorCondition($ps)
	{
		// Reset error condition
		$this->constraintError = false;
		$this->busyError = false;

		$ei = $ps->errorInfo();

		// Evaluate SQLSTATE value, class "23": Constraint violations
		if (substr($ei[0], 0, 2) == '23')
			$this->constraintError = true;

		// Driver-specific error codes
		switch ($this->driverName)
		{
			case 'sqlite':
				if ($ei[1] == 5)   // SQLITE_BUSY = 5
				{
					$this->busyError = true;
					// Wait for a while so that the query can then be repeated immediately
					usleep(mt_rand(2000, 20000));   // 2 ... 20 ms
				}
				else if ($ei[1] == 19)   // SQLITE_CONSTRAINT = 19
				{
					$this->constraintError = true;
				}
				break;
		}

		$this->errorMessage = $ei[2];
	}

	// Gets a value indicating whether the last error condition describes a constraint violation.
	//
	// This value is only valid if the previous query actually failed.
	//
	// Returns (bool) value
	//
	public function IsConstraintError()
	{
		return $this->constraintError;
	}

	// Gets the message of the last occured database error.
	//
	// Returns (string) Error message
	//
	public function GetErrorMessage()
	{
		return $this->errorMessage;
	}

	// Optimises the database.
	//
	// extensive = (bool) Set this to {{true}}, in order to optimise the database more thoroughly.
	//                    This may take a much longer time for large databases. If {{false}}, merely
	//                    the index statistics are updated which can improve performance but
	//                    does not save disk space.
	//
	// Returns {{true}}, if successful, {{false}} otherwise.
	//
	public function Optimise($extensive = false)
	{
		$r = true;
		switch ($this->driverName)
		{
			case 'mysql':
				if ($extensive)
				{
					$r &= $this->Query('OPTIMIZE TABLE "#keylist", "#keylist_keys", "#log", "#message", "#message_rating", "#message_reference", "#message_revision", "#message_revision_tag", "#tag", "#user", "#user_config", "#user_vcard"');
					// For InnoDB: OPTIMIZE TABLE rebuilds the table to update index statistics
					// and free unused space in the clustered index
				}
				else
				{
					$r &= $this->Query('ANALYZE TABLE "#keylist", "#keylist_keys", "#log", "#message", "#message_rating", "#message_reference", "#message_revision", "#message_revision_tag", "#tag", "#user", "#user_config", "#user_vcard"');
					// ANALYZE TABLE analyzes and stores the key distribution for a table.
					// During the analysis, the table is locked with a write lock for InnoDB.
				}
				break;
			case 'pgsql':
				if ($extensive)
				{
					$r &= $this->Execute('VACUUM FULL');
					$r &= $this->Execute('REINDEX DATABASE ?', $this->connectData['database']);
				}
				else
				{
					$r &= $this->Execute('VACUUM');
					$r &= $this->Execute('ANALYZE');
				}
				break;
			case 'sqlite':
				if ($extensive)
				{
					$r &= $this->Execute('VACUUM');
					// VACUUM rebuilds the database file. This can be delayed (blocking) when a
					// transaction is active.
					$r &= $this->Execute('ANALYZE');
				}
				else
				{
					$r &= $this->Execute('ANALYZE');
				}
				break;
		}
		return $r;
	}

	// Starts a new transaction.
	//
	// If a transaction is already started, only the counter is increased by 1, but no real nested
	// transaction is started.
	//
	// lock = (bool) Lock the entire database. This option is only regarded for database system
	//               that don't support table or row-level locking, like SQLite. It can only be
	//               used when no other transaction is currently started (i.e. the transaction
	//               level counter is 0).
	//
	// Throws BadMethodCallException when trying to acquire a lock for the transaction if a
	// transaction has already started.
	//
	// Returns {{true}}, if successful, {{false}} otherwise.
	//
	public function StartTransaction($lock = false)
	{
		if (!isset($this->connection))
			throw new BadMethodCallException('Database not opened');
		if ($lock && $this->transactionLevel > 0 && !$this->databaseLocked)
			throw new BadMethodCallException('You cannot lock during a transaction that has no lock.');

		$r = true;
		if ($this->transactionLevel == 0)
		{
			$start = microtime(true);
			do
			{
				if ($lock)
				{
					// Driver-specific configuration
					switch ($this->driverName)
					{
						case 'sqlite':
							ignore_user_abort(true);
							$ps = $this->connection->prepare('BEGIN EXCLUSIVE TRANSACTION');
							$r &= $ps->execute();
							break;
						default:
							$r &= $this->connection->beginTransaction();
							break;
					}
					$this->databaseLocked = true;
				}
				else
				{
					$r &= $this->connection->beginTransaction();
				}
				if (!$r) $this->DetermineErrorCondition($this->connection);
			}
			while (!$r && $this->busyError);   // SQLite busy loop workaround
			$time = microtime(true) - $start;
			$this->totalTime += $time;
			if (self::DisplayQueryErrors)
				$this->ShowQueryInfo($this->connection, $r, 'START TRANSACTION', $time);
		}
		else if (self::DisplayQueryInfo)
		{
			echo '<div class="sql-info" onclick="alert(&quot;' . htmlspecialchars(str_replace("\n", '\n', addslashes(UnbDebug::GetCallTrace()))) . '&quot;)"><b>UnbDatabase operation:</b> Called StartTransaction, changed transaction level from ' . $this->transactionLevel . ' to ' . ($this->transactionLevel + 1) . '.</div>';
		}
		$this->transactionLevel++;
		return $r;
	}

	// Rolls back a transaction.
	//
	// If the transaction counter is greater than 1, the transaction is not actually rolled back.
	// It is the callers' responsibility to pass on the error condition and call RollBack() as
	// often as required to really roll back the transaction.
	//
	// Throws BadMethodCallException if not in a transaction.
	//
	// Returns {{true}}, if successful, {{false}} otherwise.
	//
	public function RollBack()
	{
		if (!isset($this->connection))
			throw new BadMethodCallException('Database not opened');
		if ($this->transactionLevel == 0)
			throw new BadMethodCallException('No transaction is active');

		$r = true;
		$this->transactionLevel--;
		if ($this->transactionLevel == 0)
		{
			$start = microtime(true);
			if ($this->databaseLocked)
			{
				// Driver-specific configuration
				switch ($this->driverName)
				{
					case 'sqlite':
						// Transactions not started through the PDO API cannot be ended through it.
						$ps = $this->connection->prepare('ROLLBACK');
						$r &= $ps->execute();
						ignore_user_abort(false);
						break;
					default:
						$r &= $this->connection->rollBack();
						break;
				}
			}
			else
			{
				$r &= $this->connection->rollBack();
			}
			$time = microtime(true) - $start;
			$this->totalTime += $time;
			if (self::DisplayQueryErrors)
				$this->ShowQueryInfo($this->connection, $r, 'ROLLBACK', $time);
			$this->databaseLocked = false;
		}
		else if (self::DisplayQueryInfo)
		{
			echo '<div class="sql-info" onclick="alert(&quot;' . htmlspecialchars(str_replace("\n", '\n', addslashes(UnbDebug::GetCallTrace()))) . '&quot;)"><b>UnbDatabase operation:</b> Called Rollback, changed transaction level from ' . ($this->transactionLevel + 1) . ' to ' . $this->transactionLevel . '.</div>';
		}
		return $r;
	}

	// Commits a transaction.
	//
	// If the transaction counter is greater than 1, the transaction is not actually commited.
	//
	// Throws BadMethodCallException if not in a transaction.
	//
	// Returns {{true}}, if successful, {{false}} otherwise.
	//
	public function Commit()
	{
		if (!isset($this->connection))
			throw new BadMethodCallException('Database not opened');
		if ($this->transactionLevel == 0)
			throw new BadMethodCallException('No transaction is active');

		$r = true;
		$this->transactionLevel--;
		if ($this->transactionLevel == 0)
		{
			$start = microtime(true);
			if ($this->databaseLocked)
			{
				// Driver-specific configuration
				switch ($this->driverName)
				{
					case 'sqlite':
						// Transactions not started through the PDO API cannot be ended through it.
						$ps = $this->connection->prepare('COMMIT');
						$r &= $ps->execute();
						if ($r) ignore_user_abort(false);   // Committing can fail. Transactions are then rolled back.
						break;
					default:
						$r &= $this->connection->commit();
						break;
				}
			}
			else
			{
				$r &= $this->connection->commit();
			}
			$time = microtime(true) - $start;
			$this->totalTime += $time;
			if (!$r) $this->DetermineErrorCondition($this->connection);
			if (self::DisplayQueryErrors)
				$this->ShowQueryInfo($this->connection, $r, 'COMMIT', $time);
			$this->databaseLocked = false;
		}
		else if (self::DisplayQueryInfo)
		{
			echo '<div class="sql-info" onclick="alert(&quot;' . htmlspecialchars(str_replace("\n", '\n', addslashes(UnbDebug::GetCallTrace()))) . '&quot;)"><b>UnbDatabase operation:</b> Called Commit, changed transaction level from ' . ($this->transactionLevel + 1) . ' to ' . $this->transactionLevel . '.</div>';
		}
		return $r;
	}

	// Gets the next value of a sequence.
	//
	// table = (string) Table name for which the sequence was created.
	//
	// Not all database systems support sequences. Those who don't will use auto increment values
	// instead. This function returns null for them.
	//
	private function GetSequenceValue($table)
	{
		// Driver-specific configuration
		switch ($this->driverName)
		{
			case 'pgsql':
				return $this->QueryValue('nextval(\'' . $this->tablePrefix . $table . '_seq\')');
		}
		return null;
	}

	// Resets the last auto value.
	//
	private function ResetLastAutoValue()
	{
		$this->lastAutoValue = null;
	}

	// Fetches the last inserted auto increment value from the database connection and stores it
	// for later retrieval with GetLastAutoValue.
	//
	// Not all database systems support auto values. Those who don't will have to explicitly get a
	// sequence value and assign it before executing the query. This function does nothing for
	// them.
	//
	private function FetchLastAutoValue()
	{
		// Driver-specific configuration
		switch ($this->driverName)
		{
			case 'mysql':
			case 'sqlite':
				$this->lastAutoValue = intval($this->connection->lastInsertId());
				break;
			// NOTE: MSSQL needs a special handling, see http://de.php.net/manual/de/pdo.lastinsertid.php#87589
		}
	}

	// Gets the auto value generated by the last query.
	//
	// This will only be set if the last query has set a UnbDbAutoValue parameter.
	//
	// Returns (int) Last generated auto-increment/sequence value.
	//
	public function GetLastAutoValue()
	{
		return $this->lastAutoValue;
	}

	// Basic SQL query processing.
	//
	// This method strips out SQL comments (they could crash PHP/Apache during development) and
	// replaces the "#" placeholder with the specified table name prefix (if any).
	//
	// sql = (string) SQL query to process
	//
	// Returns the processed and (if appropriate) modified SQL query.
	//
	private function ProcessSql($sql)
	{
		// Driver-specific configuration
		switch ($this->driverName)
		{
			case 'sqlite':
				// SELECT ... FOR UPDATE is not supported. Instead, the database should be locked
				// for the transaction.
				if (preg_match('_^\s*SELECT\s+.+(\s+FOR\s+UPDATE)$_is', $sql, $m))
				{
					if (!$this->databaseLocked && self::DisplayQueryErrors)
					{
						echo '<div style="color: gray; font-size: 0.85em;"><b>SQL execute:</b> ' . htmlspecialchars($sql) . '</div>';
						echo '<div style="color: red; font-size: 0.85em;"><b>SQL warning:</b> A SELECT ... FOR UPDATE query was issued but the database was not locked for the transaction.</div>';
					}
					$sql = substr($sql, 0, -strlen($m[1]));
				}
				break;
		}

		return preg_replace('_\s*--\s.*?$_m', '',
			preg_replace('_(?<=")#(?=[a-zA-Z])_', $this->tablePrefix, $sql));
	}

	// Prepares an SQL statement.
	//
	// sql = (string) SQL statement.
	//
	// Throws InvalidArgumentException on invalid argument type.
	// Throws BadMethodCallException if the database is not opened.
	// Throws UnbDatabaseException if the statement could not be prepared.
	//
	// Returns a PDOStatement object
	//
	public function Prepare($sql)
	{
		if (!is_string($sql))
			throw new InvalidArgumentException('Invalid argument type: $sql');
		if (!isset($this->connection))
			throw new BadMethodCallException('Database not opened');

		$sql = $this->ProcessSql($sql);
		$ps = $this->connection->prepare($sql);
		if (!$ps)
		{
			$this->DetermineErrorCondition($this->connection);
			$this->ShowQueryInfo($this->connection, false, $sql, 0);
			throw new UnbDatabaseException('Statement could not be prepared.');
		}
		return $ps;
	}

	// Actually executes an SQL query.
	//
	// sql = (string) SQL statement
	//       (PDOStatement) Already prepared statement
	// parameters = (array) Parameters to be used in this SQL statement, {{null}} if no parameters are used.
	//                      This can only be (an array of) scalar values, UNB's objects are not allowed!
	//
	// Throws InvalidArgumentException on invalid argument type.
	// Throws BadMethodCallException if the database is not opened.
	// Throws UnbDatabaseException if the statement could not be prepared or executed.
	//
	// Returns the PDOStatement object to be used for further data retrieval.
	//
	private function InternalExecute($sql, $parameters)
	{
		if (!isset($this->connection))
			throw new BadMethodCallException('Database not opened');

		// TODO: Remove test code
		$start0 = microtime(true);
		$start = microtime(true);
		if (abs($start - $start0) > 1)
		{
			var_dump($start0);
			var_dump($start);
		}
		// Prepare SQL statement if not yet done
		if (is_string($sql))
		{
			$ps = $this->Prepare($sql);
		}
		else if ($sql instanceof PDOStatement)
		{
			$ps = $sql;
		}
		else
		{
			throw new InvalidArgumentException('Invalid argument type: $sql');
		}
		// Put single parameter argument into an array
		if (!is_array($parameters) && isset($parameters))
		{
			$parameters = array($parameters);
		}
		// Handle auto values
		$this->ResetLastAutoValue();
		$hasAutoValue = false;
		if (is_array($parameters))
		{
			foreach ($parameters as &$param)
			{
				if ($param instanceof UnbDbAutoValue)
				{
					if (!$hasAutoValue)
					{
						$hasAutoValue = true;
						$param = $this->GetSequenceValue($param->GetTable());
					}
					else
					{
						throw new InvalidArgumentException('More than one auto value passed as parameter.');
					}
				}
			}
		}
		// Execute the statement
		do
		{
			$r = $ps->execute($parameters);
			if (!$r) $this->DetermineErrorCondition($ps);
		}
		while (!$r && $this->busyError);   // SQLite busy loop workaround
		$time = microtime(true) - $start;
		$this->totalTime += $time;
		$this->queryCount++;
		if (self::DisplayQueryErrors)
			$this->ShowQueryInfo($ps, $r, $parameters, $time);
		if (!$r)
			throw new UnbDatabaseException('Statement could not be executed. ' . $this->errorMessage);
		// Get last auto value
		if ($hasAutoValue)
			$this->FetchLastAutoValue();
		return $ps;
	}

	// Executes an SQL statement that returns no data.
	//
	// sql = (string) SQL statement
	//       (PDOStatement) Already prepared statement
	// parameters = (array) Parameters to be used in this SQL statement, {{null}} if no parameters are used.
	//
	// Throws UnbDatabaseException if the statement could not be prepared or executed.
	//
	public function Execute($sql, $parameters = null)
	{
		$ps = $this->InternalExecute($sql, $parameters);
		if (isset($ps))
			$this->rowsAffected = $ps->rowCount();
		else
			$this->rowsAffected = -1;
	}

	// Executes an SQL statement and returns all records.
	//
	// sql = (string) SQL statement
	//       (PDOStatement) Already prepared statement
	// parameters = (array) Parameters to be used in this SQL statement, {{null}} if no parameters are used.
	//
	// Throws UnbDatabaseException if the statement could not be prepared or executed.
	//
	// Returns an array of all records
	//
	public function Query($sql, $parameters = null)
	{
		$ps = $this->InternalExecute($sql, $parameters);
		$records = $ps->fetchAll(PDO::FETCH_ASSOC);
		return $records;
	}

	// Executes an SQL statement and returns the first column of all records.
	//
	// sql = (string) SQL statement
	//       (PDOStatement) Already prepared statement
	// parameters = (array) Parameters to be used in this SQL statement, {{null}} if no parameters are used.
	//
	// Throws UnbDatabaseException if the statement could not be prepared or executed.
	//
	// Returns an array of all records' first column
	//
	public function QueryArray($sql, $parameters = null)
	{
		$ps = $this->InternalExecute($sql, $parameters);
		$records = $ps->fetchAll(PDO::FETCH_NUM);
		$array = array();
		foreach ($records as $record)
		{
			$array[] = $record[0];
		}
		return $array;
	}

	// Executes an SQL statement and returns the first record.
	//
	// sql = (string) SQL statement
	//       (PDOStatement) Already prepared statement
	// parameters = (array) Parameters to be used in this SQL statement, {{null}} if no parameters are used.
	//
	// Throws UnbDatabaseException if the statement could not be prepared or executed.
	//
	// Returns the first record
	//
	public function QueryRecord($sql, $parameters = null)
	{
		$ps = $this->InternalExecute($sql, $parameters);
		$record = $ps->fetch(PDO::FETCH_ASSOC);
		$ps->closeCursor();
		return $record;
	}

	// Executes an SQL statement and returns a single value.
	//
	// sql = (string) SQL statement
	//       (PDOStatement) Already prepared statement
	// parameters = (array) Parameters to be used in this SQL statement, {{null}} if no parameters are used.
	//
	// Throws UnbDatabaseException if the statement could not be prepared or executed.
	//
	// Returns the first record's first column, if successful, {{null}} if no record was fetched.
	//
	public function QueryValue($sql, $parameters = null)
	{
		$ps = $this->InternalExecute($sql, $parameters);
		$value = $ps->fetchColumn();
		$ps->closeCursor();
		if ($value === false) return null;
		return $value;
	}

	// Gets the number of rows affected by the last SQL statement (only for DELETE, INSERT and UPDATE).
	//
	// Returns (int) number of rows
	//
	public function GetRowsAffected()
	{
		return $this->rowsAffected;
	}

	// Converts a date/time value between PHP's UNIX timestamps and the SQL notation.
	//
	// in = (int) UNIX timestamp
	//      (string) SQL notation in the format "YYYY-MM-DD HH:MM:SS" and UT timezone.
	//               Only a date like "YYYY-MM-DD" is also accepted.
	//
	// Throws InvalidArgumentException on unrecognised date/time format or invalid argument type.
	//
	// Returns the date/time in the respective other format. An input value of (int) 0
	// or (string) "0000-00-00 00:00:00" is both interpreted as "no value" and converted
	// accordingly.
	//
	public static function ConvertDateTime($in)
	{
		if (!isset($in))
		{
			return null;
		}
		if (is_int($in))
		{
			if ($in == 0) return '0000-00-00 00:00:00';
			return gmdate('Y-m-d H:i:s', $in);
		}
		if (is_string($in))
		{
			if (preg_match('_^(\d{4})-(\d{2})-(\d{2})(?: (\d{2}):(\d{2}):(\d{2})(\.\d+)?)?$_', $in, $m))
			{
				if (intval($m[1]) == 0) return 0;   // No content (year is 0)
				return gmmktime(intval($m[4]), intval($m[5]), intval($m[6]), intval($m[2]), intval($m[3]), intval($m[1]));
			}
			throw new InvalidArgumentException('Unrecognised date/time format');
		}
		throw new InvalidArgumentException('Invalid argument type: $in');
	}

	// Gets all keys in a key list referenced by an entity.
	//
	// This function reads the currently assigned key list ID from tableName.columnName where
	// idColumn is idValue and returns all keys from this key list.
	//
	// tableName = (string) Table in which the key list is referenced
	// columnName = (string) Column which contains the key list ID reference
	// idColumn = (string) Identifier column (primary key) of the table
	// idValue = (int/string) Identifier value of the table, describing the record that references the key list
	// lock = (bool) Locks the returned keys for any access from other transactions.
	//
	// Throws InvalidArgumentException on invalid argument type or value.
	// Throws UnbDatabaseException
	//
	public function GetKeylistKeys($tableName, $columnName, $idColumn, $idValue, $lock = false)
	{
		if (!is_string($tableName))
			throw new InvalidArgumentException('Invalid argument type: $tableName');
		if (!preg_match('_^#?[0-9A-Z\_a-z]+$_', $tableName))
			throw new InvalidArgumentException('Invalid argument value: $tableName');
		if (!is_string($columnName))
			throw new InvalidArgumentException('Invalid argument type: $columnName');
		if (!preg_match('_^[0-9A-Z\_a-z]+$_', $columnName))
			throw new InvalidArgumentException('Invalid argument value: $columnName');
		if (!is_string($idColumn))
			throw new InvalidArgumentException('Invalid argument type: $idColumn');
		if (!preg_match('_^[0-9A-Z\_a-z]+$_', $idColumn))
			throw new InvalidArgumentException('Invalid argument value: $idColumn');

		$sql = 'SELECT kk."UserId"
			FROM "#keylist_keys" kk, "' . $tableName . '" t
			WHERE t."' . $columnName . '" = kk."KeylistId" AND t."' . $idColumn . '" = ?';
		if ($lock)
			$sql .= ' FOR UPDATE';

		$keys = $this->QueryArray($sql, $idValue);
		$keys = array_map('intval', $keys);
		return $keys;
	}

	// Adds a new key to a key list.
	//
	// This function reads the currently assigned key list ID from tableName.columnName where
	// idColumn is idValue and adds the new key to this key list. If the value is {{null}} (i.e. no key
	// list is assigned), a new one is created and assigned to tableName.columnName.
	//
	// tableName = (string) Table in which the key list is referenced
	// columnName = (string) Column which contains the key list ID reference
	// idColumn = (string) Identifier column (primary key) of the table
	// idValue = (int/string) Identifier value of the table, describing the record that references the key list
	// key = (int/UnbUser) Key to be added to the key list
	//
	// Throws InvalidArgumentException on invalid argument type or value.
	// Throws UnbDatabaseException
	//
	public function AddKeylistKey($tableName, $columnName, $idColumn, $idValue, $key)
	{
		if (!is_string($tableName))
			throw new InvalidArgumentException('Invalid argument type: $tableName');
		if (!preg_match('_^#?[0-9A-Z\_a-z]+$_', $tableName))
			throw new InvalidArgumentException('Invalid argument value: $tableName');
		if (!is_string($columnName))
			throw new InvalidArgumentException('Invalid argument type: $columnName');
		if (!preg_match('_^[0-9A-Z\_a-z]+$_', $columnName))
			throw new InvalidArgumentException('Invalid argument value: $columnName');
		if (!is_string($idColumn))
			throw new InvalidArgumentException('Invalid argument type: $idColumn');
		if (!preg_match('_^[0-9A-Z\_a-z]+$_', $idColumn))
			throw new InvalidArgumentException('Invalid argument value: $idColumn');
		if ($key instanceof UnbUser)
			$key = $key->GetUserId();
		if (!is_int($key))
			throw new InvalidArgumentException('Invalid argument type: $key');
		if (in_array($key, UnbUser::GetNonListKeys()))
			throw new InvalidArgumentException('This special key cannot be added to a key list: ' . $key);

		$this->StartTransaction(true);
		try
		{
			$keylistId = $this->QueryValue(
				'SELECT "' . $columnName . '" FROM "' . $tableName . '" WHERE "' . $idColumn . '" = ? FOR UPDATE',
				$idValue);
			$newlist = false;
			if (!$keylistId)
			{
				// Need to generate a new key list
				$newlist = true;
				$this->Execute(
					'INSERT INTO "#keylist" ("KeylistId") VALUES (?)',
					array(new UnbDbAutoValue('keylist')));
				$keylistId = $this->GetLastAutoValue();
			}
			$this->Execute(
				'INSERT INTO "#keylist_keys" ("KeylistId", "UserId") VALUES (?, ?)',
				array($keylistId, $key));
			if ($newlist)
			{
				// The key list reference can only be set when the key list exists
				$this->Execute(
					'UPDATE "' . $tableName . '" SET "' . $columnName . '" = ? WHERE "' . $idColumn . '" = ?',
					array($keylistId, $idValue));
			}
			$this->Commit();
		}
		catch (Exception $ex)
		{
			$this->RollBack();
			throw $ex;
		}
	}

	// Removes a key from a key list.
	//
	// This function reads the currently assigned key list ID from tableName.columnName where
	// idColumn is idValue and removes the key from this key list. If the key list is then empty,
	// the reference is deleted (set to {{null}}) in tableName.columnName.
	//
	// tableName = (string) Table in which the key list is referenced
	// columnName = (string) Column which contains the key list ID reference
	// idColumn = (string) Identifier column (primary key) of the table
	// idValue = (int/string) Identifier value of the table, describing the record that references the key list
	// key = (int/UnbUser) Key to be removed from the key list
	//
	// Throws InvalidArgumentException on invalid argument type or value.
	// Throws UnbDatabaseException
	//
	public function RemoveKeylistKey($tableName, $columnName, $idColumn, $idValue, $key)
	{
		if (!is_string($tableName))
			throw new InvalidArgumentException('Invalid argument type: $tableName');
		if (!preg_match('_^#?[0-9A-Z\_a-z]+$_', $tableName))
			throw new InvalidArgumentException('Invalid argument value: $tableName');
		if (!is_string($columnName))
			throw new InvalidArgumentException('Invalid argument type: $columnName');
		if (!preg_match('_^[0-9A-Z\_a-z]+$_', $columnName))
			throw new InvalidArgumentException('Invalid argument value: $columnName');
		if (!is_string($idColumn))
			throw new InvalidArgumentException('Invalid argument type: $idColumn');
		if (!preg_match('_^[0-9A-Z\_a-z]+$_', $idColumn))
			throw new InvalidArgumentException('Invalid argument value: $idColumn');
		if ($key instanceof UnbUser)
			$key = $key->GetUserId();
		if (!is_int($key))
			throw new InvalidArgumentException('Invalid argument type: $key');

		$this->StartTransaction(true);
		try
		{
			$keylistId = $this->QueryValue(
				'SELECT "' . $columnName . '" FROM "' . $tableName . '" WHERE "' . $idColumn . '" = ? FOR UPDATE',
				$idValue);
			if (!isset($keylistId))
			{
				$this->RollBack();
				return true;   // Nothing to do
			}

			$this->Execute(
				'DELETE FROM "#keylist_keys" WHERE "KeylistId" = ? AND "UserId" = ?',
				array($keylistId, $key));

			// All remaining references to now empty key lists can now be NULLed. A database
			// trigger already did that for us. :)

			$this->Commit();
		}
		catch (Exception $ex)
		{
			$this->RollBack();
			throw $ex;
		}
	}

	// Removes all keys from a key list and set new keys.
	//
	// This function reads the currently assigned key list ID from tableName.columnName where
	// idColumn is idValue and removes the key from this key list. If the key list is then empty,
	// the reference is deleted (set to {{null}}) in tableName.columnName.
	//
	// tableName = (string) Table in which the key list is referenced
	// columnName = (string) Column which contains the key list ID reference
	// idColumn = (string) Identifier column (primary key) of the table
	// idValue = (int/string) Identifier value of the table, describing the record that references the key list
	// keys = (array(int)) New keys. Set this to {{null}} to only remove all keys.
	//
	// Throws InvalidArgumentException on invalid argument type or value.
	// Throws UnbDatabaseException
	//
	public function ResetKeylistKeys($tableName, $columnName, $idColumn, $idValue, $keys)
	{
		if (!is_string($tableName))
			throw new InvalidArgumentException('Invalid argument type: $tableName');
		if (!preg_match('_^#?[0-9A-Z\_a-z]+$_', $tableName))
			throw new InvalidArgumentException('Invalid argument value: $tableName');
		if (!is_string($columnName))
			throw new InvalidArgumentException('Invalid argument type: $columnName');
		if (!preg_match('_^[0-9A-Z\_a-z]+$_', $columnName))
			throw new InvalidArgumentException('Invalid argument value: $columnName');
		if (!is_string($idColumn))
			throw new InvalidArgumentException('Invalid argument type: $idColumn');
		if (!preg_match('_^[0-9A-Z\_a-z]+$_', $idColumn))
			throw new InvalidArgumentException('Invalid argument value: $idColumn');
		if (isset($keys))
		{
			if (!is_array($keys))
				throw new InvalidArgumentException('Invalid argument type: $keys');
			if (!count($keys))
				$keys = null;
		}
		if (isset($keys))
		{
			$keys = array_unique($keys);
			$invalidKeys = UnbUser::GetNonListKeys();
			foreach ($keys as $key)
				if (in_array($key, $invalidKeys))
					throw new InvalidArgumentException('This special key cannot be added to a key list: ' . $key);
		}

		$this->StartTransaction(true);
		try
		{
			// Get all currently set keys and determine the difference sets of keys to add and to remove
			$presentKeys = $this->GetKeylistKeys($tableName, $columnName, $idColumn, $idValue, true);

			$addKeys = array();
			$removeKeys = array();
			if (isset($keys))
				foreach ($keys as $key)
					if (!in_array($key, $presentKeys))
						$addKeys[] = $key;
			foreach ($presentKeys as $key)
				if (!isset($keys) || !in_array($key, $keys))
					$removeKeys[] = $key;

			if (!count($addKeys) && !count($removeKeys))
			{
				$this->RollBack();
				return;   // Nothing to do
			}

			foreach ($addKeys as $key)
				$this->AddKeylistKey($tableName, $columnName, $idColumn, $idValue, $key);

			foreach ($removeKeys as $key)
				$this->RemoveKeylistKey($tableName, $columnName, $idColumn, $idValue, $key);

			$this->Commit();
		}
		catch (Exception $ex)
		{
			$this->RollBack();
			throw $ex;
		}
	}

	// Escape a value to use the wildcards * and ? instead of SQL's % and _.
	//
	// This value must be used in a LIKE ? ESCAPE ? context, with \ as the escape value.
	//
	// $s = (string) The value to escape
	//
	// Returns (string) the escaped value
	//
	public static function EscapeLikeWildcards($s)
	{
		return str_replace(
			array('%',   '_',   '\\',   '*', '?'),
			array('\\%', '\\_', '\\\\', '%', '_'),
			$s);
	}

	// Interprets a mixed value returned from the database as boolean.
	//
	// s = (any) Raw value
	//
	// Returns (bool) A boolean interpretation of that value.
	//
	public static function AsBool($s)
	{
		$s = strtolower(trim($s));
		return $s !== '' &&
			$s !== '0' &&
			$s !== 'false' &&
			$s !== 'no' &&
			$s !== 'off';
	}
} // class UnbDatabase

// Represents an auto-increment/sequenca value in a database query.
//
// Use an instance of this class as query parameter whenever the actual value stored in the
// database shall be automatically incremented. Only one query parameter of this type is allowed
// in a single query.
//
class UnbDbAutoValue
{
	// The database table the auto value is requested for.
	private $table;

	// Class constructor.
	//
	// Some DBMS that use native sequences require a sequence name to get the next value of. This
	// name is derived from the table name that is specified in this constructor.
	//
	// table = (string) Name of the database table the auto value is requested for.
	//
	public function __construct($table)
	{
		$this->table = $table;
	}

	// Gets the table name the auto value is requested for.
	//
	// Returns (string) Table name
	//
	public function GetTable()
	{
		return $this->table;
	}
} // UnbDbAutoValue

?>
