<?php
// UNB Components
// Copyright since 2007 by Yves Goergen
// Website: http://unclassified.de/go/unb2
// Licence: GNU General Public License (GPL) Version 3
//
// This file contains the UnbFormHandler class.

// Encoding: UTF-8 (without BOM) (auto-detect: °°°°°)

if (!class_exists('UnbDatabase', false)) require_once dirname(__FILE__) . '/database.class.php';
if (!class_exists('UnbMessage', false)) require_once dirname(__FILE__) . '/message.class.php';
if (!class_exists('UnbMessageRevision', false)) require_once dirname(__FILE__) . '/messagerevision.class.php';
if (!class_exists('UnbTag', false)) require_once dirname(__FILE__) . '/tag.class.php';
if (!class_exists('UnbSearch', false)) require_once dirname(__FILE__) . '/search.class.php';
if (!class_exists('UnbSession', false)) require_once dirname(__FILE__) . '/session.class.php';
if (!class_exists('UnbEnvironment', false)) require_once dirname(__FILE__) . '/environment.class.php';

// Handles HTML form requests and provides centralised request validating and XML conversion support.
//
class UnbFormHandler
{
	// Display some information about form handling.
	const VerboseHandling = false;
	// Let all forms fail the handling. Use this for debugging only!
	const FailForm = false;
	// Generate more verbose error messages that include a call stack and function arguments.
	// DO NOT enable this in a production environment as it may leak sensitive information!
	const VerboseErrors = true;

	// Newly created message instance, if the action was performed
	private $newMessage = null;
	// Requested message instance
	private $requestedMessage = null;
	// Newly created message revision instance, if the action was performed
	private $newMessageRevision = null;
	// Requested message revision instance
	private $requestedMessageRevision = null;
	// Newly created tag instance, if the action was performed
	private $newTag = null;
	// Requested tag instance
	private $requestedTag = null;
	// Newly created user instance, if the action was performed
	private $newUser = null;
	// Edited user instance, if the action was performed
	private $editedUser = null;
	// Requested user instance
	private $requestedUser = null;
	// List of requested actions
	private $requestedActions;
	// Actions to skip
	private $skipActions = array();

	// Gets the newly created message instance.
	//
	// Returns (UnbMessage) if the action was performed, {{null}} otherwise.
	//
	public function GetNewMessage()
	{
		return $this->newMessage;
	}

	// Gets the requested message instance that was worked with.
	//
	// Returns (UnbMessage) if one was requested, {{null}} otherwise.
	//
	public function GetRequestedMessage()
	{
		return $this->requestedMessage;
	}

	// Gets the newly created message revision instance.
	//
	// Returns (UnbMessageRevision) if the action was performed, {{null}} otherwise.
	//
	public function GetNewMessageRevision()
	{
		return $this->newMessageRevision;
	}

	// Gets the requested message revision instance that was worked with.
	//
	// Returns (UnbMessageRevision) if one was requested, {{null}} otherwise.
	//
	public function GetRequestedMessageRevision()
	{
		return $this->requestedMessageRevision;
	}

	// Gets the requested or newly created message and revision IDs.
	//
	// out messageId = (int) Message ID
	// out revisionId = (int) Message revision ID
	//
	public function GetMessageIds(&$messageId, &$revisionId)
	{
		$messageId = 0;
		$revisionId = 0;
		if ($this->GetRequestedMessage())
		{
			$messageId = $this->GetRequestedMessage()->GetMessageId();
		}
		if ($this->GetRequestedMessageRevision())
		{
			$messageId = $this->GetRequestedMessageRevision()->GetMessageId();
			$revisionId = $this->GetRequestedMessageRevision()->GetMessageRevisionId();
		}
		if ($this->GetNewMessage())
		{
			$messageId = $this->GetNewMessage()->GetMessageId();
		}
		if ($this->GetNewMessageRevision())
		{
			$messageId = $this->GetNewMessageRevision()->GetMessageId();
			$revisionId = $this->GetNewMessageRevision()->GetMessageRevisionId();
		}
	}

	// Gets the newly created tag instance.
	//
	// Returns (UnbTag) if the action was performed, {{null}} otherwise.
	//
	public function GetNewTag()
	{
		return $this->newTag;
	}

	// Gets the requested tag instance that was worked with.
	//
	// Returns (UnbTag) if one was requested, {{null}} otherwise.
	//
	public function GetRequestedTag()
	{
		return $this->requestedTag;
	}

	// Gets the newly created user instance.
	//
	// Returns (UnbUser) if the action was performed, {{null}} otherwise.
	//
	public function GetNewUser()
	{
		return $this->newUser;
	}

	// Gets the edited created user instance.
	//
	// Returns (UnbUser) if the action was performed, {{null}} otherwise.
	//
	public function GetEditedUser()
	{
		return $this->editedUser;
	}

	// Gets the requested user instance that was worked with.
	//
	// Returns (UnbUser) if one was requested, {{null}} otherwise.
	//
	public function GetRequestedUser()
	{
		return $this->requestedUser;
	}

	// Handles a submitted HTML form and performs all requested actions.
	//
	// Returns (array) Action return values
	//
	// The return array contains 3 entries at the first level: 'actions' contains the action return
	// values, 'erroraction' names the action where an error has occured (if any) and 'allsuccess'
	// is set if all actions were processed successfully.
	//
	public function HandleForm()
	{
		// First some formal checks...
		if (!self::ValidateRequest())
			return array('error' => 'Error validating request');

		$actions = $_REQUEST['action'];
		if (!is_array($actions))
			$actions = array($actions);
		$fields = array();
		$fields['actions'] = array();
		if (!count($fields)) return null;

		// These variables will be stored over multiple actions to cache some data/objects,
		// like a message that is first created and then further modified. No need to retrieve
		// it from the database for further actions, it's cached at the first time.
		$tmpVars = array();

		try
		{
			// Only acquire the database lock if an action is invoked that requires it
			$lockActions = array(
				'set_message_readaccess',
				'set_message_alteraccess',
				'set_message_replyaccess',
				'set_tag_readaccess',
				'set_tag_useaccess',
				'set_user_additionalkeys',
				'delete_user',
				);
			$lock = count(array_intersect($actions, $lockActions)) > 0;

			// Start a database transaction so that it can be rolled back. So if one action fails,
			// there is no partial execution of the entrie request.
			UnbDatabase::GetInstance()->StartTransaction($lock);

			// Bring actions in a defined order: First create things, then edit them
			usort($actions, array(__CLASS__, 'SortActions'));

			$this->requestedActions = $actions;

			foreach ($actions as $action)
			{
				// Each action can only be processed once, because it would read the same form data
				// the second time which is quite pointless. So we assume that $action is unique in
				// its array. If an action is really specified twice, it is simply ignored and skipped.
				if (isset($fields['actions'][$action])) continue;

				// Skip actions that have been marked so while processing earlier actions.
				if (in_array($action, $this->skipActions)) continue;

				// The possible exceptions are caught for every action individually so that the error
				// message can be recorded for the respective action.
				try
				{
					$this->HandleFormAction($action, $actionFields, $tmpVars);
				}
				catch (Exception $ex)
				{
					$actionFields['error'] = UnbLocale::Translate('exception:' . get_class($ex), true) . UnbLocale::Colon() . ' ' . $ex->getMessage();
					if (self::VerboseErrors)
						$actionFields['error'] .= "\n" . UnbDebug::GetCallTrace($ex);
				}

				$fields['actions'][$action] = $actionFields;

				// If an error occured while processing an action, all database changes are reverted
				// and no further actions are processed.
				if (isset($actionFields['error']))
				{
					if (self::VerboseHandling)
						echo '<p>An error was reported handling form action ' . $action . '.</p>';

					UnbDatabase::GetInstance()->RollBack();
					// This is kind of a pointer to the action that contains the error message.
					$fields['erroraction'] = $action;
					return $fields;
				}
			}

			if (self::FailForm)
			{
				if (self::VerboseHandling)
					echo '<p>Form handling is configured to fail.</p>';
				UnbDatabase::GetInstance()->RollBack();
				$fields['erroraction'] = '__failform';
				$fields['actions']['__failform']['error'] = 'Form handling is configured to fail.';
				return $fields;
			}

			UnbDatabase::GetInstance()->Commit();
			$fields['allsuccess'] = '';
		}
		catch (UnbDatabaseException $ex)
		{
			// Handle database errors with the transaction functions called in this function.
			// Be sure to invalidate all cached Unb* objects now and re-read them from the database
			// because the rollback does not affect PHP objects!
			UnbDatabase::GetInstance()->RollBack();
			$fields['error'] = UnbLocale::Translate('exception:' . get_class($ex), true) . UnbLocale::Colon() . ' ' . $ex->getMessage();
		}
		return $fields;
	}

	// Internal sorting function that ensures that actions are processed in the correct order.
	//
	private static function SortActions($a, $b)
	{
		// This array defines actions in the order they are to be processed
		$actionsOrderFirst = array(
			'login',
			'create_message',
			'alter_message',
			'create_tag',
			'create_user',
			'set_user_additionalkeys',   // Must come before set_user_{administrator,moderator,trusted}
			);
		$actionsOrderLast = array(
			'get_messages',
			// TODO: Add all the other retrieval actions here
			);
		// Every other action is processed in alphabetic order after/before the above actions are done.

		if (in_array($a, $actionsOrderFirst))
		{
			if (in_array($b, $actionsOrderFirst))
			{
				// Compare two known actions' order index
				$na = array_search($a, $actionsOrderFirst);
				$nb = array_search($b, $actionsOrderFirst);
				if ($na == $nb) return 0;
				return $na < $nb ? -1 : 1;
			}
			// $a has a known order, $b has not; so $a comes first
			return -1;
		}
		if (in_array($a, $actionsOrderLast))
		{
			if (in_array($b, $actionsOrderLast))
			{
				// Compare two known actions' order index
				$na = array_search($a, $actionsOrderLast);
				$nb = array_search($b, $actionsOrderLast);
				if ($na == $nb) return 0;
				return $na < $nb ? -1 : 1;
			}
			// $a has a known order, $b has not; so $b comes first
			return 1;
		}
		if (in_array($b, $actionsOrderFirst))
		{
			// $b has a known order, $a has not; so $b comes first
			return 1;
		}
		if (in_array($b, $actionsOrderLast))
		{
			// $b has a known order, $a has not; so $a comes first
			return -1;
		}
		// Neither one is known; now sort by name
		if ($a == $b) return 0;
		return $a < $b ? -1 : 1;
	}

	// Marks an action to be skipped in further processing.
	//
	// action = (string) Action code to be skipped
	//          (array(string)) Array of action codes to be skipped
	//
	private function SkipAction($action)
	{
		if (is_string($action))
			$this->skipActions[] = $action;
		if (is_array($action))
			foreach ($action as $a)
				if (is_string($a))
					$this->skipActions[] = $a;
	}

	// Gets a value indicating whether an action was requested and is not yet discarded.
	//
	// action = (string) Action code to be checked
	//
	// Returns (bool) {{true}} if the action is currently scheduled to be handled, {{false}} otherwise
	//
	private function IsActionRequested($action)
	{
		return in_array($action, $this->requestedActions) && !in_array($action, $this->skipActions);
	}

	// Handles a specific HTML form action.
	//
	// action = (string) Action name
	// out actionFields = (array) Array of return values from the action
	// inout tmpVars = (array) Array of temporary variables that are used in other actions, like a newly created object
	//
	private function HandleFormAction($action, &$actionFields, &$tmpVars)
	{
		if (self::VerboseHandling)
			echo '<p>Handling form action ' . $action . '...</p>';

		switch ($action)
		{
			// ---------- Session handling actions ----------

			case 'login':
				// Implicit registration is performed when posting a message while not logged in.
				// The user doesn't need to provide any information, only what is submitted in the
				// form is regarded.
				$implicitRegister = $_POST['implicit_register'] ? true : false;

				if (isset($_POST['user_loginname']))
					$loginName = trim($_POST['user_loginname']);
				if (isset($_POST['user_loginpassword']))
					$loginPassword = trim($_POST['user_loginpassword']);
				$this->SkipAction(array('set_user_loginname', 'set_user_loginpassword'));

				if (!strlen($loginName) && !$implicitRegister)
				{
					$actionFields['error'] = '[UNB#27] ' . UnbLocale::Translate('error:27');   // No user login name specified
				}
				else if (!strlen($loginPassword) && !UnbSession::IsAdministrator() && !$implicitRegister)
				{
					$actionFields['error'] = '[UNB#28] ' . UnbLocale::Translate('error:28');   // No user login password specified
				}
				else
				{
					if (strlen($loginName))
						$user = UnbUser::FindByLoginName($loginName);
					if (isset($user))
					{
						UnbSession::SetSessionUser($user, $loginPassword);
						$actionFields['success'] = '';
						$this->SkipAction('set_user_displayname');
					}
					else
					{
						if ($_REQUEST['allow_register'] || $implicitRegister)
						{
							$this->SkipAction('set_user_displayname');

							if (isset($_POST['user_displayname']))
								$displayName = trim($_POST['user_displayname']);
							if (isset($_POST['user_repeatpassword']))
								$repeatPassword = trim($_POST['user_repeatpassword']);

							if (isset($displayName) && !strlen($displayName) && !strlen($loginName))
							{
								// Only require a display name if is was included in the form and no login name was specified
								$actionFields['error'] = '[UNB#29] ' . UnbLocale::Translate('error:29');   // No user display name specified
							}
							else
							{
								if (!strlen($displayName) && strlen($loginName))
									$displayName = $loginName;

								if ($repeatPassword !== $loginPassword)
								{
									$actionFields['error'] = '[UNB#30] ' . UnbLocale::Translate('error:30');   // Incorrect password repetition
								}
								else if (UnbConfiguration::Get('captcha.enabled', true) &&
									class_exists('UnbCaptcha', false) &&
									!UnbCaptcha::CheckWord($_POST['captcha_word']))
								{
									$actionFields['error'] = '[UNB#31] ' . UnbLocale::Translate('error:31');   // Incorrect CAPTCHA word
								}
								else
								{
									$user = UnbUser::Create($displayName, $loginName, $loginPassword);
									$tmpVars['user'] = $user;
									$this->newUser = $tmpVars['user'];
									UnbSession::SetSessionUser($tmpVars['user'], null, true);
									$actionFields['success'] = '';
									if (class_exists('UnbCaptcha', false))
										UnbCaptcha::ClearWord();
								}
							}
						}
						else
						{
							$actionFields['error'] = '[UNB#32] ' . UnbLocale::Translate('error:32');   // Login name is unknown
						}
					}

					if (isset($actionFields['success']))
					{
						// Send auto-login cookie
						$p = $user->GetSingleSetting('autologin.password');
						if (!isset($p))
						{
							// Generate a new password
							$p = UnbSession::GeneratePassword();
							$user->SetSetting('autologin.password', UnbSession::Encrypt($p));
						}
						else
						{
							$p = UnbSession::Decrypt($p);
						}
						// Add the user ID
						$p = $user->GetUserId() . ' ' . $p;
						UnbSession::SetCookie('unb_autologin', $p);
					}
				}
				break;

			case 'logout':
				UnbSession::SetSessionUser(null);
				$actionFields['success'] = '';
				UnbSession::SetCookie('unb_autologin');
				break;

			// ---------- Message and revision handling actions ----------

			case 'create_message':
				$entryPoint = false;
				// Handle set_message_entrypoint action here as well
				if ($this->IsActionRequested('set_message_entrypoint'))
				{
					$entryPoint = $_POST['message_entrypoint'] ? true : false;
					$this->SkipAction('set_message_entrypoint');
				}
				$subject = null;
				if (isset($_POST['message_subject']))
					$subject = trim($_POST['message_subject']);
				$content = null;
				if (isset($_POST['message_content']))
					$content = rtrim($_POST['message_content']);
				$draft = isset($_POST['draft']) || isset($_POST['draft_addfile']);
				$data = null;
				$dataFileName = null;
				$writeToFile = null;
				if ($_POST['set_messagerevision_filedata'])
				{
					// Attachment action specified
					if (strlen(trim($_FILES['message_filedata']['tmp_name'])))
					{
						$dataFileName = $_FILES['message_filedata']['tmp_name'];
						$contentType = UnbMessage::GetContentTypeForFileName($_FILES['message_filedata']['name']);

						if (!isset($subject))
						{
							$subject = basename($_FILES['message_filedata']['name']);
						}
					}
				}
				$primaryRef = null;
				if (intval($_POST['replyto_messageid']) > 0)
					$primaryRef = intval($_POST['replyto_messageid']);
				$addToRev = null;
				if (intval($_POST['addfileto_revisionid']) > 0)
					$addToRev = intval($_POST['addfileto_revisionid']);
				$relation = null;
				if (isset($_POST['addfile_relation']))
					$relation = trim($_POST['addfile_relation']);

				if (!strlen($content) && !$draft && !isset($dataFileName))
				{
					$actionFields['error'] = '[UNB#33] ' . UnbLocale::Translate('error:33');   // No message content specified
				}
				else
				{
					$tmpVars['message'] = UnbMessage::Create($subject, $content, $contentType, $data, $dataFileName, $writeToFile, $entryPoint, $primaryRef, $draft);
					$tmpVars['messagerevision'] = $tmpVars['message']->GetSearchRevision(UnbMessageRevisionContentLevel::Content, true);

					if (isset($addToRev))
					{
						$rev = UnbMessageRevision::FindByRevisionId($addToRev);
						$rev->AddFile($tmpVars['message']->GetMessageId(), $relation);
					}

					$actionFields['success'] = '';
					$this->newMessage = $tmpVars['message'];
					$this->newMessageRevision = $tmpVars['messagerevision'];
				}
				break;

			case 'set_messagerevision_tags':
				$rev = $this->GetMessageRevision($tmpVars);

				$tagIds = $_POST['message_tagid'];
				if (!is_array($tagIds))
				{
					if (strlen($tagIds))
						$tagIds = array($tagIds);
					else
						$tagIds = array();
				}
				$tagIds = array_map('intval', $tagIds);

				// Remove all tags that are assigned in the revision but not selected by the editor
				$revTagIds = $rev->GetTagIds(false);
				if ($revTagIds)
				{
					foreach (array_diff($revTagIds, $tagIds) as $tagId)
					{
						$rev->RemoveTag($tagId);
					}
				}

				// Add newly selected tags
				foreach ($tagIds as $tagId)
				{
					$tag = UnbTag::FindByTagId($tagId);
					if (!isset($tag)) continue;   // Ignore unknown tag IDs

					if ($tag->GetFlags() & UnbTagFlag::NotSelectable) continue;
					#if ($tag->GetFlags() & UnbTagFlag::Invisible) continue;

					$tagData = null;
					if ($tag->GetFlags() & UnbTagFlag::UserData &&
						strlen(trim($_POST['message_tagdata_' . $tagId])))
					{
						$tagData = trim($_POST['message_tagdata_' . $tagId]);
					}

					$rev->AddTag($tagId, $tagData);
				}
				$actionFields['success'] = '';
				break;

			case 'set_message_pagename':
				$msg = $this->GetMessage($tmpVars);
				$msg->SetPageName($_POST['message_pagename']);
				$actionFields['success'] = '';
				break;

			case 'set_message_entrypoint':
				$msg = $this->GetMessage($tmpVars);
				$msg->SetEntryPoint($_REQUEST['message_entrypoint'] ? true : false);
				$actionFields['success'] = '';
				break;

			case 'set_message_readaccess':
				if (!$_POST['message_readaccess_empty'])
				{
					$keys = $this->GetKeys('message_readaccesslist');
					if (is_string($keys))
					{
						$actionFields['error'] = $keys;
						break;
					}
				}
				else
				{
					$keys = array();
				}

				$msg = $this->GetMessage($tmpVars);
				$msg->SetReadAccessKeys($keys);
				$actionFields['success'] = '';
				break;

			case 'set_message_alteraccess':
				if (!$_POST['message_alteraccess_empty'])
				{
					$keys = $this->GetKeys('message_alteraccesslist');
					if (is_string($keys))
					{
						$actionFields['error'] = $keys;
						break;
					}

					// Always keep the moderator key on this list
					$keys[] = UnbUserKey::Moderator;
				}
				else
				{
					$keys = array();
				}

				$msg = $this->GetMessage($tmpVars);
				$msg->SetAlterAccessKeys($keys);
				$actionFields['success'] = '';
				break;

			case 'set_message_replyaccess':
				if (!$_POST['message_replyaccess_empty'])
				{
					$keys = $this->GetKeys('message_replyaccesslist');
					if (is_string($keys))
					{
						$actionFields['error'] = $keys;
						break;
					}
				}
				else
				{
					$keys = array();
				}

				$msg = $this->GetMessage($tmpVars);
				$msg->SetReplyAccessKeys($keys);
				$actionFields['success'] = '';
				break;

			case 'alter_message':
				// Copy all data by default
				$subject = true;
				$content = true;
				$summary = null;
				$data = true;
				$dataFileName = null;
				$writeToFile = null;

				// Look for new data
				if ($_POST['message_subject_hash'] &&
					md5($_POST['message_subject']) != $_POST['message_subject_hash'])
				{
					$subject = trim($_POST['message_subject']);
				}
				if ($_POST['message_content_hash'] &&
					md5($_POST['message_content']) != $_POST['message_content_hash'])
				{
					$content = rtrim($_POST['message_content']);
				}
				$summary = trim($_POST['message_summary']);
				$draft = isset($_POST['draft']) || isset($_POST['draft_addfile']);
				if ($_POST['set_messagerevision_filedata'])
				{
					// New attachment action specified
					#$clearData = $_POST['message_cleardata'] ? true : false;
					// NOTE: File data cannot be added or removed later, it can only be replaced. A file remains a file, a non-file won't become a file.
					#if ($clearData)
					#{
					#	$data = null;
					#}
					/*else*/ if (strlen(trim($_FILES['message_filedata']['tmp_name'])))
					{
						$data = null;
						$dataFileName = $_FILES['message_filedata']['tmp_name'];
						$contentType = UnbMessage::GetContentTypeForFileName($_FILES['message_filedata']['name']);
					}
				}
				$removeFileIds = $_POST['message_removefile'];
				if (!is_array($removeFileIds))
				{
					if (strlen($removeFileIds))
						$removeFileIds = array($removeFileIds);
					else
						$removeFileIds = array();
				}
				$removeFileIds = array_map('intval', $removeFileIds);

				if (is_string($content) && !strlen($content) && !$draft)
				{
					$actionFields['error'] = '[UNB#33] ' . UnbLocale::Translate('error:33');   // No message content specified
				}
				else
				{
					// Always request the message revision so that it can be queried later
					$baseRev = $this->GetMessageRevision($tmpVars);

					// Only create a new revision if anything has changed
					if ($subject !== true ||   // Subject changed
						$content !== true ||   // Content changed
						strlen($summary) > 0 ||   // Summary set
						$data !== true ||   // Data set or unset
						$baseRev->IsDraft() && !$draft ||   // Draft finally submitted
						!$baseRev->IsDraft() && $draft && isset($_POST['draft_addfile']) ||   // File shall be added to non-draft, create draft
						!$baseRev->IsDraft() && $draft && $removeFileIds)   // Files removed from a non-draft, create draft
					{
						// When a draft is saved and the DEP shall be set, this must be done as long
						// as the message is still a draft. Later this is only allowed to moderators.
						if ($this->IsActionRequested('set_message_entrypoint'))
						{
							$msg = $this->GetMessage($tmpVars);
							$msg->SetEntryPoint($_REQUEST['message_entrypoint'] ? true : false);
							$this->SkipAction('set_message_entrypoint');
						}

						$tmpVars['messagerevision'] = $baseRev->CreateCopy($subject, $content, $summary, $contentType, $data, $dataFileName, $writeToFile, $draft, $removeFileIds);
						$actionFields['success'] = '';
						$this->newMessageRevision = $tmpVars['messagerevision'];
					}
				}
				break;

			case 'delete_message':
				// TODO: There should be some kind of confirmation question
				$msg = $this->GetMessage($tmpVars);
				$msg->Delete();
				$actionFields['success'] = '';
				break;

			case 'moderate_message':
				$flags = 0;
				if ($_POST['message_locked']) $flags |= UnbMessageModerationFlag::Locked;
				if ($_POST['message_hidden']) $flags |= UnbMessageModerationFlag::Hidden;
				if ($_POST['message_enforce_approval']) $flags |= UnbMessageModerationFlag::EnforceApproval;

				$msg = $this->GetMessage($tmpVars);
				$msg->SetModerationFlags($flags);
				$actionFields['success'] = '';
				break;

			case 'rate_message':
				$msg = $this->GetMessage($tmpVars);
				$msg->SetRating(intval($_REQUEST['message_rating']));
				$actionFields['success'] = '';
				break;

			case 'clear_rate_message':
				$msg = $this->GetMessage($tmpVars);
				$msg->ClearRating();
				$actionFields['success'] = '';
				break;

			case 'moderate_messagerevision':
				$rev = $this->GetMessageRevision($tmpVars);
				$rev->SetModerationState(intval($_REQUEST['message_modstate']));
				$actionFields['success'] = '';
				break;

			// ---------- Tag handling actions ----------

			case 'create_tag':
				$tmpVars['tag'] = UnbTag::Create($_POST['tag_name']);
				$actionFields['success'] = '';
				$this->newTag = $tmpVars['tag'];
				$this->SkipAction('set_tag_name');
				break;

			case 'set_tag_name':
				$tag = $this->GetTag($tmpVars);
				$tag->SetTagName($_POST['tag_name']);
				$actionFields['success'] = '';
				break;

			case 'set_tag_description':
				$tag = $this->GetTag($tmpVars);
				$tag->SetDescription($_POST['tag_description']);
				$actionFields['success'] = '';
				break;

			case 'set_tag_flags':
				$flags = 0;
				if ($_POST['tag_flag_notselectable'])
					$flags |= UnbTagFlag::NotSelectable;
				if ($_POST['tag_flag_invisible'])
					$flags |= UnbTagFlag::Invisible;
				if ($_POST['tag_flag_enforceapproval'])
					$flags |= UnbTagFlag::EnforceApproval;
				if ($_POST['tag_flag_userdata'])
					$flags |= UnbTagFlag::UserData;

				$tag = $this->GetTag($tmpVars);
				$tag->SetFlags($flags);
				$actionFields['success'] = '';
				break;

			case 'set_tag_parent':
				$tag = $this->GetTag($tmpVars);
				$tag->SetParentTag(intval($_POST['tag_parentid']));
				$actionFields['success'] = '';
				break;

			case 'set_tag_readaccess':
				if (!$_POST['tag_readaccess_empty'])
				{
					$keys = $this->GetKeys('tag_readaccesslist');
					if (is_string($keys))
					{
						$actionFields['error'] = $keys;
						break;
					}
				}
				else
				{
					$keys = array();
				}

				$tag = $this->GetTag($tmpVars);
				$tag->SetReadAccessKeys($keys);
				$actionFields['success'] = '';
				break;

			case 'set_tag_useaccess':
				if (!$_POST['tag_useaccess_empty'])
				{
					$keys = $this->GetKeys('tag_useaccesslist');
					if (is_string($keys))
					{
						$actionFields['error'] = $keys;
						break;
					}
				}
				else
				{
					$keys = array();
				}

				$tag = $this->GetTag($tmpVars);
				$tag->SetUseAccessKeys($keys);
				$actionFields['success'] = '';
				break;

			case 'delete_tag':
				// TODO: There should be some kind of confirmation question
				$tag = $this->GetTag($tmpVars);
				$tag->Delete();
				$actionFields['success'] = '';
				break;

			// ---------- User handling actions ----------

			case 'create_user':
				$displayName = trim($_POST['user_displayname']);
				$loginName = trim($_POST['user_loginname']);
				$loginPassword = trim($_POST['user_loginpassword']);
				$this->SkipAction(array('set_user_loginname', 'set_user_loginpassword', 'set_user_displayname'));

				if (!strlen($displayName))
					$displayName = $loginName;
				$repeatPassword = trim($_POST['user_repeatpassword']);
				if ($repeatPassword !== $loginPassword)
				{
					$actionFields['error'] = '[UNB#30] ' . UnbLocale::Translate('error:30');   // Incorrect password repetition
				}
				else
				{
					$tmpVars['user'] = UnbUser::Create($displayName, $loginName, $loginPassword);
					$actionFields['success'] = '';
					$this->newUser = $tmpVars['user'];
				}
				break;

			case 'set_user_additionalkeys':
				if (!$_POST['user_additionalkeys_empty'])
				{
					$keys = $this->GetKeys('user_additionalkeyslist');
					if (is_string($keys))
					{
						$actionFields['error'] = $keys;
						break;
					}
				}
				else
				{
					$keys = array();
				}

				// Handle set_user_{administrator,moderator,trusted} actions here as well
				if ($this->IsActionRequested('set_user_administrator'))
				{
					if ($_POST['user_is_administrator'])
						$keys[] = UnbUserKey::Administrator;
					$this->SkipAction('set_user_administrator');
				}
				if ($this->IsActionRequested('set_user_moderator'))
				{
					if ($_POST['user_is_moderator'])
						$keys[] = UnbUserKey::Moderator;
					$this->SkipAction('set_user_moderator');
				}
				if ($this->IsActionRequested('set_user_trusted'))
				{
					if ($_POST['user_is_trusted'])
						$keys[] = UnbUserKey::Trusted;
					$this->SkipAction('set_user_trusted');
				}

				$this->editedUser = $this->GetUser($tmpVars);
				$this->editedUser->SetAdditionalKeys($keys);
				$actionFields['success'] = '';
				break;

			case 'set_user_administrator':
				$this->editedUser = $this->GetUser($tmpVars);
				if ($_REQUEST['user_is_administrator'])
					$this->editedUser->AddAdditionalKey(UnbUserKey::Administrator);
				else
					$this->editedUser->RemoveAdditionalKey(UnbUserKey::Administrator);
				$actionFields['success'] = '';
				break;

			case 'set_user_moderator':
				$this->editedUser = $this->GetUser($tmpVars);
				if ($_REQUEST['user_is_moderator'])
					$this->editedUser->AddAdditionalKey(UnbUserKey::Moderator);
				else
					$this->editedUser->RemoveAdditionalKey(UnbUserKey::Moderator);
				$actionFields['success'] = '';
				break;

			case 'set_user_trusted':
				$this->editedUser = $this->GetUser($tmpVars);
				if ($_REQUEST['user_is_trusted'])
					$this->editedUser->AddAdditionalKey(UnbUserKey::Trusted);
				else
					$this->editedUser->RemoveAdditionalKey(UnbUserKey::Trusted);
				$actionFields['success'] = '';
				break;

			case 'set_user_displayname':
				$this->editedUser = $this->GetUser($tmpVars);
				$displayName = trim($_POST['user_displayname']);
				if ($displayName != $this->editedUser->GetDisplayName())
				{
					$this->editedUser->SetDisplayName($displayName);
					$actionFields['success'] = '';
				}
				break;

			case 'set_user_loginname':
				$this->editedUser = $this->GetUser($tmpVars);
				$loginName = trim($_POST['user_loginname']);
				if ($loginName != $this->editedUser->GetLoginName())
				{
					if (!UnbSession::IsAdministrator() &&
						!UnbSession::ValidateSessionUserPassword($_POST['user_currentpassword']))
					{
						$actionFields['error'] = '[UNB#34] Incorrect current password';
					}
					else
					{
						$this->editedUser->SetLoginName($loginName);
						$actionFields['success'] = '';
					}
				}
				break;

			case 'set_user_loginpassword':
				// Only change the password if it's not spaces only (likely not to be changed in the web form)
				if (!strlen($_POST['user_loginpassword']) || strlen(trim($_POST['user_loginpassword'])))
				{
					if (trim($_POST['user_loginpassword']) !== trim($_POST['user_repeatpassword']))
					{
						$actionFields['error'] = '[UNB#30] ' . UnbLocale::Translate('error:30');   // Incorrect password repetition
					}
					else if (!UnbSession::IsAdministrator() &&
						!UnbSession::ValidateSessionUserPassword($_POST['user_currentpassword']))
					{
						$actionFields['error'] = '[UNB#34] ' . UnbLocale::Translate('error:34');   // Incorrect current password
					}
					else
					{
						$this->editedUser = $this->GetUser($tmpVars);
						$this->editedUser->SetLoginPassword(trim($_POST['user_loginpassword']));
						$actionFields['success'] = '';
					}
				}
				break;

			case 'delete_user':
				// TODO: There should be some kind of confirmation question
				$this->editedUser = $this->GetUser($tmpVars);
				$this->editedUser->Delete();
				$actionFields['success'] = '';
				break;

			case 'set_user_vcard':
				$this->editedUser = $this->GetUser($tmpVars);
				$vcardTypes = array();
				foreach ($_POST as $name => $value)
				{
					if (preg_match('/^vcard_data_(.+)$/', $name, $m))
						$vcardTypes[] = $m[1];
				}
				foreach ($vcardTypes as $type)
				{
					if ($_POST['vcard_remove_' . $type])
					{
						$this->editedUser->RemoveVcardEntry($type);
					}
					else
					{
						$visibility = null;
						if (isset($_POST['vcard_visibility_' . $type]))
							$visibility = intval($_POST['vcard_visibility_' . $type]);
						$this->editedUser->SetVcardEntry($type, $_POST['vcard_data_' . $type], $visibility);
					}
				}
				if (strlen(trim($_POST['vcard_newtype'])))
				{
					$visibility = UnbUserVcardVisibility::Public_;
					if (isset($_POST['vcard_newvisibility']))
						$visibility = intval($_POST['vcard_newvisibility']);
					$this->editedUser->SetVcardEntry($_POST['vcard_newtype'], $_POST['vcard_newdata'], $visibility);
				}
				$actionFields['success'] = '';
				break;

			case 'reset_user_autologin_password':
				$this->editedUser = $this->GetUser($tmpVars);
				if ($_REQUEST['reset_user_autologin_password_value'] == '1')
				{
					// Generate a new password
					$p = UnbSession::GeneratePassword();
					$this->editedUser->SetSetting('autologin.password', UnbSession::Encrypt($p));
					// Add the user ID
					$p = $this->editedUser->GetUserId() . ' ' . $p;
					UnbSession::SetCookie('unb_autologin', $p);
				}
				break;

			// ---------- Data retrieval actions ----------

			case 'get_messages':
				$searchParameters = array();

				if (is_numeric($_POST['search_replies_of']))
				{
					$replyIds = UnbSearch::GetAllReplies(intval($_POST['search_replies_of']));
					$searchParameters['message-ids'] = array_keys($replyIds);
					if (!count($replyIds))
					{
						$actionFields['message'] = array();
						break;
					}
				}

				if (is_array($_POST['search_message_ids']))
					$searchParameters['message-ids'] = array_map('intval', $_POST['search_message_ids']);
				if (is_string($_POST['search_message_ids']) && strlen($_POST['search_message_ids']))
					$searchParameters['message-ids'] = array_map('intval', explode(',', $_POST['search_message_ids']));
				// Select newly created message
				if ($_POST['get_new_message'])
				{
					if ($this->newMessage)
					{
						$searchParameters['message-ids'] = array($this->newMessage->GetMessageId());
					}
					else
					{
						$actionFields['error'] = 'No new message to get';
						break;
					}
				}

				if (is_array($_POST['search_owner_ids']))
					$searchParameters['owner-ids'] = array_map('intval', $_POST['search_owner_ids']);
				if (is_string($_POST['search_owner_ids']) && strlen($_POST['search_owner_ids']))
					$searchParameters['owner-ids'] = array_map('intval', explode(',', $_POST['search_owner_ids']));

				if ($_POST['search_entrypoint'])
					$searchParameters['entrypoint'] = true;

				if (is_array($_POST['search_tag_ids']))
					$searchParameters['tags-set'] = array_map('intval', $_POST['search_tag_ids']);
				if (is_string($_POST['search_tag_ids']) && strlen($_POST['search_tag_ids']))
					$searchParameters['tags-set'] = array_map('intval', explode(',', $_POST['search_tag_ids']));

				if (is_numeric($_POST['search_primary_reference']))
					$searchParameters['primary-reference'] = intval($_POST['search_primary_reference']);

				if (is_numeric($_POST['search_sort_tag_id']))
					$searchParameters['sort-tagid'] = intval($_POST['search_sort_tag_id']);

				if ($_POST['search_sort_createdtime'])
					$searchParameters['sort-createdtime'] = true;

				if ($_POST['search_sort_reverse'])
					$searchParameters['sort-reverse'] = true;

				if (is_string($_POST['search_limit']) && strlen($_POST['search_limit']))
					$searchParameters['limit'] = trim($_POST['search_limit']);

				$messageIds = UnbSearch::Query($searchParameters);

				$contentLevel = UnbMessageRevisionContentLevel::Subject;
				if ($_POST['message_contentlevel'] == 1)
				{
					$contentLevel = UnbMessageRevisionContentLevel::Content;
				}
				if ($_POST['message_contentlevel'] == 2)
				{
					$contentLevel = UnbMessageRevisionContentLevel::HtmlContent;
					$markup = new UnbMarkup();
				}

				$actionFields['message'] = array();
				if ($messageIds) foreach ($messageIds as $messageId)
				{
					// Retrieve message
					$msg = UnbMessage::FindByMessageId($messageId);
					if (!isset($msg)) continue;   // Actually, this should not happen here anymore
					$searchRev = $msg->GetSearchRevision($contentLevel, true);
					$latestRev = $msg->GetLatestRevision($contentLevel, true);
					if (!isset($searchRev)) continue;   // Actually, this should not happen here anymore
					if (!isset($latestRev)) continue;   // Actually, this should not happen here anymore

					$msgData = array();
					$msgData['message-id'] = $messageId;
					$msgData['revision-id'] = $searchRev->GetMessageRevisionId();
					$msgData['primary-reference'] = $msg->GetPrimaryReference();
					$msgData['entry-point'] = $msg->IsEntryPoint();
					if ($searchRev->GetContentLevel() >= UnbMessageRevisionContentLevel::Subject)
					{
						$msgData['subject'] = $searchRev->GetSubject();
					}
					if ($searchRev->GetContentLevel() >= UnbMessageRevisionContentLevel::HtmlContent)
					{
						$html = $searchRev->GetHtmlContent();
						#$html = null;
						if (!isset($html))
						{
							if ($searchRev->GetContent() !== null)
							{
								$html = $markup->Convert($searchRev->GetContent());
								try
								{
									$searchRev->SetHtmlContent($html);
								}
								catch (Exception $ex) { }   // Ignore errors here
							}
							else
							{
								$html = '';   // Empty contents should always be empty HTML
							}
						}
						$html = $markup->PostProcess($html);
						$msgData['htmlcontent'] = $html;
					}
					else if ($searchRev->GetContentLevel() >= UnbMessageRevisionContentLevel::Content)
					{
						$msgData['content'] = $searchRev->GetContent();
					}
					$msgData['owner'] = array(
						'id' => $msg->GetOwnerUser()->GetUserId(),
						'name' => $msg->GetOwnerUser()->GetDisplayName());
					$msgData['created-time'] = array(
						'unix' => $msg->GetCreatedTime(),
						'db' => UnbDatabase::ConvertDateTime($msg->GetCreatedTime()));
					if ($searchRev->GetCreatedTime() > $msgCreatedTime)
						$msgData['altered-time'] = array(
							'unix' => $searchRev->GetCreatedTime(),
							'db' => UnbDatabase::ConvertDateTime($searchRev->GetCreatedTime()));
					$latestReply = $msg->GetLatestReply();
					if (isset($latestReply))
					{
						$msgData['replies-count'] = $msg->CountReplies();
						$msgData['last-reply-time'] = array(
							'unix' => $latestReply->GetCreatedTime(),
							'db' => UnbDatabase::ConvertDateTime($latestReply->GetCreatedTime()));
					}
					$msgData['tag'] = array();
					$tags = $searchRev->GetTags(true);
					if ($tags) foreach ($tags as $tag)
					{
						$tagData = array();
						$tagData['id'] = $tag->GetTagId();
						$tagData['name'] = $tag->GetTagName();
						if (($tag->GetFlags() & UnbTagFlag::UserData) && ($tag->GetTagData() != null))
							$tagData['data'] = $tag->GetTagData();
						$msgData['tag'][] = $tagData;
					}
					if ($msg->HasFileData())
					{
						$msgData['attach-size'] = $searchRev->GetFileDataLength();
					}
					$flags = $msg->GetModerationFlags();
					$msgData['moderation'] = array(
						'hidden' => ($flags & UnbMessageModerationFlag::Hidden) ? true : false,
						'locked' => ($flags & UnbMessageModerationFlag::Locked) ? true : false,
						'latest-rev-waiting-for-approval' => ($latestRev->GetModerationState() == UnbMessageRevisionModerationState::WaitingForApproval),
						'latest-rev-locked' => ($latestRev->GetModerationState() == UnbMessageRevisionModerationState::Locked));
					// TODO: Add read access keys

					$msgData['is-alter-allowed'] = UnbMessage::IsAlterAllowedById($messageId);
					$msgData['is-reply-allowed'] = UnbMessage::IsReplyAllowedById($messageId);

					$actionFields['message'][] = $msgData;
				}
				break;

			case 'list_revisions':
				// TODO
				break;

			case 'list_tags':
				// TODO
				break;

			case 'list_users':
				// TODO
				break;

			// ---------- General database actions ----------

			case 'optimise_database':
				if (UnbSession::IsAdministrator())
				{
					UnbDatabase::GetInstance()->Optimise($_REQUEST['optimise_db_extensive'] ? true : false);
					$actionFields['success'] = '';
				}
				else
				{
					$actionFields['error'] = 'Only allowed for administrators';
				}
				break;

			default:
				$actionFields['error'] = 'Unknown action: ' . $action;
				break;
		}

		if (self::VerboseHandling)
			echo '<p>Completed handling form action ' . $action . '.</p>';
	}

	// Gets the message object from either the message ID in the HTML form or the temporary variables
	// from earlier actions.
	//
	// inout tmpVars = (array) Array of temporary variables that are used in other actions, like a newly created object
	//
	// Returns (UnbMessage)
	//
	private function GetMessage(&$tmpVars)
	{
		if ($tmpVars['message'] instanceof UnbMessage)
		{
			$msg = $tmpVars['message'];
		}
		else
		{
			$msgId = intval($_REQUEST['messageid']);
			if ($msgId > 0)
			{
				$msg = UnbMessage::FindByMessageId($msgId);
			}
			else
			{
				$rev = $this->GetMessageRevision($tmpVars);
				if ($rev)
				{
					$msg = UnbMessage::FindByMessageId($rev->GetMessageId());
				}
			}
			$tmpVars['message'] = $msg;
			$this->requestedMessage = $msg;
		}
		return $msg;
	}

	// Gets the message revision object from either the message ID and revision number in the HTML
	// form or the temporary variables from earlier actions.
	//
	// inout tmpVars = (array) Array of temporary variables that are used in other actions, like a newly created object
	//
	// Returns (UnbMessageRevision)
	//
	private function GetMessageRevision(&$tmpVars)
	{
		if ($tmpVars['messagerevision'] instanceof UnbMessageRevision)
		{
			$rev = $tmpVars['messagerevision'];
		}
		else
		{
			$revId = intval($_REQUEST['messagerevisionid']);
			if ($revId > 0)
			{
				$rev = UnbMessageRevision::FindByRevisionId($revId, UnbMessageRevisionContentLevel::Content, true);
			}
			else
			{
				$msg = $this->GetMessage($tmpVars);
				if ($msg)
				{
					$rev = $msg->GetLatestRevision(UnbMessageRevisionContentLevel::Content, true);
				}
			}

			$tmpVars['messagerevision'] = $rev;
			$this->requestedMessageRevision = $rev;
		}
		return $rev;
	}

	// Gets the user object from either the user ID in the HTML form or the temporary variables from
	// earlier actions.
	//
	// inout tmpVars = (array) Array of temporary variables that are used in other actions, like a newly created object
	//
	// Returns (UnbUser)
	//
	private function GetUser(&$tmpVars)
	{
		if ($tmpVars['user'] instanceof UnbUser)
		{
			$user = $tmpVars['user'];
		}
		else
		{
			$user = UnbUser::FindByUserId(intval($_REQUEST['userid']));
			$tmpVars['user'] = $user;
			$this->requestedUser = $user;
		}
		return $user;
	}

	// Gets the tag object from either the tag ID in the HTML form or the temporary variables from
	// earlier actions.
	//
	// inout tmpVars = (array) Array of temporary variables that are used in other actions, like a newly created object
	//
	// Returns (UnbTag)
	//
	private function GetTag(&$tmpVars)
	{
		if ($tmpVars['tag'] instanceof UnbTag)
		{
			$tag = $tmpVars['tag'];
		}
		else
		{
			$tag = UnbTag::FindByTagId(intval($_REQUEST['tagid']));
			$tmpVars['tag'] = $tag;
			$this->requestedTag = $tag;
		}
		return $tag;
	}

	// Gets the list of user IDs from a multiline text field in the HTML form.
	//
	// fieldName = (string) HTML form field name
	//
	// Returns (array) User IDs, or (string) error message if a user name is unknown
	//
	private function GetKeys($fieldName)
	{
		$keys = array();
		foreach (explode("\n", $_POST[$fieldName]) as $line)
		{
			$line = trim($line);
			if (!strlen($line)) continue;   // Skip empty lines
			if (!is_numeric($line))
			{
				$user = UnbUser::FindByDisplayName($line);
				if (!isset($user))
					return '[UNB#40] ' . UnbLocale::Translate('error:40', 'name', UnbLocale::Quote($line));   // User name "$line" is unknown
				$userId = $user->GetUserId();
			}
			else
			{
				$userId = intval($line);
				#if ($userId == -1 && isset($this->newUser))   // Insert the newly created user
				#	$userId = $this->newUser->GetUserId();
				if ($userId == -1 && UnbSession::IsLoggedIn())   // Insert the currently logged in user
					$userId = UnbSession::GetSessionUserId();
				if ($userId <= 0)
					return '[UNB#35] ' . UnbLocale::Translate('error:35', 'id', $line);   // User ID $line is invalid
				$user = UnbUser::FindByUserId($userId);
				if (!isset($user))
					return '[UNB#36] ' . UnbLocale::Translate('error:36', 'name', $line);   // User ID $line is unknown
			}
			$keys[] = $userId;
		}
		return $keys;
	}

	// Validates the request.
	//
	// This performs checks to find out whether the request really comes from the authenticated user.
	//
	// Returns (bool) {{true}} if the request was successfully validated, {{false}} otherwise.
	//
	public static function ValidateRequest()
	{
		if (strlen(UnbEnvironment::GetReferer()) && !UnbEnvironment::IsRefererMyDomain())
		{
			// This request comes with a referer header which does not point to this server's
			// hostname. So this request is likely to be forged from another website to have an
			// authenticated user perform some action that he didn't intend, by having him visit a
			// malicious web page. This is called a Cross Site Request Forgery (CSRF) attack.
			// Since most browsers send this for every request (including XmlHttpRequest), there
			// is little chance for an attacker to use others' browsers for his attack. Of course
			// the referer can only be regarded when it's set. Some network or privacy filters may
			// remove that header on the network level and make it impossible to detect such an
			// attack with this simple and obvious approach.
			return false;
		}

		if (!UnbSession::CheckUrlKey($_REQUEST['urlkey']))
		{
			// This request doesn't come with a URL key or it is invalid. So the request was not
			// initiated by the same browser that would have gotten the previous pages from this
			// web server. The static link might have been placed by a user who wants to trick
			// other authenticated users to perform some action that they didn't intend.
			return false;
		}

		// Passed all tests, should be good
		return true;
	}
} // class UnbFormHandler

?>
