<?php
// UNB Components
// Copyright since 2007 by Yves Goergen
// Website: http://unclassified.de/go/unb2
// Licence: GNU General Public License (GPL) Version 3
//
// This file contains the UnbLog class.

if (!class_exists('UnbDatabase', false)) require_once dirname(__FILE__) . '/database.class.php';
if (!class_exists('UnbSession', false)) require_once dirname(__FILE__) . '/session.class.php';
if (!class_exists('UnbEnvironment', false)) require_once dirname(__FILE__) . '/environment.class.php';

// Enumeration of event types.
//
class UnbLogType
{
	// Information log entry.
	const Information = 1;
	// Warning log entry.
	const Warning = 2;
	// Error log entry.
	const Error = 3;
	// Security-related log entry.
	const Security = 4;
	// Debug message log entry.
	const Debug = 5;
} // class UnbLogType

// Enumeration of event IDs.
//
class UnbLogEvent
{
	// A user has successfully logged in.
	const UserLogin = 1;
	// A user has logged out.
	const UserLogout = 2;
	// A user has tried to log in and failed.
	const UserLoginFailed = 3;

	// The display name of a user was changed.
	const UserDisplayNameChanged = 11;

	// A message revision was approved by a moderator.
	const MessageRevisionApproved = 21;
	// A message revision was locked by a moderator.
	const MessageRevisionLocked = 22;
	// A message revision was unlocked by a moderator.
	const MessageRevisionUnlocked = 23;
	// A message was locked by a moderator.
	const MessageLocked = 24;
	// A message was unlocked by a moderator.
	const MessageUnlocked = 25;
	// A message was hidden by a moderator.
	const MessageHidden = 26;
	// A message was unhidden by a moderator.
	const MessageUnhidden = 27;
} // class UnbLogEvent

// The tag class provides methods for logging events and accessing the log events.
//
// It depends on the UnbDatabase class for a database connection, UnbSession for session user
// recognition and UnbEnvironment for client-specific information.
//
class UnbLog
{
	// Separate database connection
	private static $db = null;

	// Private constructor, does nothing but preventing an unwanted instantiation of this class.
	//
	private function __construct()
	{
	}

	// Gets the separate database connection used for logging
	//
	private static function GetDatabase()
	{
		if (!isset(self::$db))
		{
			self::$db = UnbDatabase::GetNewInstance();
			$connectData = array();

			// Read common database connection settings
			$entries = UnbConfiguration::Get('database.*');
			foreach ($entries as $name => $data)
			{
				$connectData[substr($name, 9)] = $data;
			}

			// Concurrent connections cannot be persistent by default, so turn it off
			$connectData['persistent'] = false;
			// Preset a different filename for SQLite
			if ($connectData['driver'] == 'sqlite')
			{
				$connectData['filename'] = $connectData['filename'] . '.log.db';
			}

			// Overwrite with log-specific settings
			$entries = UnbConfiguration::Get('database.log.*');
			foreach ($entries as $name => $data)
			{
				$connectData[substr($name, 13)] = $data;
			}

			self::$db->Open($connectData);
		}
		return self::$db;
	}

	// Adds a new log entry.
	//
	// type = (int) Event type. See the UnbLogType constants.
	// id = (int) Event ID. See the UnbLogEvent constants.
	// data = (array) Associative array of event data.
	//        (string) XML document of event data.
	// overrideUserId = (int) Use this user ID for the log instead of the session user.
	//
	// Throws InvalidArgumentException on invalid argument type or value.
	// Throws UnbDatabaseException
	//
	public static function Add($type, $id, $data = null, $overrideUserId = null)
	{
		if (!is_int($type))
			throw new InvalidArgumentException('Invalid argument type: $type');
		if ($type <= 0)
			throw new InvalidArgumentException('Argument out of range: $type');
		if (!is_int($id))
			throw new InvalidArgumentException('Invalid argument type: $id');
		if ($id <= 0)
			throw new InvalidArgumentException('Argument out of range: $id');
		if (isset($data))
		{
			if (!is_array($data) && !is_string($data))
				throw new InvalidArgumentException('Invalid argument type: $data');
			if (is_array($data) && !count($data))
				$data = null;
			if (is_string($data) && !strlen(trim($data)))
				$data = null;
		}
		if (isset($overrideUserId))
		{
			if (!is_int($overrideUserId))
				throw new InvalidArgumentException('Invalid argument type: $overrideUserId');
			if ($overrideUserId < 0)
				throw new InvalidArgumentException('Argument out of range: $overrideUserId');
		}

		// Determine the current user ID.
		// Use a user ID provided with the data in case the event is about a failed login event or such.
		if (isset($overrideUserId))
		{
			$userId = $overrideUserId;
		}
		else
		{
			$userId = UnbSession::GetSessionUserId();
		}

		$dataStr = null;
		if (is_array($data))
		{
			// Convert data array to XML string
			$dataStr = '<data>';
			foreach ($data as $key => $value)
			{
				if (!preg_match('/^[a-z][a-z0-9_-]*$/i', $key))
					throw new InvalidArgumentException('Invalid data key: ' . $key);
				$dataStr .= '<' . $key . '>' . htmlspecialchars($value, ENT_NOQUOTES) . '</' . $key . '>';
			}
			$dataStr .= '</data>';
		}
		else if (is_string($data))
		{
			// Validate XML document
			$xml = new SimpleXMLElement($data);
			$xml = null;
			$dataStr = $data;
		}

		$db = self::GetDatabase();
		$db->StartTransaction();
		try
		{
			$db->Execute(
				'INSERT INTO "#log" ("Num", "Time", "Type", "EventId", "UserId", "ClientAddress", "Data")
					VALUES (?, ?, ?, ?, ?, ?, ?)',
				array(
					new UnbDbAutoValue('log'),
					UnbDatabase::ConvertDateTime(time()),
					$type,
					$id,
					$userId,
					UnbEnvironment::GetClientAddress(),
					$dataStr));

			$db->Commit();
		}
		catch (Exception $ex)
		{
			$db->RollBack();
			throw $ex;
		}

		// While we're here, see if we can purge old log entries
		if (mt_rand(0, 100) <= 3)
		{
			self::Purge();
		}
	}

	// Gets the database table fields that may be accessed by the current session user.
	//
	// Returns (string) comma-separated field names for use in SQL SELECT query.
	//
	private static function GetReadFields()
	{
		if (UnbSession::IsAdministrator())
			return '"Num", "Time", "Type", "EventId", "UserId", "ClientAddress", "Data"';

		if (UnbSession::IsModerator())
			return '"Num", "Time", "Type", "EventId", "UserId", "ClientAddress", "Data"';

		return '"Num", "Time", "Type", "EventId", "Data"';
	}

	// Applies the read restriction that applies to the current session user.
	//
	// filter = (array) Requested filter array
	//
	private static function ApplyReadRestriction(&$filter)
	{
		if (!UnbSession::IsAdministrator() && !UnbSession::IsModerator())
		{
			// Restrict events
			$id = $filter['id'];
			if (!isset($id))
				$id = array();
			else if (!is_array($id))
				$id = array($id);

			$allowedEvents = array(
				UnbLogEvent::UserDisplayNameChanged,
				UnbLogEvent::MessageRevisionApproved,
				UnbLogEvent::MessageRevisionLocked,
				UnbLogEvent::MessageRevisionUnlocked,
				UnbLogEvent::MessageLocked,
				UnbLogEvent::MessageUnlocked,
				UnbLogEvent::MessageHidden,
				UnbLogEvent::MessageUnhidden,
				);

			if (!count($id))
			{
				$id = $allowedEvents;
			}
			else
			{
				$newId = array();
				foreach ($id as $i)
				{
					if (in_array($i, $allowedEvents))
						$newId[] = $i;
				}
				$id = $newId;
			}

			$filter['id'] = array_unique($id);
		}
	}

	// Finds log entries by their event number.
	//
	// eventNumMin = (int) First event number to include in the result
	// eventNumMax = (int) Last event number to include in the result, if specified
	//
	// Throws UnbSecurityException if not logged in or no administrator.
	// Throws InvalidArgumentException on invalid argument type or value.
	// Throws UnbDatabaseException
	//
	// Returns: (array) Array of log entries.
	//
	// Privileges: Must be logged in with administrator privilege.
	//
	public static function FindByEventNum($eventNumMin, $eventNumMax = null)
	{
		if (!is_int($eventNumMin))
			throw new InvalidArgumentException('Invalid argument type: $eventNumMin');
		if ($eventNumMin < 0)
			throw new InvalidArgumentException('Argument out of range: $eventNumMin');
		if (isset($eventNumMax))
		{
			if (!is_int($eventNumMax))
				throw new InvalidArgumentException('Invalid argument type: $eventNumMax');
			if ($eventNumMax < $eventNumMin)
				throw new InvalidArgumentException('Argument out of range: $eventNumMax');
		}

		if (!UnbSession::IsAdministrator())
			throw new UnbSecurityException('You need the administrator privilege to fully read the log.');

		$sql = 'SELECT ' . self::GetReadFields() . ' FROM "#log" WHERE "Num" >= ?';
		$params = array($eventNumMin);
		if (isset($eventNumMax))
		{
			$sql .= ' AND "Num" <= ?';
			$params[] = $eventNumMax;
		}
		$sql .= ' ORDER BY "Num"';

		$db = self::GetDatabase();
		$entries = $db->Query($sql, $params);

		foreach ($entries as &$entry)
		{
			$entry['Time'] = UnbDatabase::ConvertDateTime($entry['Time']);
		}

		return $entries;
	}

	// Finds log entries by their properties. For internal use only.
	//
	// filters = (array) Array of filter array parameters - TODO: documentation
	// countOnly = (int) true to count the events, false to return their full data
	// resolveUsers = (bool) Specifies whether the UserId column is resolved into a UserName column containing the users' display names.
	// parseData = (bool) Specifies whether the Data part is parsed into an array.
	//
	// Throws InvalidArgumentException on invalid argument type or value.
	// Throws UnbDatabaseException
	//
	// Returns: (array) Array of log entries.
	//
	private static function FindByFilterInternal($filters, $countOnly, $resolveUsers = false, $parseData = false)
	{
		if (!is_array($filters))
			throw new InvalidArgumentException('Invalid argument type: $filters');

		$allWhere = '';
		$params = array();

		foreach ($filters as $filter)
		{
			if (!$countOnly)
				self::ApplyReadRestriction($filter);

			$where = '';

			if (isset($filter['time-notbefore']))
			{
				if (!is_int($filter['time-notbefore']))
					throw new InvalidArgumentException('Invalid argument type: $filter[\'time-notbefore\']');
				$where .= ($where ? ' AND ' : '') . '"Time" >= ?';
				$params[] = UnbDatabase::ConvertDateTime($filter['time-notbefore']);
			}
			if (isset($filter['time-notafter']))
			{
				if (!is_int($filter['time-notafter']))
					throw new InvalidArgumentException('Invalid argument type: $filter[\'time-notafter\']');
				$where .= ($where ? ' AND ' : '') . '"Time" <= ?';
				$params[] = UnbDatabase::ConvertDateTime($filter['time-notafter']);
			}
			if (isset($filter['type']))
			{
				if (is_array($filter['type']))
				{
					if (count($filter['type']))
					{
						$where .= ($where ? ' AND ' : '') . '"Type" IN (';
						$c = 0;
						foreach ($filter['type'] as $type)
						{
							if (!is_int($type))
								throw new InvalidArgumentException('Invalid argument type: $filter[\'type\'][...]');
							if ($type <= 0)
								throw new InvalidArgumentException('Argument out of range: $filter[\'type\'][...]');
							$where .= ($c ? ', ' : '') . '?';
							$params[] = $type;
							$c++;
						}
						$where .= ')';
					}
				}
				else if (is_int($filter['type']))
				{
					if ($filter['type'] <= 0)
						throw new InvalidArgumentException('Argument out of range: $filter[\'type\']');
					$where .= ($where ? ' AND ' : '') . '"Type" = ?';
					$params[] = $filter['type'];
				}
				else
				{
					throw new InvalidArgumentException('Invalid argument type: $filter[\'type\']');
				}
			}
			if (isset($filter['id']))
			{
				if (is_array($filter['id']))
				{
					if (count($filter['id']))
					{
						$where .= ($where ? ' AND ' : '') . '"EventId" IN (';
						$c = 0;
						foreach ($filter['id'] as $id)
						{
							if (!is_int($id))
								throw new InvalidArgumentException('Invalid argument type: $filter[\'id\'][...]');
							if ($id <= 0)
								throw new InvalidArgumentException('Argument out of range: $filter[\'id\'][...]');
							$where .= ($c ? ', ' : '') . '?';
							$params[] = $id;
							$c++;
						}
						$where .= ')';
					}
				}
				else if (is_int($filter['id']))
				{
					if ($filter['id'] <= 0)
						throw new InvalidArgumentException('Argument out of range: $filter[\'id\']');
					$where .= ($where ? ' AND ' : '') . '"EventId" = ?';
					$params[] = $filter['id'];
				}
				else
				{
					throw new InvalidArgumentException('Invalid argument type: $filter[\'id\']');
				}
			}
			if (isset($filter['userid']))
			{
				if (is_array($filter['userid']))
				{
					if (count($filter['userid']))
					{
						$where .= ($where ? ' AND ' : '') . '"UserId" IN (';
						$c = 0;
						foreach ($filter['userid'] as $userid)
						{
							if (!is_int($userid))
								throw new InvalidArgumentException('Invalid argument type: $filter[\'userid\'][...]');
							if ($userid <= 0)
								throw new InvalidArgumentException('Argument out of range: $filter[\'userid\'][...]');
							$where .= ($c ? ', ' : '') . '?';
							$params[] = $userid;
							$c++;
						}
						$where .= ')';
					}
				}
				else if (is_int($filter['userid']))
				{
					if ($filter['userid'] <= 0)
						throw new InvalidArgumentException('Argument out of range: $filter[\'userid\']');
					$where .= ($where ? ' AND ' : '') . '"UserId" = ?';
					$params[] = $filter['userid'];
				}
				else
				{
					throw new InvalidArgumentException('Invalid argument type: $filter[\'userid\']');
				}
			}
			if (isset($filter['clientaddress']))
			{
				if (!is_string($filter['clientaddress']))
					throw new InvalidArgumentException('Invalid argument type: $filter[\'clientaddress\']');
				$where .= ($where ? ' AND ' : '') . '"ClientAddress" = ?';
				$params[] = $filter['clientaddress'];
			}
			if (isset($filter['data']))
			{
				if (!is_array($filter['data']))
					throw new InvalidArgumentException('Invalid argument type: $filter[\'data\']');
				foreach ($filter['data'] as $path => $value)
				{
					$where .= ($where ? ' AND ' : '') . 'ExtractValue("Data", ?) = ?';
					$params[] = $path;
					$params[] = $value;
				}
			}

			if (strlen($where))
				$allWhere .= ($allWhere ? ' OR ' : '') . '(' . $where . ')';
		}

		if ($countOnly)
			$sql = 'SELECT COUNT(*)';
		else
			$sql = 'SELECT ' . self::GetReadFields();
		$sql .= ' FROM "#log"';
		if ($allWhere)
			$sql .= ' WHERE ' . $allWhere;
		if (!$countOnly)
			$sql .= ' ORDER BY "Num"';

		$db = self::GetDatabase();
		if ($countOnly)
		{
			return $db->QueryValue($sql, $params);
		}
		else
		{
			$entries = $db->Query($sql, $params);

			foreach ($entries as &$entry)
			{
				$entry['Time'] = UnbDatabase::ConvertDateTime($entry['Time']);

				if ($resolveUsers && isset($entry['UserId']))
				{
					$userId = intval($entry['UserId']);
					if (!isset($userCache[$userId]))
					{
						$user = UnbUser::FindByUserId($userId);
						if (isset($user))
						{
							$userCache[$userId] = $user->GetDisplayName();
						}
						else
						{
							$userCache[$userId] = false;
						}
					}
					$entry['UserName'] = $userCache[$userId];
				}

				if ($parseData && isset($entry['Data']))
				{
					$data = array();
					$xml = new SimpleXMLElement($entry['Data']);
					foreach ($xml->children() as $child)
					{
						$data[$child->getName()] = $child->__toString();
					}
					$entry['ParsedData'] = $data;
				}
			}

			return $entries;
		}
	}

	// Finds log entries by their properties.
	//
	// filter = (array) Array of filter array parameters. See FindByFilterInternal for details.
	// resolveUsers = (bool) Specifies whether the UserId column is resolved into a UserName column containing the users' display names.
	// parseData = (bool) Specifies whether the Data part is parsed into an array.
	//
	// Throws UnbSecurityException if not logged in or no administrator.
	// Throws InvalidArgumentException on invalid argument type or value.
	// Throws UnbDatabaseException
	//
	// Returns: (array) Array of log entries.
	//
	// Privileges: Must be logged in with administrator privilege. Lesser privileged users are restricted.
	//
	public static function FindByFilter($filters, $resolveUsers = false, $parseData = false)
	{
		return self::FindByFilterInternal($filters, false, $resolveUsers, $parseData);
	}

	// Counts log entries filtered by their properties.
	//
	// filter = (array) Array of filter array parameters. See FindByFilterInternal for details.
	//
	// Throws InvalidArgumentException on invalid argument type or value.
	// Throws UnbDatabaseException
	//
	// Returns: (int) Count of matched log entries.
	//
	public static function CountByFilter($filters)
	{
		return self::FindByFilterInternal($filters, true);
	}

	// Finds log entries relatet to the moderation history of a message and revision.
	//
	// This function is a shortcut to FindByFilter with an appropriate set of filters.
	//
	// msg = (UnbMessage) Message object.
	// rev = (UnbMessageRevision) Message revision object. If unset, no revision-related history will be retrieved.
	//
	// Returns: See FindByFilter
	//
	public static function FindMessageModeration($msg, $rev = null)
	{
		if (isset($msg))
		{
			if (!($msg instanceof UnbMessage))
				throw new InvalidArgumentException('Invalid argument type: $msg');
		}
		if (isset($rev))
		{
			if (!($rev instanceof UnbMessageRevision))
				throw new InvalidArgumentException('Invalid argument type: $rev');
		}

		$filters = array();
		if (isset($msg))
		{
			$filters[] = array(
				'time-notbefore' => $msg->GetCreatedTime(),
				'id' => array(
					UnbLogEvent::MessageLocked,
					UnbLogEvent::MessageUnlocked,
					UnbLogEvent::MessageHidden,
					UnbLogEvent::MessageUnhidden,
					),
				'data' => array(
					'messageid' => $msg->GetMessageId(),
					),
				);
		}
		if (isset($rev))
		{
			$filters[] = array(
				'time-notbefore' => $rev->GetCreatedTime(),
				'id' => array(
					UnbLogEvent::MessageRevisionApproved,
					UnbLogEvent::MessageRevisionLocked,
					UnbLogEvent::MessageRevisionUnlocked,
					),
				'data' => array(
					'messagerevisionid' => $rev->GetMessageRevisionId(),
					),
				);
		}
		return UnbLog::FindByFilter($filters, true, true);
	}

	// Deletes old log entries. For internal use only.
	//
	// time = (int) Timestamp before which events will be deleted. For security reasons, the timestamp must not be newer than 7 days ago.
	//
	// Throws InvalidArgumentException on invalid argument type or value.
	// Throws UnbDatabaseException
	//
	private static function DeleteBeforeInternal($time)
	{
		if (!is_int($time))
			throw new InvalidArgumentException('Invalid argument type: $time');
		if ($time > time() - 7 * 86400)
			throw new InvalidArgumentException('Argument out of range: $time');

		// TODO: Don't specify a single time but rather define individual times per event in the code

		$db = self::GetDatabase();
		$db->StartTransaction();
		try
		{
			$db->Execute('DELETE FROM "#log" WHERE "Time" < ?', UnbDatabase::ConvertDateTime($time));
			$db->Commit();
		}
		catch (Exception $ex)
		{
			$db->RollBack();
			throw $ex;
		}
	}

	// Deletes old log entries.
	//
	// time = (int) Timestamp before which events will be deleted. For security reasons, the timestamp must not be newer than 7 days ago.
	//
	// Throws UnbSecurityException if not logged in or no administrator.
	// Throws InvalidArgumentException on invalid argument type or value.
	// Throws UnbDatabaseException
	//
	// Privileges: Must be logged in with administrator privilege
	//
	public static function DeleteBefore($time)
	{
		if (!UnbSession::IsAdministrator())
			throw new UnbSecurityException('You need the administrator privilege to delete log entries.');

		self::DeleteBeforeInternal($time);
	}

	// Deletes old log entries as specified in the configuration.
	//
	// Throws InvalidArgumentException on invalid argument type or value.
	// Throws UnbDatabaseException
	//
	private static function Purge()
	{
		// Make sure that no log entries younger than 7 days are deleted.
		$time = time() - max(7, intval(UnbConfiguration::Get('log.keep-days'))) * 86400;

		self::DeleteBeforeInternal($time);
	}

} // class UnbTag

?>
