<?php
// UNB Components
// Copyright since 2007 by Yves Goergen
// Website: http://unclassified.de/go/unb2
// Licence: GNU General Public License (GPL) Version 3
//
// This file contains the UnbMessageRevision class.

if (!class_exists('UnbDatabase', false)) require_once dirname(__FILE__) . '/database.class.php';
if (!class_exists('UnbMessage', false)) require_once dirname(__FILE__) . '/message.class.php';
if (!class_exists('UnbTag', false)) require_once dirname(__FILE__) . '/tag.class.php';
if (!class_exists('UnbSession', false)) require_once dirname(__FILE__) . '/session.class.php';
if (!class_exists('UnbConfiguration', false)) require_once dirname(__FILE__) . '/configuration.class.php';
if (!class_exists('UnbSecurityException', false)) require_once dirname(__FILE__) . '/unb.class.php';

// Enumeration of moderation states for message revisions.
//
class UnbMessageRevisionModerationState
{
	// This revision is waiting to be approved
	const WaitingForApproval = 0;
	// This message revision has been approved
	const Approved = 1;
	// This message revision has been approved but is now locked
	const Locked = 2;
} // class UnbMessageRevisionModerationState

// Enumeration of message revision retrieval levels regarding content fields (content level).
//
class UnbMessageRevisionContentLevel
{
	// Don't retrieve any content
	const NoContent = 0;
	// Only retrieve the subject and summary fields
	const Subject = 1;
	// Also retrieve the source content field
	const Content = 2;
	// Retrieve both the source content and the cached HTML content
	const HtmlContent = 3;
} // class UnbMessageRevisionContentLevel

// The message revision class represents a single message revision in the UNB system and provides
// methods to find and manipulate message revision instances.
//
// It depends on the UnbDatabase class for a database connection. It also uses the class UnbTag to
// manage message revision tags.
//
class UnbMessageRevision extends UnbEntity
{
	// Message revision ID
	private $messageRevisionId = 0;
	// ID of the message that this revision belongs to
	private $messageId = 0;
	// Time when this revision was created (UNIX timestamp)
	private $createdTime = 0;
	// User who created this revision
	private $author = 0;
	// Subject of the message (revision)
	private $subject = '';
	// Content
	private $content = '';
	// Cached HTML output of the parsed content, used to accelerate displaying of the message (revision)
	private $htmlContent = '';
	// Summary of the changes in this revivion to the previous one
	private $summary = '';
	// Moderation state of this single revision (see UnbMessageRevision::MS_* constants)
	private $moderationState = 0;
	// Is this message revision a draft?
	private $draft = false;
	// Attachment data MIME content type
	private $contentType = '';

	// Cached author UnbUser object
	private $authorUser = null;
	// Remembers which level of content was requested for this revision
	private $contentLevel = 0;
	// Cached revision number
	private $revisionNumber = 0;
	// Cached file attachments list
	private $files = null;

	// Cached UnbDatabase instance
	private $db;

	private function __construct()
	{
		$this->db = UnbDatabase::GetInstance();
	}

	// Creates a new message revision instance for a message.
	//
	// The newly created message revision instance is also stored in the database. If the session
	// user is a moderator, the new message revision is automatically approved.
	//
	// NOTE: This method is only meant to be called from UnbMessage::Create, although this is not
	//       strictly enforced. If calling from somewhere else, take care not to add or remove
	//       attachment data other than indicated in the message instance!
	//
	// messageId = (int/UnbMessage) Message to add the revision to
	// subject = (string) Subject of the new message revision or {{null}} (max. 255 characters)
	// content = (string) Content of the new message revision or {{null}} (max. 16,777,215 characters)
	// summary = (string) Summary of the changes in the new revision to the previous one or {{null}} (max. 255 characters)
	// contentType = (string) MIME content type of attached file data (max. 255 characters, ignored if no data is specified)
	// data = (string) Attached file data contents or {{null}}
	// dataFileName = (string) Local file name for attached file data or {{null}} (do not use together with {{date}})
	// writeToFile = (bool) Write the attached file data into a file instead of the database or {{null}} for automatic decision
	// draft = (bool) Declare the new message revision as draft
	//
	// Throws InvalidArgumentException
	// Throws LengthException
	// Throws Exception if the input file could not be read or copied.
	// Throws UnbSecurityException if not logged in or no read/alter access on the message.
	// Throws UnbDatabaseException
	//
	// Returns: The newly created UnbMessageRevision object, if successful, {{null}} otherwise.
	//
	// Privileges: Must be logged in and need read and alter access on the message
	//
	public static function Create($messageId, $subject, $content, $summary = null, $contentType = null, $data = null, $dataFileName = null, $writeToFile = null, $draft = false)
	{
		if ($messageId instanceof UnbMessage)
			$messageId = $messageId->GetMessageId();
		if (!is_int($messageId))
			throw new InvalidArgumentException('Invalid argument type: $messageId');
		if ($messageId <= 0)
			throw new InvalidArgumentException('Argument out of range: $messageId');
		if (isset($subject))
		{
			if (!is_string($subject))
				throw new InvalidArgumentException('Invalid argument type: $subject');
			$subject = trim($subject);
			if (mb_strlen($subject) > 255)
				throw new LengthException('[UNB#4] ' . UnbLocale::Translate('error:4'));   // Subject is too long
			if (!strlen($subject)) $subject = null;
		}
		if (isset($content))
		{
			if (!is_string($content))
				throw new InvalidArgumentException('Invalid argument type: $content');
			// UNIX line endings
			$content = str_replace("\r", '', $content);
			// Trim all lines at their end
			$content = preg_replace('_[ \t]+(?=\n)_', '', $content);
			// Trim content at the end
			$content = rtrim($content);
			if (mb_strlen($content) > 16777215)
				throw new LengthException('[UNB#3] ' . UnbLocale::Translate('error:3'));   // Content is too long
			if (!strlen(trim($content))) $content = null;
		}
		if (isset($summary))
		{
			if (!is_string($summary))
				throw new InvalidArgumentException('Invalid argument type: $summary');
			$summary = trim($summary);
			if (mb_strlen($summary) > 255)
				throw new LengthException('[UNB#5] ' . UnbLocale::Translate('error:5'));   // Summary is too long
			if (!strlen($summary)) $summary = null;
		}
		if (isset($contentType))
		{
			if (!is_string($contentType))
				throw new InvalidArgumentException('Invalid argument type: $contentType');
			$contentType = trim($contentType);
			if (mb_strlen($contentType) > 255)
				throw new LengthException('[UNB#39] Content type is too long');
			if (!strlen($contentType)) $contentType = null;
		}
		if (isset($data))
		{
			if (!is_string($data))
				throw new InvalidArgumentException('Invalid argument type: $data');
		}
		if (isset($dataFileName))
		{
			if (!is_string($dataFileName))
				throw new InvalidArgumentException('Invalid argument type: $dataFileName');
			if (isset($data))
				throw new InvalidArgumentException('You cannot set $data and $dataFileName at the same time.');
			if (!strlen($dataFileName)) $dataFileName = null;
			if (isset($dataFileName) && !file_exists($dataFileName))
				throw new InvalidArgumentException('The specified file does not exist.');
		}
		if (!is_bool($draft))
			throw new InvalidArgumentException('Invalid argument type: $draft');
		if ((isset($data) || isset($dataFileName)) && $draft)
			throw new Exception('Drafts cannot be used with attached data.');

		if (!UnbSession::IsLoggedIn())
			throw new UnbSecurityException('You must be logged in to create a message revision.');

		// Check for read and alter access on the message
		if (!UnbMessage::IsReadAllowedById($messageId))
			throw new UnbSecurityException('You are not allowed to read this message.');
		if (!UnbMessage::IsAlterAllowedById($messageId))
			throw new UnbSecurityException('You are not allowed to alter this message.');

		$now = time();

		$db = UnbDatabase::GetInstance();
		$db->StartTransaction();
		try
		{
			$fields = array(
				'MessageRevisionId',
				'MessageId',
				'CreatedTime',
				'Author',
				'Subject',
				'Content',
				'Summary',
				'Draft');
			$parameters = array(
				new UnbDbAutoValue('message_revision'),
				$messageId,
				UnbDatabase::ConvertDateTime($now),
				UnbSession::GetSessionUserId(),
				$subject,
				$content,
				$summary,
				$draft ? 1 : 0);

			$autoApprove = false;
			if (UnbSession::IsModerator())
				$autoApprove = true;
			if (UnbSession::IsTrusted() && UnbConfiguration::Get('moderation.approve-from-trusted', false))
				$autoApprove = true;
			if ($autoApprove)
			{
				// Moderators' new message (revisions) are always approved by default (self-approval)
				// TODO: Self-approval should be disabled for a specified list of "banned" users by option
				//       Use another system key for that option? Like UnbUserKey::Banned
				//       (Moderators cannot be banned - they must be excluded from this option)
				// TODO: This is the place to regard the EnforceApproval flag from referenced messages
				$fields[] = 'ModerationState';
				$parameters[] = UnbMessageRevisionModerationState::Approved;
			}

			// Handle attached file data
			if (isset($data) || isset($dataFileName))
			{
				if (!isset($writeToFile))
				{
					// No indication whether to save data in a file or in the database, ask our configuration
					$dbMaxSize = UnbConfiguration::Get('attachment.storage.database-max-size', -1);
					if ($dbMaxSize == -1)
					{
						// Always write to the database (default if unconfigured)
						$writeToFile = false;
					}
					else if ($dbMaxSize == 0)
					{
						// Always write to a file
						$writeToFile = true;
					}
					else
					{
						// Write to the database up to a limit, larger data goes into a file
						if (is_string($data))
							$dataLength = strlen($data);
						else
							$dataLength = filesize($dataFileName);
						$writeToFile = $dataLength > $dbMaxSize;
					}
				}

				if (!$writeToFile)
				{
					// Output is database
					if (is_string($dataFileName))
					{
						// Input is a file
						$data = file_get_contents($dataFileName);
						if ($data === false)
							throw new Exception('[UNB#8] The data could not be read from the input file.');
					}

					$fields[] = 'Data';
					$parameters[] = $data;
				}

				$fields[] = 'ContentType';
				$parameters[] = $contentType;
			}
			else
			{
				$contentType = null;
			}

			$sql = 'INSERT INTO "#message_revision" (';
			for ($i = 0; $i < count($fields); $i++)
				$sql .= ($i ? ', ' : '') . '"' . $fields[$i] . '"';
			$sql .= ') VALUES (';
			for ($i = 0; $i < count($fields); $i++)
				$sql .= ($i ? ', ' : '') . '?';
			$sql .= ')';

			$db->Execute($sql, $parameters);

			$messageRevisionId = $db->GetLastAutoValue();
			UnbMessage::UpdateSearchFields($messageId);

			// Write attached data to file, now that the revision ID is known
			if ((isset($data) || isset($dataFileName)) && $writeToFile)
			{
				// Output is a file
				// Now write the new file
				$newFilename = self::GetUploadPath() . $messageId . '_' . $messageRevisionId;
				if (isset($dataFileName))
				{
					// Input is a file
					if (is_uploaded_file($dataFileName))
					{
						if (!move_uploaded_file($dataFileName, $newFilename))
							throw new Exception('[UNB#6] The uploaded file could not be moved to its new location.');
					}
					else
					{
						if (!copy($dataFileName, $newFilename))
							throw new Exception('[UNB#6] The input file could not be copied to its new location.');
					}
				}
				else
				{
					// Input is a string
					if (file_put_contents($newFilename, $data) === false)
						throw new Exception('[UNB#7] The data could not be written to the local file.');
				}
			}

			$db->Commit();
		}
		catch (Exception $ex)
		{
			$db->RollBack();
			throw $ex;
		}

		$messageRevision = new self();
		$messageRevision->messageRevisionId = $messageRevisionId;
		$messageRevision->messageId = $messageId;
		$messageRevision->createdTime = $now;
		$messageRevision->author = UnbSession::GetSessionUserId();
		$messageRevision->subject = $subject;
		$messageRevision->content = $content;
		$messageRevision->summary = $summary;
		$messageRevision->draft = $draft;
		$messageRevision->contentType = $contentType;
		return $messageRevision;
	}

	// Creates a new message revision instance for a message, based on this revision.
	//
	// The newly created message revision instance is also stored in the database. The same
	// conditions as for the static {{Create}} method apply. The data for one or multiple fields
	// of the new message revision can be copied from this revision by setting the respective
	// argument to {{true}}. All tags are copied from this revision to the new one.
	//
	// This message revision must have been retrieved with content level UnbMessageRevisionContentLevel::Content.
	//
	// subject = (string) Subject of the new message revision, {{true}} to copy or {{null}} (max. 255 characters)
	// content = (string) Content of the new message revision, {{true}} to copy or {{null}} (max. 16,777,215 characters)
	// summary = (string) Summary of the changes in the new revision to the previous one or {{null}} (max. 255 characters)
	// contentType = (string) MIME content type of attached file data, {{true}} to copy or {{null}} (max. 255 characters)
	// data = (string) Attached file data contents, {{true}} to copy or {{null}}
	// dataFileName = (string) Local file name for attached file data or {{null}} (do not use together with {{data}})
	// writeToFile = (bool) Write the attached file data into a file instead of the database or {{null}} for automatic decision
	// draft = (bool) Declare the new message revision as draft
	// removeFileIds = (array(int)) Message IDs of the assigned files to remove from the copied revision
	//
	// Throws InvalidArgumentException
	// Throws LengthException
	// Throws Exception if the input file could not be read or copied.
	// Throws UnbSecurityException if not logged in or no read/alter access on the message.
	// Throws UnbDatabaseException
	//
	// Returns: The newly created UnbMessageRevision object, if successful, {{null}} otherwise.
	//
	// Privileges: Must be logged in and need read and alter access on the message
	//
	public function CreateCopy($subject, $content, $summary, $contentType, $data, $dataFileName, $writeToFile, $draft = false, $removeFileIds = null)
	{
		if (isset($subject) && $subject !== true)
		{
			if (!is_string($subject))
				throw new InvalidArgumentException('Invalid argument type: $subject');
			$subject = trim($subject);
			if (mb_strlen($subject) > 255)
				throw new LengthException('[UNB#4] ' . UnbLocale::Translate('error:4'));   // Subject is too long
			if (!strlen($subject)) $subject = null;
		}
		if (isset($content) && $content !== true)
		{
			if (!is_string($content))
				throw new InvalidArgumentException('Invalid argument type: $content');
			// UNIX line endings
			$content = str_replace("\r", '', $content);
			// Trim all lines at their end
			$content = preg_replace('_[ \t]+(?=\n)_', '', $content);
			// Trim content at the end
			$content = rtrim($content);
			if (mb_strlen($content) > 16777215)
				throw new LengthException('[UNB#3] ' . UnbLocale::Translate('error:3'));   // Content is too long
			if (!strlen(trim($content))) $content = null;
		}
		if (isset($summary))
		{
			if (!is_string($summary))
				throw new InvalidArgumentException('Invalid argument type: $summary');
			$summary = trim($summary);
			if (mb_strlen($summary) > 255)
				throw new LengthException('[UNB#5] ' . UnbLocale::Translate('error:5'));   // Summary is too long
			if (!strlen($summary)) $summary = null;
		}
		if (!is_bool($draft))
			throw new InvalidArgumentException('Invalid argument type: $draft');
		if (isset($removeFileIds))
		{
			if (!is_array($removeFileIds))
				throw new InvalidArgumentException('Invalid argument type: $removeFileIds');
			if (!count($removeFileIds)) $removeFileIds = null;
		}

		if ($this->contentLevel < UnbMessageRevisionContentLevel::Content)
			throw new UnbSecurityException('This revision was not retrieved with an appropriate content level to get the content.');

		if ($data === true)
		{
			$data = $this->GetFileData(2);   // Only copy in-database data to this variable
			$dataFileName = $this->GetFileDataPath();   // Ignore possible local argument

			// NOTE:
			// MySQL doesn't currently support SELECTing from a table being UPDATEd. Sad but true.
			// So even if the data is copied from one revision to the new all inside the database,
			// it needs to be read into PHP and written again the regular way.
			// See http://dev.mysql.com/doc/refman/5.1/en/update.html (at the very bottom)
		}
		else if ($this->IsFileDataAvailable() && !isset($data) && !isset($dataFileName))
		{
			throw new Exception('Attachment data cannot be removed from a message revision.');
		}
		else if (!$this->IsFileDataAvailable() && (isset($data) || isset($dataFileName)))
		{
			throw new Exception('Attachment data cannot be added to a message revision.');
		}

		if (isset($removeFileIds))
			$removeFileIdsSql = join(', ', $removeFileIds);

		$db = UnbDatabase::GetInstance();
		$db->StartTransaction();
		try
		{
			if ($this->draft)
			{
				$now = time();

				// This revision is a draft. It can only be updated but no new revision can be
				// created after a draft revision.

				$modState = UnbMessageRevisionModerationState::WaitingForApproval;
				if (UnbSession::IsModerator())
					$modState = UnbMessageRevisionModerationState::Approved;
				if (UnbSession::IsTrusted() && UnbConfiguration::Get('moderation.approve-from-trusted', false))
					$modState = UnbMessageRevisionModerationState::Approved;
				// TODO: See TODOs about auto approval in Create() method

				$sql = 'UPDATE "#message_revision"
					SET "CreatedTime" = ?,
						"Author" = ?,
						"Subject" = ?,
						"Content" = ?,
						"Summary" = ?,
						"ModerationState" = ?,
						"Draft" = ?
					WHERE "MessageRevisionId" = ?
						AND "Draft"';
				$parameters = array(
					UnbDatabase::ConvertDateTime($now),
					UnbSession::GetSessionUserId(),
					($subject === true ? $this->subject : $subject),
					($content === true ? $this->content : $content),
					$summary,   // Do not copy the summary field
					$modState,
					$draft,
					$this->messageRevisionId);
				$db->Execute($sql, $parameters);

				// Verify effect
				if ($db->GetRowsAffected() == 0)
					throw new Exception('No message revision was created. Has the draft been submitted already?');

				// Remove selected files
				if (isset($removeFileIds))
				{
					$db->Execute(
						'DELETE FROM "#message_revision_file" WHERE "MessageRevisionId" = ? AND "MessageId" IN (' . $removeFileIdsSql . ')',
						$this->messageRevisionId);
				}

				// Update the message's creation time if it's a draft, too
				$db->Execute(
					'UPDATE "#message" SET "CreatedTime" = ? WHERE "MessageId" = ? AND "Draft"',
					array(UnbDatabase::ConvertDateTime($now), $this->messageId));

				if (!$draft)
				{
					// Make sure that the message's draft state is also reset
					$db->Execute(
						'UPDATE "#message" SET "Draft" = 0 WHERE "MessageId" = ? AND "Draft"',
						$this->messageId);
				}

				UnbMessage::UpdateSearchFields($this->messageId);

				// No new revision was created, but one must be returned
				$newRevision = $this;
			}
			else
			{
				// Create a new revision
				$newRevision = self::Create(
					$this->messageId,
					($subject === true ? $this->subject : $subject),
					($content === true ? $this->content : $content),
					$summary,   // Do not copy the summary field
					($contentType === true ? $this->contentType : $contentType),
					$data,
					$dataFileName,
					$writeToFile,
					$draft);

				// Copy all tags
				$sql = 'INSERT INTO "#message_revision_tag"
					("MessageRevisionId", "TagId", "TagData")
					SELECT ? "MessageRevisionId", "TagId", "TagData"
					FROM "#message_revision_tag"
					WHERE "MessageRevisionId" = ?';
				$parameters = array(
					$newRevision->GetMessageRevisionId(),
					$this->messageRevisionId);
				$db->Execute($sql, $parameters);

				// Copy all files except those selected to remove
				$sql = 'INSERT INTO "#message_revision_file"
					("MessageRevisionId", "MessageId", "Relation")
					SELECT ? "MessageRevisionId", "MessageId", "Relation"
					FROM "#message_revision_file"
					WHERE "MessageRevisionId" = ?';
				if (isset($removeFileIds))
				{
					$sql .= ' AND "MessageId" NOT IN (' . $removeFileIdsSql . ')';
				}
				$parameters = array(
					$newRevision->GetMessageRevisionId(),
					$this->messageRevisionId);
				$db->Execute($sql, $parameters);
			}

			$db->Commit();
		}
		catch (Exception $ex)
		{
			$db->RollBack();
			throw $ex;
		}

		return $newRevision;
	}

	// Creates a UnbMessageRevision object from a database record array.
	//
	// rec = (array) Database record
	//
	// Throws InvalidArgumentException on invalid argument type or value.
	//
	// Returns (UnbMessageRevision) object constructed from the database record
	//
	protected static function FromRecord($rec)
	{
		if (!is_array($rec))
			throw new InvalidArgumentException('Invalid argument type: $rec');
		if (!count($rec))
			throw new InvalidArgumentException('Record array is empty');

		$messageRevision = new self();
		$messageRevision->messageRevisionId = intval($rec['MessageRevisionId']);
		$messageRevision->messageId = intval($rec['MessageId']);
		$messageRevision->createdTime = UnbDatabase::ConvertDateTime($rec['CreatedTime']);
		$messageRevision->author = intval($rec['Author']);
		$messageRevision->subject = $rec['Subject'];
		$messageRevision->content = $rec['Content'];
		$messageRevision->htmlContent = $rec['HtmlContent'];
		$messageRevision->summary = $rec['Summary'];
		$messageRevision->moderationState = intval($rec['ModerationState']);
		$messageRevision->draft = UnbDatabase::AsBool($rec['Draft']);
		$messageRevision->contentType = $rec['ContentType'];
		return $messageRevision;
	}

	// Reads a message revision from the database.
	//
	// This function is kind of a "service provider" for UnbMessageRevision::FindByRevisionId()
	// and UnbMessageRevision::FindLatestByMessageId().
	//
	// Only one of both parameters messageId and messageRevisionId must be passed since either one
	// is unique on its own.
	//
	// messageId = (int/UnbMessage) Message or its ID to read
	// messageRevisionId = (int) Message revision ID
	// contentLevel = (int) See UnbMessageRevision::FindByRevisionId()
	// allowFallback = (bool) Allow falling back to a lower content level if moderation restrictions require it
	//
	// Throws InvalidArgumentException on invalid argument type or value.
	// Throws UnbDatabaseException
	//
	// Privileges: Need read access on the message. Additionally, message and message revision
	// must not have a moderation lock of any kind. Exceptions are documented separately.
	//
	private static function ReadInternal($messageId, $messageRevisionId, $contentLevel, $allowFallback)
	{
		if (isset($messageId))
		{
			if ($messageId instanceof UnbMessage)
				$messageId = $messageId->GetMessageId();
			if (!is_int($messageId))
				throw new InvalidArgumentException('Invalid argument type: $messageId');
			if ($messageId <= 0)
				throw new InvalidArgumentException('Argument out of range: $messageId');
		}
		if (isset($messageRevisionId))
		{
			if (!is_int($messageRevisionId))
				throw new InvalidArgumentException('Invalid argument type: $messageRevisionId');
			if ($messageRevisionId <= 0)
				throw new InvalidArgumentException('Argument out of range: $messageRevisionId');
		}
		if (!isset($messageId) && !isset($messageRevisionId))
			throw new InvalidArgumentException('Neither $messageId nor $messageRevisionId is set.');
		if (isset($messageId) && isset($messageRevisionId))
			throw new InvalidArgumentException('Both $messageId and $messageRevisionId are set.');
		if (!isset($contentLevel))
			$contentLevel = UnbMessageRevisionContentLevel::Content;
		if (!is_int($contentLevel))
			throw new InvalidArgumentException('Invalid argument type: $contentLevel');
		if ($contentLevel < 0)
			throw new InvalidArgumentException('Argument out of range: $contentLevel');

		$contentFields = '';
		if ($contentLevel == UnbMessageRevisionContentLevel::Subject)
			$contentFields .= 'mr."Subject", mr."Summary", ';
		else if ($contentLevel == UnbMessageRevisionContentLevel::Content)
			$contentFields .= 'mr."Subject", mr."Summary", mr."Content", ';
		else if ($contentLevel == UnbMessageRevisionContentLevel::HtmlContent)
			$contentFields .= 'mr."Subject", mr."Summary", mr."Content", mr."HtmlContent", ';

		$sessionUser = UnbSession::GetSessionUser();
		$sessionUserId = UnbSession::GetSessionUserId();

		if ($allowFallback)
		{
			// Probe all messages with the least content level that can be fallen back to.
			// But first, remember what content level we would like to get at best.
			$maxContentLevel = $contentLevel;
			$contentLevel = UnbMessageRevisionContentLevel::NoContent;
		}

		$db = UnbDatabase::GetInstance();

		if (!isset($messageId))
		{
			$messageId = $db->QueryValue('SELECT "MessageId" FROM "#message_revision" WHERE "MessageRevisionId" = ?', $messageRevisionId);
			if (!isset($messageId)) return null;
			$messageId = intval($messageId);
		}

		// Is reading the message (by the current session user) generally allowed?
		$isReadAllowed = UnbMessage::IsReadAllowedById($messageId);
		// If not:
		// Administrators and the message owner must be able to get the revision object to see the
		// tags and be able to change their access key lists. This is only allowed when no content
		// more than the subject (for handling convenience) is requested.  (Message owners are
		// handled directly in SQL, not here.)
		$isMsgAccessException = UnbSession::IsAdministrator() && $contentLevel <= UnbMessageRevisionContentLevel::Subject;

		// This is an exception value that determines the conditions when the current session user
		// will get access to a message revision when the message has restricting moderation flags
		// set. (Message owners are handled directly in SQL, not here.)
		$isMsgModException = UnbSession::IsModerator() ||
			$contentLevel == UnbMessageRevisionContentLevel::NoContent ||
			UnbSession::IsAdministrator() && $contentLevel <= UnbMessageRevisionContentLevel::Subject;
		// TODO: Administrators may also be allowed to read hidden/locked/unapproved message revisions.
		//       If so, add this:
		//       $isMsgModException |= UnbSession::IsAdministrator();

		$sql = 'SELECT mr."MessageRevisionId", mr."MessageId", mr."CreatedTime", mr."Author", ' . $contentFields . 'mr."Draft", mr."ModerationState", mr."ContentType", m."ModerationFlags"
			FROM "#message" m
				JOIN "#message_revision" mr USING ("MessageId")
			WHERE mr."MessageId" = ? AND

				(? OR   -- Message access by keylists OR Exempted user condition for message access by keylists
				m."Owner" = ? AND ?) AND   -- Condition for message owner to get access to the message

				(m."ModerationFlags" & ' . (UnbMessageModerationFlag::Locked | UnbMessageModerationFlag::Hidden) . ' = 0 OR
				? OR   -- Exempted user condition for message access by moderation flags
				m."Owner" = ?) AND
				mr."MessageRevisionId" = ';

		$parameters = array(
			$messageId,

			$isReadAllowed || $isMsgAccessException,   // Message access by keylists OR Exempted user condition for message access by keylists
			$sessionUserId,
			$contentLevel <= UnbMessageRevisionContentLevel::Subject,   // Condition for message owner to get access to the message

			$isMsgModException,   // Exempted user condition for message access by moderation flags
			$sessionUserId);

		// This is an exception value that determines the conditions when the current session user
		// will get access to a message revision when the message revision has a restricting
		// moderation state. (Revision authors are handled directly in SQL, not here.)
		#$isRevModException = UnbSession::IsModerator() ||
		#	$contentLevel == UnbMessageRevisionContentLevel::NoContent ||
		#	UnbSession::IsAdministrator() && $contentLevel <= UnbMessageRevisionContentLevel::Subject;
		$isRevModException = $isMsgModException;

		if (isset($messageRevisionId))
		{
			// Continue the query like we know what revision number we're looking for
			$sql .= '? AND
					(mr."ModerationState" = ' . UnbMessageRevisionModerationState::Approved . ' OR
					? OR   -- This will be a boolean value indicating whether the sessionUser is an exempted user
					mr."Author" = ?)';
			$parameters[] = $messageRevisionId;
			$parameters[] = $isRevModException;
			$parameters[] = $sessionUserId;
		}
		else
		{
			// Continue the query like get us the latest accessible revision
			$sql .= '(SELECT MAX(mr2."MessageRevisionId")
						FROM "#message_revision" mr2
						WHERE mr2."MessageId" = mr."MessageId" AND
							(NOT mr2."Draft" OR m."Draft") AND
							(mr2."ModerationState" = ' . UnbMessageRevisionModerationState::Approved . ' OR
							? OR   -- This will be a boolean value indicating whether the sessionUser is an exempted user
							mr2."Author" = ?))';
			$parameters[] = $isRevModException;
			$parameters[] = $sessionUserId;
		}

		$rec = $db->QueryRecord($sql, $parameters);
		if (!$rec) return null;

		if ($allowFallback)
		{
			// Now is the time to see what maximum content level we can reach for this message.
			// The data has already been read for the maximum desired content level, now remove
			// the fields that are actually forbidden to read.

			// Start with the assumption that everything is allowed. This will be narrowed down
			// with the following checks.
			$possibleContentLevel = UnbMessageRevisionContentLevel::HtmlContent;

			if (!$isReadAllowed)
			{
				// Message read access by keylists
				// The message owner is handled just like an administrator in this case.
				if (UnbSession::IsAdministrator() || intval($rec['Owner']) == $sessionUserId)
				{
					$possibleContentLevel = min(UnbMessageRevisionContentLevel::Subject, $possibleContentLevel);
				}
				else
				{
					$possibleContentLevel = UnbMessageRevisionContentLevel::NoContent;
				}
			}

			if (intval($rec['ModerationFlags']) & (UnbMessageModerationFlag::Locked | UnbMessageModerationFlag::Hidden) ||
				intval($rec['ModerationState']) != UnbMessageRevisionModerationState::Approved)
			{
				// At least one of these: (they have the same effects)
				// * Message.ModerationFlags contains Hidden or Locked
				// * MessageRevision.ModerationState is not Approved (either Waiting or Locked)
				if (UnbSession::IsModerator() || intval($rec['Author']) == $sessionUserId)
				{
					// No restriction
				}
				else if (UnbSession::IsAdministrator())
				{
					$possibleContentLevel = min(UnbMessageRevisionContentLevel::Subject, $possibleContentLevel);
				}
				else
				{
					$possibleContentLevel = UnbMessageRevisionContentLevel::NoContent;
				}
			}

			// Now apply the maximum possible AND desired content level and clear excess fields
			$contentLevel = min($possibleContentLevel, $maxContentLevel);
			if ($contentLevel < UnbMessageRevisionContentLevel::Subject)
				$rec['Subject'] = $rec['Summary'] = null;
			if ($contentLevel < UnbMessageRevisionContentLevel::Content)
				$rec['Content'] = null;
			if ($contentLevel < UnbMessageRevisionContentLevel::HtmlContent)
				$rec['HtmlContent'] = null;
		}

		$rev = self::FromRecord($rec);
		$rev->contentLevel = $contentLevel;
		return $rev;
	}

	// Finds a message revision instance by its message revision ID.
	//
	// messageRevisionId = (int) Message revision ID
	// contentLevel = (int) Specifies whether the content should be retrieved from the database, too.
	//                     This can be one of the following values:
	//                     * UnbMessageRevisionContentLevel::NoContent - Don't retrieve any content
	//                     * UnbMessageRevisionContentLevel::Subject - Retrieve the subject and summary only
	//                     * UnbMessageRevisionContentLevel::Content - Also retrieve the normal content field (default if unset)
	//                     * UnbMessageRevisionContentLevel::HtmlContent - Retrieve the normal content field and its cached HTML version.
	// allowFallback = (bool) Allow falling back to a lower content level if moderation restrictions require it
	//
	// Throws InvalidArgumentException on invalid argument type or value.
	// Throws UnbDatabaseException
	//
	// Returns: A UnbMessageRevision object, if successful, {{null}} otherwise.
	//
	// Privileges: Need read access on the message. Additionally, message and message revision
	// must not have a moderation lock of any kind. Exceptions are documented separately.
	//
	public static function FindByRevisionId($messageRevisionId, $contentLevel = null, $allowFallback = false)
	{
		if (!is_int($messageRevisionId))
			throw new InvalidArgumentException('Invalid argument type: $messageRevisionId');
		if ($messageRevisionId <= 0)
			throw new InvalidArgumentException('Argument out of range: $messageRevisionId');

		return self::ReadInternal(null, $messageRevisionId, $contentLevel, $allowFallback);
	}

	// Finds the latest accessible message revision instance by its message ID and moderation state.
	//
	// This method will not find message revisions that are a draft.
	//
	// messageId = (int/UnbMessage) Message or its ID of which to read the revision
	// contentLevel = (int) See UnbMessageRevision::FindByRevisionId()
	// allowFallback = (bool) Allow falling back to a lower content level if moderation restrictions require it
	//
	// Throws UnbDatabaseException
	//
	// Returns: A UnbMessageRevision object, if successful, {{null}} otherwise.
	//
	// Privileges: Need read access on the message. Additionally, message and message revision
	// must not have a moderation lock of any kind. Exceptions are documented separately.
	//
	public static function FindLatestByMessageId($messageId, $contentLevel = null, $allowFallback = false)
	{
		return self::ReadInternal($messageId, null, $contentLevel, $allowFallback);
	}

	// Finds all message revisions by their message ID.
	//
	// messageId = (int/UnbMessage) Message or its ID of which to read the revisions
	// includeDrafts = (bool) Indicates whether drafts are included
	//
	// Throws UnbDatabaseException
	//
	// Returns: Array of UnbMessageRevision objects, if successful, {{null}} otherwise.
	//
	// Privileges: Need read access on the message. Additionally, message and message revision
	// must not have a moderation lock of any kind. Exceptions are documented separately.
	//
	public static function FindAllByMessageId($messageId, $includeDrafts = false)
	{
		if ($messageId instanceof UnbMessage)
			$messageId = $messageId->GetMessageId();
		if (!is_int($messageId))
			throw new InvalidArgumentException('Invalid argument type: $messageId');
		if ($messageId <= 0)
			throw new InvalidArgumentException('Argument out of range: $messageId');
		if (!is_bool($includeDrafts))
			throw new InvalidArgumentException('Invalid argument type: $includeDrafts');

		// Is reading the message (by the current session user) generally allowed?
		$isReadAllowed = UnbMessage::IsReadAllowedById($messageId);
		// If not:
		// Administrators and the message owner must be able to get the revision object to see the
		// tags and be able to change their access key lists. This is only allowed when no content
		// more than the subject (for handling convenience) is requested. (Message owners are
		// handled directly in SQL, not here.)
		$isMsgAccessException = UnbSession::IsAdministrator();

		$sql = 'SELECT mr."MessageRevisionId", mr."MessageId", mr."CreatedTime", mr."Author", mr."Summary", mr."Draft", mr."ModerationState", mr."ContentType"
			FROM "#message" m
				JOIN "#message_revision" mr USING ("MessageId")
			WHERE mr."MessageId" = ? AND
				(NOT mr."Draft" OR ?) AND
				(? OR   -- Message access by keylists OR Exempted user condition for message access by keylists
				m."Owner" = ?)
			ORDER BY mr."MessageRevisionId"';
		$parameters = array(
			$messageId,
			$includeDrafts,
			$isReadAllowed || $isMsgAccessException,   // Message access by keylists OR Exempted user condition for message access by keylists
			UnbSession::GetSessionUserId());

		$db = UnbDatabase::GetInstance();
		$recs = $db->Query($sql, $parameters);
		if (!isset($recs)) return null;
		$revs = array();
		foreach ($recs as $rec)
		{
			$contentLevel = UnbMessageRevisionContentLevel::Subject;   // NOTE: This also includes the subject, but it's unset here.

			if ($rec['ModerationState'] != UnbMessageRevisionModerationState::Approved &&
				!UnbSession::IsModerator() &&
				UnbSession::GetSessionUserId() != intval($rec['Author']))
			{
				// This revision is not approved and we're neither a moderator nor the revision's author.
				// Hide all user-content fields.
				$rec['Summary'] = null;
				$contentLevel = UnbMessageRevisionContentLevel::NoContent;
			}

			$rev = self::FromRecord($rec);
			$rev->contentLevel = $contentLevel;
			$revs[] = $rev;
		}
		return $revs;
	}

	// Gets the number of revisions of a message.
	//
	// messageId = (int/UnbMessage) Message or its ID of which to count the revisions
	// includeDrafts = (bool) Indicates whether drafts are included
	//
	// Throws UnbDatabaseException
	//
	// Returns: (int) Number of revisions for this message.
	//
	// Privileges: Need read access on the message.
	//
	public static function CountByMessageId($messageId, $includeDrafts = false)
	{
		if ($messageId instanceof UnbMessage)
			$messageId = $messageId->GetMessageId();
		if (!is_int($messageId))
			throw new InvalidArgumentException('Invalid argument type: $messageId');
		if ($messageId <= 0)
			throw new InvalidArgumentException('Argument out of range: $messageId');
		if (!is_bool($includeDrafts))
			throw new InvalidArgumentException('Invalid argument type: $includeDrafts');

		// Is reading the message (by the current session user) generally allowed?
		$isReadAllowed = UnbMessage::IsReadAllowedById($messageId);

		$sql = 'SELECT COUNT(*)
			FROM "#message" m
				JOIN "#message_revision" mr USING ("MessageId")
			WHERE mr."MessageId" = ? AND
				(NOT mr."Draft" OR ?) AND
				(? OR   -- Message access by keylists OR Exempted user condition for message access by keylists
				m."Owner" = ?)
			ORDER BY mr."MessageRevisionId"';
		$parameters = array(
			$messageId,
			$includeDrafts,
			$isReadAllowed,
			UnbSession::GetSessionUserId());

		$db = UnbDatabase::GetInstance();
		$count = $db->QueryValue($sql, $parameters);
		return $count;
	}

	// Finds all draft message revisions from the logged in user.
	//
	// Throws UnbDatabaseException
	//
	// Returns: Array of (int) MessageRevisionIDs
	//
	public static function FindAllDrafts()
	{
		$db = UnbDatabase::GetInstance();
		$ids = $db->QueryArray(
			'SELECT "MessageRevisionId"
			FROM "#message_revision"
			WHERE "Author" = ? AND
				"Draft"
			ORDER BY "MessageRevisionId"',
			UnbSession::GetSessionUserId());
		$ids = array_map('intval', $ids);
		return $ids;
	}

	// Gets the revision's content level.
	//
	// Returns (int) Content level value, see UnbMessageRevisionContentLevel constants.
	//
	public function GetContentLevel()
	{
		return $this->contentLevel;
	}

	// Gets the revision's message revision ID.
	//
	// Returns (int) Message revision ID
	//
	public function GetMessageRevisionId()
	{
		return $this->messageRevisionId;
	}

	// Gets the revision's message ID.
	//
	// Returns (int) Message ID
	//
	public function GetMessageId()
	{
		return $this->messageId;
	}

	// Gets the time when the revision was created.
	//
	// Returns (int) UNIX timestamp
	//
	public function GetCreatedTime()
	{
		return $this->createdTime;
	}

	// Gets the user ID of the author who created this revision.
	//
	// Returns (int) Author user ID
	//
	public function GetAuthorUserId()
	{
		return $this->author;
	}

	// Gets the UnbUser object of the message revision's author.
	//
	// Returns: (UnbUser) Author user
	//
	public function GetAuthorUser()
	{
		if (!isset($this->authorUser))
			$this->authorUser = UnbUser::FindByUserId($this->author);
		return $this->authorUser;
	}

#	If enabled, this method needs a documentation with parameter, exceptions and return value
#	description and argument type checking. (And exception handling and privileges)
#
#	public function SetAuthor($author)
#	{
#		$this->db->StartTransaction();
#		$r = $this->db->Execute(
#			'UPDATE "#message_revision" SET "Author" = ? WHERE "MessageRevisionId" = ?',
#			array($author, $this->messageRevisionId));
#		if (!$r)
#		{
#			$this->db->RollBack();
#			return false;
#		}
#		$r = $this->db->Commit();
#		if ($r)
#		{
#			$this->author = $author;
#			$this->authorUser = null;   // TODO: Save if passed as argument
#		}
#		return $r;
#	}

	// Gets the message revision's subject.
	//
	// Returns (string) Subject
	//
	public function GetSubject()
	{
		if ($this->contentLevel < UnbMessageRevisionContentLevel::Subject)
			throw new UnbSecurityException('This revision was not retrieved with an appropriate content level to get the subject.');

		return $this->subject;
	}

	// Gets the message revision's full content (if retrieved).
	//
	// Returns (string) Content
	//
	public function GetContent()
	{
		if ($this->contentLevel < UnbMessageRevisionContentLevel::Content)
			throw new UnbSecurityException('This revision was not retrieved with an appropriate content level to get the content.');

		return $this->content;
	}

	// Gets the message revision's cached HTML version of the content (if available and retrieved).
	//
	// autoGenerate = (bool) Automatically generate and store the HTML content if it is not available yet
	// markup = (UnbMarkup) Markup processor to use for generating HTML content
	//
	// Returns (string) HTML content
	//
	public function GetHtmlContent($autoGenerate = false, $markup = null)
	{
		if ($this->contentLevel < UnbMessageRevisionContentLevel::HtmlContent)
			throw new UnbSecurityException('This revision was not retrieved with an appropriate content level to get the HTML content.');

		if (!isset($this->htmlContent) && $autoGenerate)
		{
			if ($this->GetContent() !== null)
			{
				$html = $markup->Convert($this->GetContent());
			}
			else
			{
				$html = '';   // Empty contents should always be empty HTML
			}
			try
			{
				$this->SetHtmlContent($html);
			}
			catch (Exception $ex)
			{
				// Ignore errors here
			}
		}

		return $this->htmlContent;
	}

	// Sets the message revision's cached HTML version of the content.
	//
	// htmlContent = (string) HTML content or {{null}} (max. 16,777,215 characters)
	//
	// Throws InvalidArgumentException
	// Throws LengthException
	// Throws UnbDatabaseException
	//
	public function SetHtmlContent($htmlContent)
	{
		if (isset($htmlContent))
		{
			if (!is_string($htmlContent))
				throw new InvalidArgumentException('Invalid argument type: $htmlContent');
			$htmlContent = trim($htmlContent);
			if (mb_strlen($htmlContent) > 16777215)
				throw new LengthException('[UNB#3] HTML content is too long');
			if (!strlen($htmlContent))
				$htmlContent = null;
		}

		$this->db->StartTransaction();
		try
		{
			$this->db->Execute(
				'UPDATE "#message_revision" SET "HtmlContent" = ? WHERE "MessageRevisionId" = ?',
				array($htmlContent, $this->messageRevisionId));
			$this->db->Commit();
		}
		catch (Exception $ex)
		{
			$this->db->RollBack();
			throw $ex;
		}
		$this->htmlContent = $htmlContent;
	}

	// Gets this revision's summary of changes to the previous one.
	//
	// Returns (string) Summary
	//
	public function GetSummary()
	{
		if ($this->contentLevel < UnbMessageRevisionContentLevel::Subject)
			throw new UnbSecurityException('This revision was not retrieved with an appropriate content level to get the summary.');

		return $this->summary;
	}

	// Gets a value indicating whether this message revision is a draft.
	//
	// Returns: (bool) Message is a draft
	//
	public function IsDraft()
	{
		return $this->draft;
	}

	// Gets this revision's attachment data MIME content type.
	//
	// Returns (string) MIME content type
	//
	public function GetContentType()
	{
		if ($this->contentLevel < UnbMessageRevisionContentLevel::Content)
			throw new UnbSecurityException('This revision was not retrieved with an appropriate content level to get the file content type.');

		return $this->contentType;
	}

	// Gets the length of an attachment file in bytes.
	//
	// searchMode = (int) Specifies how to read the data.
	//                    * 0: Find the data whereever it is, look in the database first.
	//                    * 1: Find the data whereever it is, look in the filesystem first.
	//
	// Throws UnbDatabaseException
	// Throws Exception if the attached data was not found although it's supposed to be there.
	//
	// Returns: (int) The size of the attached file in bytes, or 0 if there is no file attached.
	//
	public function GetFileDataLength($searchMode = 0)
	{
		if ($searchMode == 0)
		{
			$val = $this->db->QueryValue(
				'SELECT LENGTH("Data") FROM "#message_revision"
				WHERE "MessageRevisionId" = ?',
				array($this->messageRevisionId));
		}
		if (!isset($val))
		{
			// There is no data in the database, look for the local file
			$filename = $this->GetFileDataPath();
			if (isset($filename))
			{
				$val = filesize($filename);
			}
			else
			{
				if ($searchMode == 1)
				{
					$val = $this->db->QueryValue(
						'SELECT LENGTH("Data") FROM "#message_revision"
						WHERE "MessageRevisionId" = ?',
						array($this->messageRevisionId));
				}
				if (!isset($val))
					throw new Exception('[UNB#9] Attached data not found (' . $this->messageId . '_' . $this->messageRevisionId . ').');
			}
		}
		return intval($val);
	}

	// Gets the name of the attachment file on disk.
	//
	// nonExistent = (bool) {{true}} to return non-existent file name, useful when the file shall be created
	//
	// Throws UnbDatabaseException
	//
	// This method only returns a file name if the file exists. It is not verified that the
	// attached data exists in the database otherwise. If the constructed file name does not
	// exist, it will not be returned.
	//
	// Returns the name of the attached file, or {{null}} if there is no file attached.
	//
	public function GetFileDataPath($nonExistent = false)
	{
		$filename = self::GetUploadPath() . $this->messageId . '_' . $this->messageRevisionId;
		if ($nonExistent || file_exists($filename))
		{
			return $filename;
		}
		return null;
	}

	// Gets the contents of an attachment file.
	//
	// searchMode = (int) Specifies how to read the data.
	//                    * 0: Find the data whereever it is, look in the database first.
	//                    * 1: Find the data whereever it is, look in the filesystem first.
	//                    * 2: Only read from the database, return {{null}} if it's not there.
	//                    * 3: Use {{readfile}} to read from the filesystem. Returns {{true}} or {{false}} depending on success.
	//
	// Throws UnbDatabaseException
	// Throws UnbSecurityException if the message revision was requested with NoContent mode.
	// Throws Exception if the attached data was not found although it's supposed to be there.
	//
	// Returns the contents of the attached file, or {{null}} if there is no file attached.
	// If searchMode is 3, the return value is a boolean value.
	//
	public function GetFileData($searchMode = 0)
	{
		if ($this->contentLevel < UnbMessageRevisionContentLevel::Content)
			throw new UnbSecurityException('This revision was not retrieved with an appropriate content level to get the file data.');

		if ($searchMode == 0 || $searchMode == 2)
		{
			$val = $this->db->QueryValue(
				'SELECT "Data" FROM "#message_revision"
				WHERE "MessageRevisionId" = ?',
				array($this->messageRevisionId));
		}
		if (!isset($val))
		{
			if ($searchMode == 2) return null;

			// There is no data in the database, look for the local file
			$filename = $this->GetFileDataPath();
			if (isset($filename))
			{
				if ($searchMode == 3)
					return readfile($filename) !== false;

				$val = file_get_contents($filename);
			}
			else
			{
				if ($searchMode == 3)
					return false;   // readfile() doesn't work. Possible data in the database is irrelevant in this case.
				if ($searchMode == 1)
				{
					$val = $this->db->QueryValue(
						'SELECT "Data" FROM "#message_revision"
						WHERE "MessageRevisionId" = ?',
						array($this->messageRevisionId));
				}
				if (!isset($val))
					throw new Exception('[UNB#9] Attached data not found.');
			}
		}
		return $val;
	}

	// Determines whether attachment file data is available.
	//
	// searchMode = (int) Specifies where to find the data.
	//                    * 0: Find the data whereever it is, look in the database first.
	//                    * 2: Only read from the database, return {{false}} if it's not there.
	//
	// Throws UnbDatabaseException
	// Throws UnbSecurityException if the message revision was requested with NoContent mode.
	//
	// Returns {{true}} if data is attached to this revision and available, {{false}} otherwise.
	//
	public function IsFileDataAvailable($searchMode = 0)
	{
		if ($this->contentLevel < UnbMessageRevisionContentLevel::Content)
			throw new UnbSecurityException('This revision was not retrieved with an appropriate content level to get the file data.');

		$val = $this->db->QueryValue(
			'SELECT COUNT("Data") FROM "#message_revision"
			WHERE "MessageRevisionId" = ?',
			$this->messageRevisionId);
		if ($val == 1) return true;
		if ($searchMode == 2) return false;

		// There is no data in the database, look for the local file
		$filename = $this->GetFileDataPath();
		return isset($filename);
	}

	// Gets this revision's moderation state.
	//
	// Returns (int) Moderation state
	//
	public function GetModerationState()
	{
		return $this->moderationState;
	}

	// Sets this revision's moderation state.
	//
	// If the message revision is waiting for approval and is now being approved, all attached
	// files that are waiting for approval will automatically be approved, too.
	//
	// moderationState = (int) See the UnbMessageRevision::MS_* constants.
	//
	// Throws InvalidArgumentException on invalid argument type or value.
	// Throws UnbSecurityException if not logged in or no moderator.
	// Throws UnbDatabaseException
	//
	// Privileges: Must be logged in with moderator privilege
	//
	public function SetModerationState($moderationState)
	{
		if (!is_int($moderationState))
			throw new InvalidArgumentException('Invalid argument type: $moderationState');
		if ($moderationState < 0)
			throw new InvalidArgumentException('Argument out of range: $moderationState');

		if (!UnbSession::IsModerator())
			throw new UnbSecurityException('You need the moderator privilege to set the moderation state.');

		$this->db->StartTransaction();
		try
		{
			$this->db->Execute(
				'UPDATE "#message_revision" SET "ModerationState" = ? WHERE "MessageRevisionId" = ?',
				array($moderationState, $this->messageRevisionId));

			// Log event
			$data = array(
				'messageid' => $this->messageId,
				'messagerevisionid' => $this->messageRevisionId,
				);

			if ($this->moderationState == UnbMessageRevisionModerationState::WaitingForApproval &&
				$moderationState == UnbMessageRevisionModerationState::Approved)
			{
				UnbLog::Add(UnbLogType::Information, UnbLogEvent::MessageRevisionApproved, $data);
			}
			if ($this->moderationState == UnbMessageRevisionModerationState::Approved &&
				$moderationState == UnbMessageRevisionModerationState::Locked)
			{
				UnbLog::Add(UnbLogType::Information, UnbLogEvent::MessageRevisionLocked, $data);
			}
			if ($this->moderationState == UnbMessageRevisionModerationState::Locked&&
				$moderationState == UnbMessageRevisionModerationState::Approved)
			{
				UnbLog::Add(UnbLogType::Information, UnbLogEvent::MessageRevisionUnlocked, $data);
			}

			if ($this->moderationState == UnbMessageRevisionModerationState::WaitingForApproval &&
				$moderationState == UnbMessageRevisionModerationState::Approved)
			{
				// Also approve all unapproved attached files
				$fileIds = $this->GetFileIds();

				#$this->db->Execute(
				#	'UPDATE "#message_revision"
				#	SET "ModerationState" = ?
				#	WHERE "ModerationState" = ?
				#		AND "MessageRevisionId" IN (
				#			SELECT "SearchRevisionId"
				#			FROM "#message"
				#			WHERE "MessageId" IN (
				#				SELECT "MessageId"
				#				FROM "#message_revision_file"
				#				WHERE "MessageRevisionId" = ?))',
				#	array(
				#		UnbMessageRevisionModerationState::Approved,
				#		UnbMessageRevisionModerationState::WaitingForApproval,
				#		$this->messageRevisionId));
				// NOTE: This doesn't let us write log events for each file message revision approved.

				$files = $this->db->Query(
					'SELECT "MessageId", "SearchRevisionId" AS "MessageRevisionId"
					FROM "#message"
					WHERE "MessageId" IN (
						SELECT "MessageId"
						FROM "#message_revision_file"
						WHERE "MessageRevisionId" = ?)',
					array(
						$this->messageRevisionId));

				foreach ($files as $file)
				{
					$this->db->Execute(
						'UPDATE "#message_revision"
						SET "ModerationState" = ?
						WHERE "ModerationState" = ?
							AND "MessageRevisionId" = ?',
						array(
							UnbMessageRevisionModerationState::Approved,
							UnbMessageRevisionModerationState::WaitingForApproval,
							$file['MessageRevisionId']));

					// Log event
					if ($this->db->GetRowsAffected() > 0)
					{
						$data = array(
							'messageid' => $file['MessageId'],
							'messagerevisionid' => $file['MessageRevisionId'],
							);

						UnbLog::Add(UnbLogType::Information, UnbLogEvent::MessageRevisionApproved, $data);
					}
				}
			}

			UnbMessage::UpdateSearchFields($this->messageId);

			$this->db->Commit();
		}
		catch (Exception $ex)
		{
			$this->db->RollBack();
			throw $ex;
		}
		$this->moderationState = $moderationState;
	}

	// Gets the counting number of this revision.
	//
	// Returns (int)
	//
	public function GetRevisionNumber()
	{
		if (!$this->revisionNumber)
		{
			$count = $this->db->QueryValue('
				SELECT COUNT(*)
				FROM "#message_revision"
				WHERE "MessageId" = ?
					AND "MessageRevisionId" < ?',
				array($this->messageId, $this->messageRevisionId));
			$this->revisionNumber = intval($count) + 1;
		}
		return $this->revisionNumber;
	}

	// Deletes this instance of the message revision with all its references.
	//
	// Throws UnbSecurityException if not logged in or no moderator.
	// Throws UnbDatabaseException
	//
	// Privileges: Must be logged in with moderator privilege (except for drafts)
	//
	public function Delete()
	{
		if (!UnbSession::IsModerator() && !$this->draft)
			throw new UnbSecurityException('You need the moderator privilege to delete a message revision.');

		$this->db->StartTransaction();
		try
		{
			// All references are automatically resolved by the database foreign key contraints. :)

			$this->db->Execute(
				'DELETE FROM "#message_revision" WHERE "MessageRevisionId" = ? AND "Draft"',
				array($this->messageRevisionId));

			// Verify effect
			if ($this->db->GetRowsAffected() == 0)
				throw new Exception('No message revision was deleted.');

			// Determine whether a file is attached to this message revision that needs to be deleted
			$dataFilename = $this->GetFileDataPath();
			if (isset($dataFilename))
				if (!unlink($dataFilename))
					throw new Exception('[UNB#10] The attached file could not be deleted.');

			$this->db->Commit();
		}
		catch (Exception $ex)
		{
			$this->db->RollBack();
			throw $ex;
		}
	}

	// Gets UnbTag objects of all tags assigned to this message revision.
	//
	// onlyVisible = (bool) Only get tags that don't have the UnbTagFlag::Invisible flag set
	// names = (array) Tag names to retrieve
	//
	// Throws UnbDatabaseException
	//
	// If a list of tag names was specified, the return value is an associative array with the
	// tags' display name as key. If no names were specified, all tags are returned in a normal
	// array indexed numerically sorted in the correct hierarchical tag order.
	//
	// Returns: (array(UnbTag)) Tags
	//
	public function GetTags($onlyVisible = true, $names = null)
	{
		if (isset($names))
		{
			if (!is_array($names))
				throw new InvalidArgumentException('Invalid argument type: $names');
		}

		// t.TagId would be ambiguous (from both tables), so an alias for this one is required
		$sql = 'SELECT t."TagId" AS "TagId", t.*, mrt."TagData"
			FROM "#message_revision_tag" mrt
				JOIN "#tag" t USING ("TagId")
			WHERE mrt."MessageRevisionId" = ?';
		$parameters = array($this->messageRevisionId);

		if ($onlyVisible)
			$sql .= ' AND NOT (t.Flags & ' . UnbTagFlag::Invisible . ')';

		if (isset($names))
		{
			$sql .= ' AND (';
			$count = 0;
			foreach ($names as $name)
			{
				if (!is_string($name))
					throw new InvalidArgumentException('Invalid argument type: $names');

				if ($count++)
					$sql .= ' OR ';
				$sql .= 't.TagName = ?';
				$parameters[] = $name;
			}
			$sql .= ')';
		}

		#$sql .= ' ORDER BY t."TagName"';
		$order = UnbTag::GetHierarchicalOrder();
		if (count($order))
		{
			$sql .= ' ORDER BY CASE t."TagId"';
			foreach ($order as $index => $tagId)
				$sql .= ' WHEN ' . $tagId . ' THEN ' . $index;
			$sql .= ' END';
		}

		$recs = $this->db->Query($sql, $parameters);
		$tags = array();
		foreach ($recs as $rec)
		{
			if (isset($names))
			{
				$tag = UnbTag::FromRecord($rec);
				$tags[$tag->GetTagName()] = $tag;
			}
			else
			{
				$tags[] = UnbTag::FromRecord($rec);
			}
		}
		return $tags;
	}

	// Gets (unsorted) tag IDs of all tags assigned to this message revision.
	//
	// withData = (bool) {{true}} to include tag data, {{false}} otherwise.
	//
	// Throws UnbDatabaseException
	//
	// Returns: An array of all tag ID numbers. If {{withData}} is set, a tag has the UserData
	// flag set and tag data stored (i.e. not null), the array item will be an
	// array(TagID => (int), TagData => (string)) instead of the (int) number only.
	//
	public function GetTagIds($withData = true)
	{
		$recs = $this->db->Query(
			'SELECT "TagId", "TagData" FROM "#message_revision_tag" WHERE "MessageRevisionId" = ?',
			array($this->messageRevisionId));
		$tags = array();
		foreach ($recs as $rec)
		{
			$rec['TagId'] = intval($rec['TagId']);
			if ($withData && $rec['TagData'] != null)
			{
				$tags[] = $rec;
			}
			else
			{
				$tags[] = $rec['TagId'];
			}
		}
		return $tags;
	}

	// Assigns a tag to this message revision.
	//
	// tagId = (int/UnbTag) Tag or its ID to assign.
	// tagData = (string) Tag data to use for this message revision. (Only for tags with UserData flag,
	//                    {{null}} otherwise. Max. 255 characters)
	//
	// Throws InvalidArgumentException
	// Throws LengthException
	// Throws UnbSecurityException if not logged in or not the same user/no moderator or tag use disallowed.
	// Throws UnbDatabaseException
	//
	// Privileges: Must be logged in as the same user or with moderator privilege. You must be allowed
	// to use the specified tag.
	//
	// TODO: Reset moderation state when a UserData tag has been altered - when? (Except for moderators)
	//
	public function AddTag($tagId, $tagData = null)
	{
		if ($tagId instanceof UnbTag)
			$tagId = $tagId->GetTagId();
		if (!is_int($tagId))
			throw new InvalidArgumentException('Invalid argument type: $tagId');
		if ($tagId <= 0)
			throw new InvalidArgumentException('Argument out of range: $tagId');
		if (isset($tagData))
		{
			if (!is_string($tagData))
				throw new InvalidArgumentException('Invalid argument type: $tagData');
			$tagData = trim($tagData);
			if (mb_strlen($tagData) > 255)
				throw new LengthException('[UNB#11] ' . UnbLocale::Translate('error:11'));   // Tag data is too long
		}

		if (UnbSession::GetSessionUserId() != $this->author && !UnbSession::IsModerator())
			throw new UnbSecurityException('You must be logged in as this user or need the moderator privilege to add a tag.');
		if (!UnbTag::IsUseAllowedById($tagId))
			throw new UnbSecurityException('You are not allowed to use this tag.');

		UnbMessage::ClearAccessCache();
		$this->db->StartTransaction();
		try
		{
			$this->db->Execute(
				'UPDATE "#message_revision_tag"
				SET "TagData" = ?
				WHERE "MessageRevisionId" = ? AND "TagId" = ?',
				array($tagData, $this->messageRevisionId, $tagId));
			if ($this->db->GetRowsAffected() == 0)
			{
				// The entry doesn't exist yet, so insert it.
				try
				{
					$this->db->Execute(
						'INSERT INTO "#message_revision_tag" ("MessageRevisionId", "TagId", "TagData") VALUES (?, ?, ?)',
						array($this->messageRevisionId, $tagId, $tagData));
				}
				catch (UnbDatabaseException $ex)
				{
					// If the entry already exists, it is unchanged.
					// GetRowsAffected() has returned 0 in this case although the record has been found.
					// But it has not been altered because it is unchanged.
					if (!$this->db->IsConstraintError()) throw $ex;
				}
			}
			$this->db->Commit();
		}
		catch (Exception $ex)
		{
			$this->db->RollBack();
			throw $ex;
		}
	}

	// Removes a tag from this message revision.
	//
	// tagId = (int/UnbTag) Tag or its ID to remove.
	//
	// Throws InvalidArgumentException on invalid argument type or value.
	// Throws UnbSecurityException if not logged in or not the same user/no moderator.
	// Throws UnbDatabaseException
	//
	// Privileges: Must be logged in as the same user or with moderator privilege
	//
	public function RemoveTag($tagId)
	{
		if ($tagId instanceof UnbTag)
			$tagId = $tagId->GetTagId();
		if (!is_int($tagId))
			throw new InvalidArgumentException('Invalid argument type: $tagId');
		if ($tagId <= 0)
			throw new InvalidArgumentException('Argument out of range: $tagId');

		if (UnbSession::GetSessionUserId() != $this->author && !UnbSession::IsModerator())
			throw new UnbSecurityException('You must be logged in as this user or need the moderator privilege to remove a tag.');

		UnbMessage::ClearAccessCache();
		$this->db->StartTransaction();
		try
		{
			$this->db->Execute(
				'DELETE FROM "#message_revision_tag" WHERE "MessageRevisionId" = ? AND "TagId" = ?',
				array($this->messageRevisionId, $tagId));
			$this->db->Commit();
		}
		catch (Exception $ex)
		{
			$this->db->RollBack();
			throw $ex;
		}
	}

	// Gets (unsorted) message IDs of all file messages assigned to this message revision.
	//
	// relation = (string) Relation type to restrict the search for, {{null}} for unconditional search.
	//
	// Throws UnbDatabaseException
	//
	// Returns: An array of all message IDs. This data may be cached if requested with no relation
	// type restriction.
	//
	public function GetFileIds($relation = null)
	{
		if (isset($relation))
		{
			if (!is_string($relation))
				throw new InvalidArgumentException('Invalid argument type: $relation');
			$relation = trim($relation);
			if (!strlen($relation))
				$relation = null;
		}

		if (!isset($relation) && isset($this->files))
			return $this->files;   // Return cached data

		$sql = 'SELECT "MessageId" FROM "#message_revision_file" WHERE "MessageRevisionId" = ?';
		$params = array($this->messageRevisionId);

		if (isset($relation))
		{
			$sql .= ' AND "Relation" = ?';
			$params[] = $relation;
		}

		$ids = $this->db->QueryArray($sql, $params);
		$files = array_map('intval', $ids);

		if (!isset($relation))
			$this->files = $files;   // Cache data

		return $files;
	}

	// Gets (unsorted) message IDs mapped to relation types for all file messages assigned to this
	// message revision.
	//
	// Throws UnbDatabaseException
	//
	// Returns: array(Message ID => Relation type)
	//
	public function GetFileRelations()
	{
		$recs = $this->db->Query(
			'SELECT "MessageId", "Relation" FROM "#message_revision_file" WHERE "MessageRevisionId" = ?',
			$this->messageRevisionId);

		$out = array();
		foreach ($recs as $rec)
		{
			$out[intval($rec['MessageId'])] = $rec['Relation'];
		}
		return $out;
	}

	// Assigns a file message to this message revision.
	//
	// messageId = (int/UnbMessage) Message or its ID to assign
	// relation = (string) Relation type of the file to the message revision (max. 255 characters)
	//
	// Throws InvalidArgumentException
	// Throws LengthException
	// Throws UnbSecurityException if not logged in or not the author user.
	// Throws UnbDatabaseException
	//
	// Privileges: Must be logged in as the author user.
	//
	public function AddFile($messageId, $relation = null)
	{
		if ($messageId instanceof UnbMessage)
			$messageId = $messageId->GetMessageId();
		if (!is_int($messageId))
			throw new InvalidArgumentException('Invalid argument type: $messageId');
		if ($messageId <= 0)
			throw new InvalidArgumentException('Argument out of range: $messageId');
		if (isset($relation))
		{
			if (!is_string($relation))
				throw new InvalidArgumentException('Invalid argument type: $relation');
			$relation = trim($relation);
			if (mb_strlen($relation) > 255)
				throw new LengthException('Relation type is too long');
			if (!strlen($relation))
				$relation = null;
		}

		if (UnbSession::GetSessionUserId() != $this->author)
			throw new UnbSecurityException('You must be logged in as the author user.');

		$this->db->StartTransaction();
		try
		{
			$this->db->Execute(
				'INSERT INTO "#message_revision_file" ("MessageRevisionId", "MessageId", "Relation") VALUES (?, ?, ?)',
				array($this->messageRevisionId, $messageId, $relation));

			$this->db->Execute(
				'UPDATE "#message" SET "FileType" = ? WHERE "MessageId" = ? AND "FileType" = ?',
				array(UnbMessageFileType::Embedded, $messageId, UnbMessageFileType::Independent));

			$this->db->Commit();

			if (isset($this->files))
				$this->files[] = $messageId;   // Update cache
		}
		catch (Exception $ex)
		{
			$this->db->RollBack();
			throw $ex;
		}
	}

	// Gets the base upload path. For internal use.
	//
	public static function GetUploadPath()
	{
		$path = UnbConfiguration::Get('path.file-uploads', '.');
		if (strlen($path) == 0 ||   // Empty path
			($path{0} != '/' &&   // Unix rooted path
			 !preg_match('_^[A-Za-z]:(\\\\|/)_', $path) &&   // Windows local rooted path
			 substr($path, 0, 2) != '\\\\'))   // Windows UNC path
		{
			$path = dirname(__FILE__) . '/' . $path;
		}
		if (substr($path, -1) != '/')
		{
			$path .= '/';
		}
		return $path;
	}

} // class UnbMessageRevision

?>
