<?php
// UNB Components
// Copyright since 2007 by Yves Goergen
// Website: http://unclassified.de/go/unb2
// Licence: GNU General Public License (GPL) Version 3
//
// This file contains the UnbTag class.

if (!class_exists('UnbDatabase', false)) require_once dirname(__FILE__) . '/database.class.php';
if (!class_exists('UnbSession', false)) require_once dirname(__FILE__) . '/session.class.php';
if (!class_exists('UnbSecurityException', false)) require_once dirname(__FILE__) . '/unb.class.php';

// Enumeration of tag flags.
//
class UnbTagFlag
{
	// This tag is a hierarchy separator only and cannot be set on messages itself.
	const NotSelectable = 1;
	// This tag is not visible to user selection and will not be displayed,
	// i.e. it is for programme-internal use only.
	const Invisible = 2;
	// Messages which have this tag set need to be approved by a moderator,
	// no matter what other rights the author would normally have.
	const EnforceApproval = 4;
	// This tag can be set to a message with a custom data value set.
	const UserData = 8;
} // class UnbTagFlag

// The tag class represents a single tag in the UNB system and provides methods to find and
// manipulate tag instances.
//
// It depends on the UnbDatabase class for a database connection.
//
class UnbTag extends UnbEntity
{
	// Tag ID
	private $tagId = 0;
	// Displayed name of the tag
	private $tagName = '';
	// ID of the parent tag for this tag, used to organise tags in a tree hierarchy
	private $parentTag = 0;
	// Special meaning of this tag
	private $meaning = 0;
	// Description of this tag that explains the user when and why this tag should be used
	private $description = '';
	// Flags for special tag behaviour (see UnbTagFlag constants)
	private $flags = 0;

	// For use with a tag instance assigned to a message
	private $tagData = null;

	// Cached UnbDatabase instance
	private $db;

	// Caches the correct hierarchical order of all tags
	private static $tagHierarchicalOrder = null;

	private function __construct()
	{
		$this->db = UnbDatabase::GetInstance();
	}

	// Creates a new tag instance.
	//
	// The newly created tag instance is also stored in the database.
	//
	// tagName = (string) New tag's displayed name (max. 255 characters)
	//
	// Throws InvalidArgumentException
	// Throws LengthException
	// Throws UnbSecurityException if not logged in or no moderator.
	// Throws UnbDatabaseException
	//
	// Returns: The newly created UnbTag object
	//
	// Privileges: Must be logged in with moderator privilege
	//
	public static function Create($tagName)
	{
		if (!is_string($tagName))
			throw new InvalidArgumentException('Invalid argument type: $tagName');
		$tagName = trim($tagName);
		if (!strlen($tagName))
			throw new InvalidArgumentException('Tag name is empty');
		if (mb_strlen($tagName) > 255)
			throw new LengthException('[UNB#14] ' . UnbLocale::Translate('error:14'));   // Tag name is too long

		if (!UnbSession::IsModerator())
			throw new UnbSecurityException('You need the moderator privilege to create a tag.');

		$db = UnbDatabase::GetInstance();
		$db->StartTransaction();
		try
		{
			$db->Execute(
				'INSERT INTO "#tag" ("TagId", "TagName") VALUES (?, ?)',
				array(new UnbDbAutoValue('tag'), $tagName));
			$tagId = $db->GetLastAutoValue();

			$db->Commit();
		}
		catch (Exception $ex)
		{
			$db->RollBack();
			throw $ex;
		}

		$tag = new self();
		$tag->tagId = $tagId;
		$tag->tagName = $tagName;
		return $tag;
	}

	// Creates a UnbTag object from a database record array.
	//
	// rec = (array) Database record
	//
	// Throws InvalidArgumentException on invalid argument type or value.
	//
	// Returns: (UnbTag) Object constructed from the database record
	//
	// This method is declared public because it is called by UnbMessageRevision::GetTags().
	//
	protected static function FromRecord($rec)
	{
		if (!is_array($rec))
			throw new InvalidArgumentException('Invalid argument type: $rec');
		if (!count($rec))
			throw new InvalidArgumentException('Record array is empty');

		$tag = new self();
		$tag->tagId = intval($rec['TagId']);
		$tag->tagName = $rec['TagName'];
		$tag->parentTag = intval($rec['ParentTag']);
		$tag->meaning = intval($rec['Meaning']);
		$tag->description = $rec['Description'];
		$tag->flags = intval($rec['Flags']);
		$tag->tagData = $rec['TagData'];
		return $tag;
	}

	// Finds a tag instance or an array of tag instances by its/their tag ID.
	//
	// tagId = (int) Tag ID
	//         (array) Array of tag IDs
	//
	// Throws InvalidArgumentException on invalid argument type or value.
	// Throws UnbDatabaseException
	//
	// Returns: (UnbTag) Tag object if a single tag ID was specified, or (array(UnbTag)) array of
	// tag instances if an array of tag IDs was specified, or {{null}} on error.
	//
	public static function FindByTagId($tagId)
	{
		if (!is_int($tagId) && !is_array($tagId))
			throw new InvalidArgumentException('Invalid argument type: $tagId');
		if (is_int($tagId) && $tagId <= 0)
			throw new InvalidArgumentException('Argument out of range: $tagId');
		if (is_array($tagId) && !count($tagId))
			throw new InvalidArgumentException('Argument out of range: $tagId');

		$db = UnbDatabase::GetInstance();
		if (is_int($tagId))
		{
			$rec = $db->QueryRecord('SELECT * FROM "#tag" WHERE "TagId" = ?', $tagId);
			if (!$rec) return null;
			return self::FromRecord($rec);
		}
		if (is_array($tagId))
		{
			$tagId = array_map('intval', $tagId);
			$rec = $db->Query('SELECT * FROM "#tag" WHERE "TagId" IN (' . join(', ', $tagId) . ')');
			if (!$rec) return null;
			$a = array();
			foreach ($rec as $r)
			{
				$a[] = self::FromRecord($r);
			}
			return $a;
		}
	}

	// Finds one or more tag instances by their tag name.
	//
	// tagName = (string) Tag name (max. 255 characters)
	// caseless = (bool) Compare tag names case-insensitive (default is case-sensitive)
	// getAll = (bool) Return all search results instead only the first (default is only the first result)
	//
	// Throws InvalidArgumentException
	// Throws LengthException
	// Throws UnbDatabaseException
	//
	// Returns a UnbTag object or an array of UnbTag objects, if successful, {{null}} otherwise.
	//
	// Note that this method will only return a single tag if the getAll parameter is false. So if
	// the specified name is not unique (which is possible by the application design), it is
	// undetermined, which of the tags is returned.
	//
	public static function FindByTagName($tagName, $caseless = false, $getAll = false)
	{
		if (!is_string($tagName))
			throw new InvalidArgumentException('Invalid argument type: $tagName');
		$tagName = trim($tagName);
		if (!strlen($tagName))
			throw new InvalidArgumentException('Tag name is empty');
		if (mb_strlen($tagName) > 255)
			throw new LengthException('[UNB#14] ' . UnbLocale::Translate('error:14'));   // Tag name is too long

		$db = UnbDatabase::GetInstance();
		if ($getAll)
		{
			if ($caseless)
				$recs = $db->Query('SELECT * FROM "#tag" WHERE LOWER("TagName") = LOWER(?)', $tagName);
			else
				$recs = $db->Query('SELECT * FROM "#tag" WHERE "TagName" = ?', $tagName);
			if (!$recs) return null;
			$tags = array();
			foreach ($recs as $rec)
				$tags[] = self::FromRecord($rec);
			return $tags;
		}
		else
		{
			if ($caseless)
				$rec = $db->QueryRecord('SELECT * FROM "#tag" WHERE LOWER("TagName") = LOWER(?)', $tagName);
			else
				$rec = $db->QueryRecord('SELECT * FROM "#tag" WHERE "TagName" = ?', $tagName);
			if (!$rec) return null;
			return self::FromRecord($rec);
		}
	}

	// Finds tag instances by their tag meaning value.
	//
	// meaning = (int) Tag meaning value
	//
	// Throws InvalidArgumentException
	// Throws UnbDatabaseException
	//
	// Returns: An array of UnbTag objects
	//
	public static function FindByMeaning($meaning)
	{
		if (!is_int($meaning))
			throw new InvalidArgumentException('Invalid argument type: $meaning');

		$tag = array();
		$db = UnbDatabase::GetInstance();
		$recs = $db->Query('SELECT * FROM "#tag" WHERE "Meaning" = ?', $meaning);
		if (!$recs) return $tags;
		foreach ($recs as $rec)
			$tags[] = self::FromRecord($rec);
		return $tags;
	}

	// Finds all tags and returns them in their hierarchical order.
	//
	// withReadAccess = (bool) Filter tags for those where the session user has read access for.
	// withUseAccess = (bool) Filter tags for those where the session user has use access for.
	//
	// Throws UnbDatabaseException
	//
	// Returns: UnbTag objects for the found tags.
	//
	public static function Find($withReadAccess = false, $withUseAccess = false)
	{
		$sessionUserId = UnbSession::GetSessionUserId();

		if (!isset(self::$tagHierarchicalOrder))
			self::UpdateHierarchicalOrder();

		$sql = 'SELECT * FROM "#tag" t';

		$where = '';
		if ($withReadAccess)
		{
			$where .= ($where ? ' AND ' : ' WHERE ') . ' (
				-- Tag has no read access keylist
				t."ReadAccessKeylistId" IS NULL
				OR EXISTS
					-- Tag\'s read access keylist contains personal or additional user key
					(SELECT 1
					FROM "#keylist_keys" tkk
					WHERE tkk."KeylistId" = t."ReadAccessKeylistId" AND
						(tkk."UserId" = ? OR
							tkk."UserId" IN
								(SELECT ukk."UserId"
								FROM "#user" u, "#keylist_keys" ukk
								WHERE ukk."KeylistId" = u."AdditionalKeylistId" AND
									u."UserId" = ?))))';
			$parameters = array($sessionUserId, $sessionUserId);
		}
		if ($withUseAccess)
		{
			$where .= ($where ? ' AND ' : ' WHERE ') . ' (
				-- Tag has no use access keylist
				t."UseAccessKeylistId" IS NULL
				OR EXISTS
					-- Tag\'s use access keylist contains personal or additional user key
					(SELECT 1
					FROM "#keylist_keys" tkk
					WHERE tkk."KeylistId" = t."UseAccessKeylistId" AND
						(tkk."UserId" = ? OR
							tkk."UserId" IN
								(SELECT ukk."UserId"
								FROM "#user" u, "#keylist_keys" ukk
								WHERE ukk."KeylistId" = u."AdditionalKeylistId" AND
									u."UserId" = ?))))';
			$parameters = array($sessionUserId, $sessionUserId);
		}
		$sql .= $where;

		if (count(self::$tagHierarchicalOrder))
		{
			$sql .= ' ORDER BY CASE "TagId"';
			foreach (self::$tagHierarchicalOrder as $index => $tagId)
				$sql .= ' WHEN ' . $tagId . ' THEN ' . $index;
			$sql .= ' END';
		}

		$db = UnbDatabase::GetInstance();
		$recs = $db->Query($sql, $parameters);
		if (!isset($recs)) return null;
		$tags = array();
		foreach ($recs as $rec)
			$tags[] = self::FromRecord($rec);
		return $tags;
	}

	// Updates the cached correct hierarchical order of all tags.
	//
	private static function UpdateHierarchicalOrder()
	{
		self::$tagHierarchicalOrder = array();

		$db = UnbDatabase::GetInstance();
		$recs = $db->Query('SELECT * FROM "#tag" ORDER BY "TagName"');
		if (!$recs) return;
		$tags = array();
		foreach ($recs as $rec)
			$tags[] = self::FromRecord($rec);

		self::UpdateHierarchicalOrder_Rec($tags, 0, self::$tagHierarchicalOrder);
	}

	// Recursion function for UpdateHierarchicalOrder().
	//
	// tags = (array) Array of UnbTag all instances, ordered by tag name
	// parentId = (int) This iteration's parent tag ID
	// inout order = (array) Array of cached tag IDs that all tag IDs with the given parent will be appended
	//
	private static function UpdateHierarchicalOrder_Rec($tags, $parentId, &$order)
	{
		foreach ($tags as $tag)
		{
			if ($tag->GetParentTagId() == $parentId)
			{
				$order[] = $tag->GetTagId();
				self::UpdateHierarchicalOrder_Rec($tags, $tag->GetTagId(), $order);
			}
		}
	}

	// Gets the cached correct hierarchical order of all tags.
	//
	// If not yet done, the order will be determined and cached for future requests.
	//
	// Returns (array(int)) Array of tag IDs in the order they should be displayed
	//
	public static function GetHierarchicalOrder()
	{
		if (!isset(self::$tagHierarchicalOrder))
			self::UpdateHierarchicalOrder();
		return self::$tagHierarchicalOrder;
	}

	// Gets the tag ID.
	//
	// Returns: (int) Tag ID
	//
	public function GetTagId()
	{
		return $this->tagId;
	}

	// Gets the tag's displayed name.
	//
	// Returns: (string) Tag name
	//
	public function GetTagName()
	{
		return $this->tagName;
	}

	// Sets the tag's displayed name.
	//
	// tagName = (string) New tag name (max. 255 characters)
	//
	// Throws InvalidArgumentException
	// Throws LengthException
	// Throws UnbSecurityException if not logged in or no moderator.
	// Throws UnbDatabaseException
	//
	// Privileges: Must be logged in with moderator privilege
	//
	public function SetTagName($tagName)
	{
		if (!is_string($tagName))
			throw new InvalidArgumentException('Invalid argument type: $tagName');
		$tagName = trim($tagName);
		if (!strlen($tagName))
			throw new InvalidArgumentException('Tag name is empty');
		if (mb_strlen($tagName) > 255)
			throw new LengthException('[UNB#14] ' . UnbLocale::Translate('error:14'));   // Tag name is too long

		if (!UnbSession::IsModerator())
			throw new UnbSecurityException('You need the moderator privilege to set a tag name.');

		$this->db->StartTransaction();
		try
		{
			$this->db->Execute(
				'UPDATE "#tag" SET "TagName" = ? WHERE "TagId" = ?',
				array($tagName, $this->tagId));
			$this->db->Commit();
		}
		catch (Exception $ex)
		{
			$this->db->RollBack();
			throw $ex;
		}
		$this->tagName = $tagName;
	}

	// Gets the tag's assigned parent tag.
	//
	// Returns: (UnbTag) Parent tag
	//
	public function GetParentTag()
	{
		if ($this->parentTag == 0) return null;
		return self::FindByTagId($this->parentTag);
	}

	// Gets the ID of the tag's assigned parent tag.
	//
	// Returns: (int) Parent tag ID
	//
	public function GetParentTagId()
	{
		return $this->parentTag;
	}

	// Sets the tag's parent tag ID.
	//
	// parentTag = (int/UnbTag) New parent tag or its ID or {{null}}
	//
	// Throws InvalidArgumentException on invalid argument type or value.
	// Throws UnbSecurityException if not logged in or no moderator.
	// Throws UnbDatabaseException
	//
	// Privileges: Must be logged in with moderator privilege
	//
	public function SetParentTag($parentTag)
	{
		if (isset($parentTag))
		{
			if ($parentTag instanceof UnbTag)
				$parentTag = $parentTag->GetTagId();
			if (!is_int($parentTag))
				throw new InvalidArgumentException('Invalid argument type: $parentTag');
			if ($parentTag < 0)
				throw new InvalidArgumentException('Argument out of range: $parentTag');
			if ($parentTag == 0)
				$parentTag = null;
		}

		if (!UnbSession::IsModerator())
			throw new UnbSecurityException('You need the moderator privilege to set a parent tag.');

		$this->db->StartTransaction();
		try
		{
			$this->db->Execute(
				'UPDATE "#tag" SET "ParentTag" = ? WHERE "TagId" = ?',
				array($parentTag, $this->tagId));
			$this->db->Commit();
		}
		catch (Exception $ex)
		{
			$this->db->RollBack();
			throw $ex;
		}
		$this->parentTag = $parentTag;
	}

	// Gets the tag's meaning value.
	//
	// Returns: (int) Tag meaning value
	//
	public function GetMeaning()
	{
		return $this->meaning;
	}

	// Sets the tag's meaning value.
	//
	// meaning = (int) New tag meaning value
	//
	// Throws InvalidArgumentException on invalid argument type.
	// Throws UnbSecurityException if not logged in or no moderator.
	// Throws UnbDatabaseException
	//
	// Privileges: Must be logged in with moderator privilege
	//
	public function SetMeaning($meaning)
	{
		if (isset($meaning))
		{
			if (!is_int($meaning))
				throw new InvalidArgumentException('Invalid argument type: $meaning');
		}

		if (!UnbSession::IsModerator())
			throw new UnbSecurityException('You need the moderator privilege to set a tag meaning.');

		$this->db->StartTransaction();
		try
		{
			$this->db->Execute(
				'UPDATE "#tag" SET "Meaning" = ? WHERE "TagId" = ?',
				array($meaning, $this->tagId));
			$this->db->Commit();
		}
		catch (Exception $ex)
		{
			$this->db->RollBack();
			throw $ex;
		}
		$this->meaning = $meaning;
	}

	// Gets the tag's description.
	//
	// Returns: (string) Description
	//
	public function GetDescription()
	{
		return $this->description;
	}

	// Sets the tag's description.
	//
	// description = (string) New tag description or {{null}} (max. 255 characters)
	//
	// Throws InvalidArgumentException
	// Throws LengthException
	// Throws UnbSecurityException if not logged in or no moderator.
	// Throws UnbDatabaseException
	//
	// Privileges: Must be logged in with moderator privilege
	//
	public function SetDescription($description)
	{
		if (isset($description))
		{
			if (!is_string($description))
				throw new InvalidArgumentException('Invalid argument type: $description');
			$description = trim($description);
			if (mb_strlen($description) > 255)
				throw new LengthException('[UNB#15] ' . UnbLocale::Translate('error:15'));   // Description is too long
			if (!strlen($description))
				$description = null;
		}

		if (!UnbSession::IsModerator())
			throw new UnbSecurityException('You need the moderator privilege to set a tag description.');

		$this->db->StartTransaction();
		try
		{
			$this->db->Execute(
				'UPDATE "#tag" SET "Description" = ? WHERE "TagId" = ?',
				array($description, $this->tagId));
			$this->db->Commit();
		}
		catch (Exception $ex)
		{
			$this->db->RollBack();
			throw $ex;
		}
		$this->description = $description;
	}

	// Gets the tag's flags.
	//
	// Returns: (int) Flags
	//
	public function GetFlags()
	{
		return $this->flags;
	}

	// Sets the tag's flags.
	//
	// flags = (int) New flags. See the UnbTagFlag constants.
	//
	// Throws InvalidArgumentException on invalid argument type.
	// Throws UnbSecurityException if not logged in or no moderator.
	// Throws UnbDatabaseException
	//
	// Privileges: Must be logged in with moderator privilege
	//
	public function SetFlags($flags)
	{
		if (!is_int($flags))
			throw new InvalidArgumentException('Invalid argument type: $flags');

		if (!UnbSession::IsModerator())
			throw new UnbSecurityException('You need the moderator privilege to set tag flags.');

		$this->db->StartTransaction();
		try
		{
			$this->db->Execute(
				'UPDATE "#tag" SET "Flags" = ? WHERE "TagId" = ?',
				array($flags, $this->tagId));
			$this->db->Commit();
		}
		catch (Exception $ex)
		{
			$this->db->RollBack();
			throw $ex;
		}
		$this->flags = $flags;
	}

	// Gets a list of all keys that allow using this tag.
	//
	// Throws UnbDatabaseException
	//
	// Returns: (array(int)) Keys array
	//
	public function GetUseAccessKeys()
	{
		return $this->db->GetKeylistKeys('tag', 'UseAccessKeylistId', 'TagId', $this->tagId);
	}

	// Adds a key to this tag that allows using it.
	//
	// key = (int/UnbUser) Key to add
	//
	// Throws InvalidArgumentException on invalid argument type or value.
	// Throws UnbSecurityException if not logged in or no administrator.
	// Throws UnbDatabaseException
	//
	// See UnbDatabase::AddKeylistKey()
	//
	// Privileges: Must be logged in with administrator privilege
	//
	public function AddUseAccessKey($key)
	{
		if (!UnbSession::IsAdministrator())
			throw new UnbSecurityException('You need the administrator privilege to add use access keys.');

		$this->db->AddKeylistKey('#tag', 'UseAccessKeylistId', 'TagId', $this->tagId, $key);
	}

	// Removes a key from this tag that allows using it.
	//
	// key = (int/UnbUser) Key to remove
	//
	// Throws InvalidArgumentException on invalid argument type or value.
	// Throws UnbSecurityException if not logged in or no administrator.
	// Throws UnbDatabaseException
	//
	// See UnbDatabase::RemoveKeylistKey()
	//
	// Privileges: Must be logged in with administrator privilege
	//
	public function RemoveUseAccessKey($key)
	{
		if (!UnbSession::IsAdministrator())
			throw new UnbSecurityException('You need the administrator privilege to remove use access keys.');

		$this->db->RemoveKeylistKey('#tag', 'UseAccessKeylistId', 'TagId', $this->tagId, $key);
	}

	// Sets all keys for this tag that allow using it. This removes any keys set before.
	//
	// keys = (array(int/UnbUser)) Keys to set
	//
	// Throws InvalidArgumentException on invalid argument type or value.
	// Throws UnbSecurityException if not logged in or no administrator.
	// Throws UnbDatabaseException
	//
	// See UnbDatabase::ResetKeylistKeys()
	//
	// Privileges: Must be logged in with administrator privilege
	//
	public function SetUseAccessKeys($keys)
	{
		if (!UnbSession::IsAdministrator())
			throw new UnbSecurityException('You need the administrator privilege to set use access keys.');

		$this->db->ResetKeylistKeys('#tag', 'UseAccessKeylistId', 'TagId', $this->tagId, $keys);
	}

	// Gets a list of all keys that allow reading messages with this tag set.
	//
	// Throws UnbDatabaseException
	//
	// Returns: (array(int)) Keys array
	//
	public function GetReadAccessKeys()
	{
		return $this->db->GetKeylistKeys('tag', 'ReadAccessKeylistId', 'TagId', $this->tagId);
	}

	// Adds a key to this tag that allows reading it.
	//
	// key = (int/UnbUser) Key to add
	//
	// Throws InvalidArgumentException on invalid argument type or value.
	// Throws UnbSecurityException if not logged in or no administrator.
	// Throws UnbDatabaseException
	//
	// See UnbDatabase::AddKeylistKey()
	//
	// Privileges: Must be logged in with administrator privilege
	//
	public function AddReadAccessKey($key)
	{
		if (!UnbSession::IsAdministrator())
			throw new UnbSecurityException('You need the administrator privilege to add read access keys.');

		$this->db->AddKeylistKey('#tag', 'ReadAccessKeylistId', 'TagId', $this->tagId, $key);
	}

	// Removes a key from this tag that allows reading it.
	//
	// key = (int/UnbUser) Key to remove
	//
	// Throws InvalidArgumentException on invalid argument type or value.
	// Throws UnbSecurityException if not logged in or no administrator.
	// Throws UnbDatabaseException
	//
	// See UnbDatabase::RemoveKeylistKey()
	//
	// Privileges: Must be logged in with administrator privilege
	//
	public function RemoveReadAccessKey($key)
	{
		if (!UnbSession::IsAdministrator())
			throw new UnbSecurityException('You need the administrator privilege to remove read access keys.');

		$this->db->RemoveKeylistKey('#tag', 'ReadAccessKeylistId', 'TagId', $this->tagId, $key);
	}

	// Sets all keys for this tag that allow reading it. This removes any keys set before.
	//
	// keys = (array(int/UnbUser)) Keys to set
	//
	// Throws InvalidArgumentException on invalid argument type or value.
	// Throws UnbSecurityException if not logged in or no administrator.
	// Throws UnbDatabaseException
	//
	// See UnbDatabase::ResetKeylistKeys()
	//
	// Privileges: Must be logged in with administrator privilege
	//
	public function SetReadAccessKeys($keys)
	{
		if (!UnbSession::IsAdministrator())
			throw new UnbSecurityException('You need the administrator privilege to set read access keys.');

		$this->db->ResetKeylistKeys('#tag', 'ReadAccessKeylistId', 'TagId', $this->tagId, $keys);
	}

	// Gets a value indicating whether using this tag is allowed to the current session user.
	//
	// Throws UnbDatabaseException
	//
	// Returns: (bool)
	//
	public function IsUseAllowed()
	{
		return self::IsUseAllowedById($this->tagId);
	}

	// Gets a value indicating whether using a tag specified by its tag ID is allowed to
	// the current session user.
	//
	// tagId = (int/UnbTag) Tag or its ID to check.
	//
	// Throws InvalidArgumentException
	// Throws UnbDatabaseException
	//
	// Returns: (bool)
	//
	public static function IsUseAllowedById($tagId)
	{
		if ($tagId instanceof UnbTag)
			$tagId = $tagId->GetTagId();
		if (!is_int($tagId))
			throw new InvalidArgumentException('Invalid argument type: $tagId');

		$sessionUserId = UnbSession::GetSessionUserId();

		$sql = 'SELECT
			(EXISTS
				-- Tag has no use access keylist
				(SELECT 1
				FROM "#tag" t
				WHERE t."TagId" = ? AND
					t."UseAccessKeylistId" IS NULL)
			OR EXISTS
				-- Tag\'s use access keylist contains personal or additional user key
				(SELECT 1
				FROM "#tag" t, "#keylist_keys" tkk
				WHERE tkk."KeylistId" = t."UseAccessKeylistId" AND
					t."TagId" = ? AND
					(tkk."UserId" = ? OR
						tkk."UserId" IN
							(SELECT ukk."UserId"
							FROM "#user" u, "#keylist_keys" ukk
							WHERE ukk."KeylistId" = u."AdditionalKeylistId" AND
								u."UserId" = ?))))';
		$parameters = array($tagId, $tagId, $sessionUserId, $sessionUserId);

		$db = UnbDatabase::GetInstance();
		$val = $db->QueryValue($sql, $parameters);
		$val = $val ? true : false;
		return $val;
	}

	// Gets the tag data that is assigned to a message.
	//
	// Returns: (string) Tag data
	//
	public function GetTagData()
	{
		return $this->tagData;
	}

	// Deletes this instance of the tag with all its references.
	//
	// Throws UnbSecurityException if not logged in or no moderator.
	// Throws UnbDatabaseException
	//
	// Returns: {{true}}, if successful, {{false}} otherwise.
	//
	// Privileges: Must be logged in with moderator privilege
	//
	public function Delete()
	{
		if (!UnbSession::IsModerator())
			throw new UnbSecurityException('You need the moderator privilege to delete a tag.');

		$this->db->StartTransaction();
		try
		{
			// Deleting tags may leave keylists unreferenced. Delete them now.
			$this->db->Execute(
				'DELETE FROM "#keylist"
				WHERE "KeylistId" =
					(SELECT "UseAccessKeylistId"
					FROM "#tag"
					WHERE "TagId" = ?)',
				$this->tagId);
			$this->db->Execute(
				'DELETE FROM "#keylist"
				WHERE "KeylistId" =
					(SELECT "ReadAccessKeylistId"
					FROM "#tag"
					WHERE "TagId" = ?)',
				$this->tagId);

			$this->db->Execute('DELETE FROM "#tag" WHERE "TagId" = ?', $this->tagId);
			$this->db->Commit();
		}
		catch (Exception $ex)
		{
			$this->db->RollBack();
			throw $ex;
		}
	}

	// Determines the differences between two sets of tags.
	//
	// tags1 = (array) First tag set. {{null}} is accepted for an empty set.
	// tags2 = (array) Second tag set. {{null}} is accepted for an empty set.
	// out addedTags = (array) Array to store all tags that were added in the second set
	// out removedTags = (array) Array to store all tags that were removed in the second set
	// out changedTags = (array) Array to store all tags whose user data was changed in the second set
	//
	// Throws InvalidArgumentException on invalid argument type or value.
	//
	public static function GetDifferences($tags1, $tags2, &$addedTags, &$removedTags, &$changedTags)
	{
		if (!isset($tags1)) $tags1 = array();
		if (!is_array($tags1))
			throw new InvalidArgumentException('Invalid argument type: $tags1');
		if (!isset($tags2)) $tags2 = array();
		if (!is_array($tags2))
			throw new InvalidArgumentException('Invalid argument type: $tags2');

		$addedTags = array();
		$removedTags = array();
		$changedTags = array();

		foreach ($tags1 as $tag1)
		{
			$found = false;
			foreach ($tags2 as $tag2)
			{
				if ($tag1->GetTagId() === $tag2->GetTagId())
				{
					$found = true;
					if (($tag1->GetFlags() & UnbTagFlag::UserData) &&
						($tag1->GetTagData() !== $tag2->GetTagData()))
					{
						$changedTags[] = $tag2;
					}
					break;
				}
			}
			if (!$found)
			{
				$removedTags[] = $tag1;
			}
		}

		foreach ($tags2 as $tag2)
		{
			$found = false;
			foreach ($tags1 as $tag1)
			{
				if ($tag1->GetTagId() === $tag2->GetTagId())
				{
					$found = true;
					break;
				}
			}
			if (!$found)
			{
				$addedTags[] = $tag2;
			}
		}
	}

} // class UnbTag

?>
